/******************************************************************************
 *
 * AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: 
 * Simon Vey, Thomas Witkowski, Andreas Naumann, Simon Praetorius, et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * This file is part of AMDiS
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/



/** \file SurfaceOperator.h */

#ifndef AMDIS_SURFACEOPERATOR_H
#define AMDIS_SURFACEOPERATOR_H

#include "FiniteElemSpace.h"
#include "Mesh.h"
#include "Operator.h"
#include "SurfaceQuadrature.h"

namespace AMDiS {

  /** 
   * \ingroup Integration
   *
   * \brief
   * A SurfaceOperator is a Operator used for surface integration. Instead of a 
   * normal Quadrature it uses a SurfaceQuadrature with integration points at
   * one element side instead of integration points in the inner of the element.
   * The determinant in ElInfo is replaced by the surface determinant, so the
   * assemblage can be done with the standard Assembler classes.
   * The SurfaceQuadrature is used for the implementation of Neumann and Robin
   * boundary conditions.
   */
  class SurfaceOperator : public Operator
  {
  public:
    /// Creates a SurfaceOperator conforming to operat for the given \ref coords.
    SurfaceOperator(Operator *operat, 
		    VectorOfFixVecs<DimVec<double> > &coords) 
      : Operator(*operat), 
	coords_(coords),
	quad2(nullptr),
	quad1GrdPsi(nullptr),
	quad1GrdPhi(nullptr),
	quad0(nullptr)
    {
      assembler = nullptr;

      int dim = rowFeSpace->getMesh()->getDim();
      int degree;

      // create surface quadratures
      if (secondOrder.size() > 0) {
	degree = getQuadratureDegree(2);
	quad2 = 
	  new SurfaceQuadrature(Quadrature::provideQuadrature(dim - 1, degree), coords);
      }

      if (firstOrderGrdPsi.size() > 0) {
	degree = getQuadratureDegree(1, GRD_PSI);
	quad1GrdPsi = 
	  new SurfaceQuadrature(Quadrature::provideQuadrature(dim - 1, degree), coords);
      }

      if (firstOrderGrdPhi.size() > 0) {
	degree = getQuadratureDegree(1, GRD_PHI);
	quad1GrdPhi = 
	  new SurfaceQuadrature(Quadrature::provideQuadrature(dim - 1, degree), coords);
      }

      if (zeroOrder.size() > 0) {
	degree = getQuadratureDegree(0);
	quad0 = 
	  new SurfaceQuadrature(Quadrature::provideQuadrature(dim - 1, degree), coords);
      }

      // initialize assembler with surface quadratures
      optimized = false;
      initAssembler(quad2, quad1GrdPsi, quad1GrdPhi, quad0);
    }

    /// Adapt surface quadratures to \ref coords.
    void adaptSurfaceOperator(VectorOfFixVecs<DimVec<double> > &coords)
    {
      coords_ = coords;

      if (quad2)
	quad2->scaleSurfaceQuadrature(coords);
      if (quad1GrdPsi)
	quad1GrdPsi->scaleSurfaceQuadrature(coords);
      if (quad1GrdPhi)
	quad1GrdPhi->scaleSurfaceQuadrature(coords);
      if (quad0)
	quad0->scaleSurfaceQuadrature(coords);
    }

    /** \brief
     * Implementation of \ref Operator::getElementMatrix(). Repalces the 
     * determinant by the surface determinant and deligates the call to
     * the base class function.
     */
    virtual void getElementMatrix(const ElInfo *elInfo, 
				  ElementMatrix& userMat, 
				  double factor = 1.0)
    {
      int dim = rowFeSpace->getMesh()->getDim();
      double origDet = elInfo->getDet();

      FixVec<WorldVector<double>, VERTEX> worldCoords(dim-1, NO_INIT);

      // transform barycentric coords to world coords
      for (int i = 0; i < dim; i++)
	elInfo->coordToWorld(coords_[i], worldCoords[i]);

      // set determinant for world coords of the side
      const_cast<ElInfo*>(elInfo)->setDet(elInfo->calcDet(worldCoords));

      // calc element matrix
      Operator::getElementMatrix(elInfo, userMat, factor);

      // set determinant for world coords of the side
      const_cast<ElInfo*>(elInfo)->setDet(origDet);
    }
  
    /** \brief
     * Implementation of \ref Operator::getElementVector(). Repalces the 
     * determinant by the surface determinant and deligates the call to
     * the base class function.
     */
    virtual void getElementVector(const ElInfo *elInfo, 
				  ElementVector& userVec, 
				  double factor = 1.0) 
    {
      int dim = rowFeSpace->getMesh()->getDim();
      double origDet = elInfo->getDet();

      FixVec<WorldVector<double>, VERTEX> worldCoords(dim-1, NO_INIT);

      // transform barycentric coords to world coords
      for (int i = 0; i < dim; i++)
	elInfo->coordToWorld(coords_[i], worldCoords[i]);

      // set determinant for world coords of the side
      const_cast<ElInfo*>(elInfo)->setDet(elInfo->calcDet(worldCoords));

      // calc element vector
      Operator::getElementVector(elInfo, userVec, factor);

      const_cast<ElInfo*>(elInfo)->setDet(origDet);
    }

  protected:
    VectorOfFixVecs<DimVec<double> > coords_;

    /// Surface quadratures
    SurfaceQuadrature *quad2;
    SurfaceQuadrature *quad1GrdPsi;
    SurfaceQuadrature *quad1GrdPhi;
    SurfaceQuadrature *quad0;
  };

}

#endif
