/******************************************************************************
 *
 * AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: 
 * Simon Vey, Thomas Witkowski, Andreas Naumann, Simon Praetorius, et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * This file is part of AMDiS
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/


#include <vector>
#include <boost/numeric/mtl/mtl.hpp>
#include "Assembler.h"
#include "ZeroOrderAssembler.h"
#include "Operator.h"
#include "QPsiPhi.h"
#include "FiniteElemSpace.h"
#include "Quadrature.h"
#include "DOFVector.h"

using namespace std;

namespace AMDiS {

  ThreadPrivate<vector<SubAssembler*> > 
  ZeroOrderAssembler::optimizedSubAssemblers;
  
  ThreadPrivate<vector<SubAssembler*> >
  ZeroOrderAssembler::standardSubAssemblers;


  ZeroOrderAssembler::ZeroOrderAssembler(Operator *op,
					 Assembler *assembler,
					 Quadrature *quad,
					 bool optimized)
    : SubAssembler(op, assembler, quad, 0, optimized)
  {}


  ZeroOrderAssembler* ZeroOrderAssembler::getSubAssembler(Operator* op,
							  Assembler *assembler,
							  Quadrature *quad,
							  bool optimized)
  {
    // check if an assembler is needed at all
    if (op->zeroOrder.size() == 0)
      return nullptr;   

    ZeroOrderAssembler *newAssembler;

    vector<SubAssembler*> subAssemblers =
      optimized ? optimizedSubAssemblers.get() : standardSubAssemblers.get();

    vector<OperatorTerm*> opTerms = op->zeroOrder;

    sort(opTerms.begin(), opTerms.end());

    // check if a new assembler is needed
    if (quad) {
      for (int i = 0; i < static_cast<int>(subAssemblers.size()); i++) {
	vector<OperatorTerm*> assTerms = *(subAssemblers[i]->getTerms());

	sort(assTerms.begin(), assTerms.end());

	if (opTerms == assTerms && subAssemblers[i]->getQuadrature() == quad)
	  return dynamic_cast<ZeroOrderAssembler*>(subAssemblers[i]);
      }
    }
 
    // check if all terms are pw_const
    bool pwConst = true;
    for (unsigned int i = 0; i < op->zeroOrder.size(); i++) {
      if (!op->zeroOrder[i]->isPWConst()) {
	pwConst = false;
	break;
      }
    }  

    // create new assembler
    if (!optimized) {
      newAssembler = new StandardZOA(op, assembler, quad);
    } else {
      if (pwConst)
	newAssembler = new PrecalcZOA(op, assembler, quad);
      else
     	newAssembler = new FastQuadZOA(op, assembler, quad);      
    }

    subAssemblers.push_back(newAssembler);

    return newAssembler;
  }


  StandardZOA::StandardZOA(Operator *op, Assembler *assembler, Quadrature *quad)
    : ZeroOrderAssembler(op, assembler, quad, false)      
  {
    name = "standard zero order assembler";
  }


  void StandardZOA::calculateElementMatrix(const ElInfo *elInfo, 
					   ElementMatrix& mat)
  {
    const BasisFunction *psi = rowFeSpace->getBasisFcts();
    const BasisFunction *phi = colFeSpace->getBasisFcts();

    int nPoints = quadrature->getNumPoints();
    ElementVector c(nPoints, 0.0);
    vector<double> phival(nCol);

    for (vector<OperatorTerm*>::iterator termIt = terms.begin(); 
	 termIt != terms.end(); ++termIt)
      (static_cast<ZeroOrderTerm*>((*termIt)))->getC(elInfo, nPoints, c);

    if (symmetric) {
      TEST_EXIT_DBG(nCol == nRow)("nCol != nRow, but symmetric assembling!\n");
            
      for (int iq = 0; iq < nPoints; iq++) {
	c[iq] *= elInfo->getDet();

	// calculate phi at QPs only once!
	for (int i = 0; i < nCol; i++)
	  phival[i] = (*(phi->getPhi(i)))(quadrature->getLambda(iq));

	for (int i = 0; i < nRow; i++) {
	  double psival = (*(psi->getPhi(i)))(quadrature->getLambda(iq));
	  mat[i][i] += quadrature->getWeight(iq) * c[iq] * psival * phival[i];
	  for (int j = i + 1; j < nCol; j++) {
	    double val = quadrature->getWeight(iq) * c[iq] * psival * phival[j];
	    mat[i][j] += val;
	    mat[j][i] += val;
	  }
	}
      }
    } else {      //  non symmetric assembling 
      for (int iq = 0; iq < nPoints; iq++) {
	c[iq] *= elInfo->getDet();

	// calculate phi at QPs only once!
	for (int i = 0; i < nCol; i++)
	  phival[i] = (*(phi->getPhi(i)))(quadrature->getLambda(iq));

	for (int i = 0; i < nRow; i++) {
	  double psival = (*(psi->getPhi(i)))(quadrature->getLambda(iq));
	  for (int j = 0; j < nCol; j++)
	    mat[i][j] += quadrature->getWeight(iq) * c[iq] * psival * phival[j];	  
	}
      }
    }
  }


  void StandardZOA::calculateElementVector(const ElInfo *elInfo, ElementVector& vec)
  {
    int nPoints = quadrature->getNumPoints();
    ElementVector c(nPoints, 0.0);

    vector<OperatorTerm*>::iterator termIt;
    for (termIt = terms.begin(); termIt != terms.end(); ++termIt)
      (static_cast<ZeroOrderTerm*>((*termIt)))->getC(elInfo, nPoints, c);

    for (int iq = 0; iq < nPoints; iq++) {
      c[iq] *= elInfo->getDet();

      for (int i = 0; i < nRow; i++) {
	double psi = (*(rowFeSpace->getBasisFcts()->getPhi(i)))
	  (quadrature->getLambda(iq));
	vec[i] += quadrature->getWeight(iq) * c[iq] * psi;
      }
    }
  }


  FastQuadZOA::FastQuadZOA(Operator *op, Assembler *assembler, Quadrature *quad)
    : ZeroOrderAssembler(op, assembler, quad, true)
  {
    name = "fast quadrature zero order assembler";
  }


  void FastQuadZOA::calculateElementMatrix(const ElInfo *elInfo, ElementMatrix& mat)
  {
    int nPoints = quadrature->getNumPoints();

    if (firstCall) {
      const BasisFunction *basFcts = rowFeSpace->getBasisFcts();
      psiFast = updateFastQuadrature(psiFast, basFcts, INIT_PHI);
      basFcts = colFeSpace->getBasisFcts();
      phiFast = updateFastQuadrature(phiFast, basFcts, INIT_PHI);
      firstCall = false;      
    }

    if (num_rows(c) < static_cast<unsigned int>(nPoints))
      c.change_dim(nPoints);
    c = 0.0;

    for (vector<OperatorTerm*>::iterator termIt = terms.begin(); 
	 termIt != terms.end(); ++termIt)
      (static_cast<ZeroOrderTerm*>((*termIt)))->getC(elInfo, nPoints, c);

    const mtl::dense2D<double>& psi = psiFast->getPhi();
    const mtl::dense2D<double>& phi = phiFast->getPhi();

    if (symmetric) {
      TEST_EXIT_DBG(nCol == nRow)("nCol != nRow, but symmetric assembling!\n");

      for (int iq = 0; iq < nPoints; iq++) {
	c[iq] *= elInfo->getDet() * quadrature->getWeight(iq);

	for (int i = 0; i < nRow; i++) {
	  mat[i][i] += c[iq] * psi[iq][i] * phi[iq][i];
	  for (int j = i + 1; j < nCol; j++) {
	    double val = c[iq] * psi[iq][i] * phi[iq][j];
	    mat[i][j] += val;
	    mat[j][i] += val;
	  }
	}
      }
    } else {      /*  non symmetric assembling   */
      for (int iq = 0; iq < nPoints; iq++)
	c[iq] *= elInfo->getDet() * quadrature->getWeight(iq);

      for (int i = 0; i < nRow; i++)
	for (int j = 0; j < nCol; j++) {
	  double v = 0.0;
	  for (int iq = 0; iq < nPoints; iq++)
	    v += c[iq] * psi[iq][i] * phi[iq][j];
	  mat[i][j] += v;
	}
    }
  }


  void FastQuadZOA::calculateElementVector(const ElInfo *elInfo, ElementVector& vec)
  {
    int nPoints = quadrature->getNumPoints();

    if (firstCall) {
      const BasisFunction *basFcts = rowFeSpace->getBasisFcts();
      psiFast = updateFastQuadrature(psiFast, basFcts, INIT_PHI);
      basFcts = colFeSpace->getBasisFcts();
      phiFast = updateFastQuadrature(phiFast, basFcts, INIT_PHI);
      firstCall = false;      
    }

    ElementVector c(nPoints, 0.0);
    vector<OperatorTerm*>::iterator termIt;
    for (termIt = terms.begin(); termIt != terms.end(); ++termIt)
      (static_cast<ZeroOrderTerm*>((*termIt)))->getC(elInfo, nPoints, c);

    const mtl::dense2D<double>& psi = psiFast->getPhi();

    for (int iq = 0; iq < nPoints; iq++) {
      c[iq] *= elInfo->getDet();

      for (int i = 0; i < nRow; i++)
	vec[i] += quadrature->getWeight(iq) * c[iq] * psi[iq][i];
    }
  }


  PrecalcZOA::PrecalcZOA(Operator *op, Assembler *assembler, Quadrature *quad) 
    : ZeroOrderAssembler(op, assembler, quad, true)
  {
    name = "precalculated zero order assembler";
  }


  void PrecalcZOA::calculateElementMatrix(const ElInfo *elInfo, 
					  ElementMatrix& mat)
  {
    if (firstCall) {
      q00 = Q00PsiPhi::provideQ00PsiPhi(rowFeSpace->getBasisFcts(), 
					colFeSpace->getBasisFcts(), 
					quadrature);
      q0 = Q0Psi::provideQ0Psi(rowFeSpace->getBasisFcts(), quadrature);
      firstCall = false;      
    }

    ElementVector c(1, 0.0);
    int size = static_cast<int>(terms.size());

    for (int i = 0; i < size; i++)
      (static_cast<ZeroOrderTerm*>((terms[i])))->getC(elInfo, 1, c);    

    c[0] *= elInfo->getDet();
    
    if (symmetric) {
      for (int i = 0; i < nRow; i++) {
    	mat[i][i] += c[0] * q00->getValue(i,i);
    	for (int j = i + 1; j < nCol; j++) {
    	  double val = c[0] * q00->getValue(i, j);
    	  mat[i][j] += val;
    	  mat[j][i] += val;
    	}
      }
    } else {
      for (int i = 0; i < nRow; i++)
	for (int j = 0; j < nCol; j++)
	  mat[i][j] += c[0] * q00->getValue(i, j);
    }
  }


  void PrecalcZOA::calculateElementVector(const ElInfo *elInfo, 
					  ElementVector& vec)
  {
    if (firstCall) {
      q00 = Q00PsiPhi::provideQ00PsiPhi(rowFeSpace->getBasisFcts(), 
					colFeSpace->getBasisFcts(), 
					quadrature);
      q0 = Q0Psi::provideQ0Psi(rowFeSpace->getBasisFcts(), quadrature);	
      firstCall = false;      
    }

    vector<OperatorTerm*>::iterator termIt;

    ElementVector c(1, 0.0);
    for (termIt = terms.begin(); termIt != terms.end(); ++termIt)
      (static_cast<ZeroOrderTerm*>(*termIt))->getC(elInfo, 1, c);

    c[0] *= elInfo->getDet();

    for (int i = 0; i < nRow; i++)
      vec[i] += c[0] * q0->getValue(i);
  }

}
