/******************************************************************************
 *
 * AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: 
 * Simon Vey, Thomas Witkowski, Andreas Naumann, Simon Praetorius, et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * This file is part of AMDiS
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/



/** \file FileWriter.h */

/** \defgroup Output Output module
 * @{ <img src="output.png"> @}
 */

#ifndef AMDIS_FILEWRITER_H
#define AMDIS_FILEWRITER_H

#include <vector>
#include <string>
#include "AMDiS_fwd.h"
#include "Mesh.h"
#include "DataCollector.h"
#include "FileCompression.h"

namespace AMDiS {

  class FileWriterInterface
  {
  public:
    FileWriterInterface()
      : filename(""),
	traverseLevel(-1),
	traverseFlag(Mesh::CALL_LEAF_EL),
	writeElement(NULL)
    {}

    virtual ~FileWriterInterface() {}

    /** \brief
     * Interface. Must be overridden in subclasses.
     * \param time time index of solution std::vector.
     * \param force enforces the output operation for the last timestep.
     */
    virtual void writeFiles(AdaptInfo *adaptInfo, bool force,
			    int level = -1,
			    Flag traverseFlag = Mesh::CALL_LEAF_EL,
			    bool (*writeElem)(ElInfo*) = NULL) = 0;

    void setTraverseProperties(int level, 
			       Flag flag,
			       bool (*writeElem)(ElInfo*))
    {
      traverseLevel = level;
      traverseFlag |= flag;
      writeElement = writeElem;
    }

    std::string getFilename() 
    { 
      return filename; 
    }
    
    void setFilename(std::string n) 
    { 
      filename = n; 
    }

  protected:
    /// Used filename prefix.
    std::string filename;

    int traverseLevel;

    Flag traverseFlag;

    bool (*writeElement)(ElInfo*);  
  };
  
  namespace detail
  {

    /**  
    * \ingroup Output
    *
    * \brief
    * Base class of FileWriterScal and FileWriterVec. Manages the file
    * output of solution vectors.
    */
    template<typename T>
    class FileWriter : public ::AMDiS::FileWriterInterface
    {
    public:
      /// Constructor for a filewriter for one data component.
      FileWriter(std::string name, Mesh *mesh, DOFVector<T> *vec);

      /// Constructor for a filewriter with more than one data component.
      FileWriter(std::string name,
			  Mesh *mesh,
			  std::vector< DOFVector<T>* > vecs);

      /// Constructor for a filewriter with more than one data component.
      FileWriter(std::string name,
			  Mesh *mesh,
			  SystemVector *vecs);
      
      /// Destructor
      virtual ~FileWriter();

      /// Implementation of FileWriterInterface::writeFiles().
      virtual void writeFiles(AdaptInfo *adaptInfo, bool force,
			      int level = -1,
			      Flag traverseFlag = Mesh::CALL_LEAF_EL,
			      bool (*writeElem)(ElInfo*) = NULL);
    
      std::vector<std::pair<double, std::string> >& getParaviewAnimationFrames()
      {
	return paraviewAnimationFrames;
      }

    protected:
      /// Initialization of the filewriter.
      void initialize();

      /// Reads all file writer dependend parameters from the init file.
      void readParameters();

      /// Name of the writer.
      std::string name;

      /// AMDiS mesh-file extension.
      std::string amdisMeshExt;

      /// AMDiS solution-file extension.
      std::string amdisDataExt;

      /// VTK file extension.
      std::string paraviewFileExt;

      /// Parallel VTK file extension.
      std::string paraviewParallelFileExt;

      /// Periodic file extension.
      std::string periodicFileExt;

      /// 0: Don't write AMDiS files; 1: Write AMDiS files. 
      int writeAMDiSFormat;

      /// 0: Don't write ParaView files; 1: Write ParaView files.
      int writeParaViewFormat;

      /// 0: Don't write ParaView std::vector files; 1: Write ParaView std::vector files.
      int writeParaViewVectorFormat;

      /// 1: extend number of component to 3, so that paraview can display the std::vector as worldstd::vector
      bool writeAs3dVector;

      /// 0: Don't write ParaView animation file; 1: Write ParaView animation file.
      int writeParaViewAnimation;

      /// 0: Don't write periodic files; 1: Write periodic files.
      int writePeriodicFormat;

      /// 0: Don't write png files; 1: Write png image files.
      int writePngFormat;

      /// 0: Gray color picture; 1: RGB picture.
      int pngType;

      /// 0: Don't write Povray scripts; 1: Write Povray scripts
      int writePovrayFormat;

      /// 0: Don't write DOF files; 1: Write DOF files
      int writeDofFormat;

      /// 0: Don't write ARH files; 1: Write ARH files
      int writeArhFormat;

      /// 0: Don't write ARH2 files; 1: Write ARH2 files
      int writeArh2Format;

      /// camera position for povray script files
      std::string povrayCameraLocation;

      /// orientation for camera in povray script files
      std::string povrayCameraLookAt;

      /// name of the template file that will be prepended to all created *.pov files 
      std::string povrayTemplate;

      /// 0: Don't append time index to filename prefix.
      /// 1: Append time index to filename prefix.
      int appendIndex;

      /// Total length of appended time index.
      int indexLength;

      /// Number of decimals in time index.
      int indexDecimals;

      /// Timestep modulo: write only every tsModulo-th timestep! 
      int tsModulo;
      
      /// Time modulo: write at first iteration after lastWriteTime + timeModulo
      double timeModulo;
      double lastWriteTime;

      /// Stores a set of std::pairs of timepoint and filename to write a ParaView 
      /// animation file.
      std::vector<std::pair<double, std::string> > paraviewAnimationFrames;

      ///
      int timestepNumber;

      /// Mesh used for output.
      Mesh *mesh;

      /// fespace used for output.
      const FiniteElemSpace *feSpace;

      /// Pointers to the std::vectors which store the solution.
      std::vector<DOFVector<T>*> solutionVecs;

      /** \brief
      * Stores the number of temporal solutions std::vectors, which have been created
      * in the constructor. If this number is greater than zero, the std::vectors
      * stored in solutionVecs_ must be deleted in the destructor.
      */
      int nTmpSolutions;

      /** \brief
      * Defines if, and with what kind of compression, the file should be compressed
      * during writing.
      */
      FileCompression compression;
    };

    template<>
    FileWriter<double>::FileWriter(std::string name_,
			  Mesh *mesh_,
			  SystemVector *vecs);

    template<>
    void FileWriter<double>::writeFiles(AdaptInfo *adaptInfo,
				bool force,
				int level,
				Flag flag,
				bool (*writeElem)(ElInfo*));
    
  } // end namespace detail
  
  typedef detail::FileWriter<double> FileWriter;
  typedef detail::FileWriter<WorldVector<double> > FileVectorWriter;
}

#include "FileWriter.hh"

#endif // AMDIS_FILEWRITER_H
