/******************************************************************************
 *
 * AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: 
 * Simon Vey, Thomas Witkowski, Andreas Naumann, Simon Praetorius, et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * This file is part of AMDiS
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/


#include <stdio.h>
#include <string>
#include <vector>
#include <fstream>
#include <sstream>
#include <cmath>

#include "boost/filesystem/operations.hpp"
#include "boost/filesystem/convenience.hpp"

#include "boost/iostreams/filtering_stream.hpp"
#include "boost/iostreams/device/file_descriptor.hpp"
#ifdef HAVE_COMPRESSION
#include "boost/iostreams/filter/gzip.hpp"
#include "boost/iostreams/filter/bzip2.hpp"
#endif

#include "AdaptInfo.h"
#include "VtkWriter.h"

namespace AMDiS  { namespace io {
  
    namespace VtkWriter
    {
      
      using namespace std;

      int Aux::writeFile(string name)
      {
	FUNCNAME("VtkWriter::writeFile()");
	
	boost::iostreams::filtering_ostream file;
#ifdef HAVE_COMPRESSION
	switch (compress) {
	case GZIP:
	  file.push(boost::iostreams::gzip_compressor());
	  name.append(".gz");
	  break;
	case BZIP2:
	  file.push(boost::iostreams::bzip2_compressor());
	  name.append(".bz2");
	  break;
	default:
	  break;
	}
#endif
	{
	  ofstream swapfile(name.c_str(), ios::out | ios::trunc);
	  TEST_EXIT(swapfile.is_open())
	    ("Cannot open file %s for writing!\n", name.c_str());
	  swapfile.close();
	}

	file.push(boost::iostreams::file_descriptor_sink(name, ios::trunc));
	writeFileToStream(file);

	return 0;
      }
      
// __________________________________________________________________________ //
      
      namespace detail 
      {

	int updateAnimationFile(AdaptInfo *adaptInfo,
				string valueFilename,
				vector< pair<double, string> > *paraViewAnimationFrames,
				string animationFilename)
	{
	  FUNCNAME("updateAnimationFile()");

	  size_t found = valueFilename.find_last_of("/\\");
	  paraViewAnimationFrames->push_back(make_pair(adaptInfo->getTime(),
						      valueFilename.substr(found + 1)));

	  boost::iostreams::filtering_ostream file;
	  {
	    ofstream swapfile(animationFilename.c_str(), 
				  ios::out | ios::trunc);
	    TEST_EXIT(swapfile.is_open())
	      ("Cannot open file %s for writing!\n", animationFilename.c_str());
	    swapfile.close();
	  }
	  file.push(boost::iostreams::file_descriptor_sink(animationFilename, 
							  ios::trunc)); 

	  file << "<?xml version=\"1.0\"?>\n";
	  file << "<VTKFile type=\"Collection\" version=\"0.1\" >"  << "\n";
	  file << "<Collection>\n";

	  for (vector<pair<double, string> >::iterator it = paraViewAnimationFrames->begin();
	      it < paraViewAnimationFrames->end(); ++it) {
	    file << "<DataSet timestep=\"" << it->first
		<< "\" part=\"0\" file=\"" << it->second << "\"/>\n";
	  }

	  file << "</Collection>\n";
	  file << "</VTKFile>\n";

	  return 0;
	}
	

	void writeParallelFile(string name, int nRanks,
			      string fnPrefix, string fnPostfix,
			      vector<string> &componentNames)
	{
	  FUNCNAME("writeParallelFile()");

	  using boost::lexical_cast;

	  boost::iostreams::filtering_ostream file;
	  {
	    ofstream swapfile(name.c_str(), ios::out | ios::trunc);
	    TEST_EXIT(swapfile.is_open())
	      ("Cannot open file %s for writing!\n", name.c_str());
	    swapfile.close();
	  }
	  file.push(boost::iostreams::file_descriptor_sink(name, ios::trunc)); 

	  file << "<?xml version=\"1.0\"?>\n";
	  file << "<VTKFile type=\"PUnstructuredGrid\">\n";
	  file << "  <PUnstructuredGrid GhostLevel=\"0\">\n";
	  file << "    <PPoints>\n"
	      << "      <PDataArray type=\"Float32\" NumberOfComponents=\"3\" format=\"ascii\"/>\n"
	      << "    </PPoints>\n";
	  file << "    <PCells>\n"
	      << "      <PDataArray type=\"Int32\" Name=\"offsets\"/>\n"
	      << "      <PDataArray type=\"UInt8\" Name=\"types\"/>\n"
	      << "      <PDataArray type=\"Int32\" Name=\"connectivity\"/>\n"
	      << "    </PCells>\n";
	  file << "    <PPointData>\n";

	  for (size_t i = 0; i < componentNames.size(); i++)
	    file << "      <PDataArray type=\"Float32\" Name=\"" 
		<< componentNames[i] 
		<< "\" format=\"ascii\"/>\n";

	  file << "    </PPointData>\n";

	  for (int i = 0; i < nRanks; i++) {
	    string pname(fnPrefix + "-p" + lexical_cast<string>(i) + "-" + fnPostfix);
	    boost::filesystem::path filepath(pname);
	    file << "    <Piece Source=\"" 
		<< boost::filesystem::basename(filepath)
		<< boost::filesystem::extension(filepath) << "\"/>\n";      
	  }

	  file << "  </PUnstructuredGrid>\n";
	  file << "</VTKFile>\n";
	}
	
      } // end namespace detail
    } // end namespace VtkWriter
} } // end namespace io, AMDiS
