/******************************************************************************
 *
 * AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: 
 * Simon Vey, Thomas Witkowski, Andreas Naumann, Simon Praetorius, et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * This file is part of AMDiS
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/



/** \file VtkVector.h */

#ifndef AMDIS_VTKWRITER_DETAIL_H
#define AMDIS_VTKWRITER_DETAIL_H

#include <string>
#include <vector>

#include "AdaptInfo.h"
#include "io/DataCollector.h"

#ifdef HAVE_COMPRESSION
#include "io/FileCompression.h"
#endif

namespace AMDiS  { namespace io {
  
    namespace VtkWriter
    {
      // A class definition of a VtkWriter, to accept Datacollectors in 
      // the constructor.
      class Aux
      {
      public:
	Aux(std::vector<DataCollector<>*> *dc)
	  : dataCollector(dc)
#ifdef HAVE_COMPRESSION
	    , compress(NONE)
#endif
	{
	  degree = (*dataCollector)[0]->getFeSpace()->getBasisFcts()->getDegree();    
	  dim = (*dataCollector)[0]->getMesh()->getDim();
	}  

	/// Writes a ParaView-VTK file.
	int writeFile(std::string name);

#ifdef HAVE_COMPRESSION
	/// Set a compressing method for file output.
	void setCompression(FileCompression c) 
	{
	  compress = c;
	}
#endif

      protected:
	/// Writes the VTK file to an arbitrary stream.
	template<typename T>
	void writeFileToStream(T &file);

	/// Writes all coordinates of vertices and interpolation points to an 
	/// output file.
	template<typename T>
	void writeVertexCoords(T &file);

	/// Writes all values of vertices and interpolation point to an output file.
	template<typename T>
	void writeVertexValues(T &file, int componentNo);

	/// Writes the connectivity of all simplices to an output file.
	template<typename OutputStream>
	void writeConnectivity(OutputStream &file);

      private:
	/// List of DataCollectors, for each component of the problem one.
	std::vector<DataCollector<>*> *dataCollector;

#ifdef HAVE_COMPRESSION
	/// Defines if the file has to be compressed for ouput, and with which
	/// kind of compress method.
	FileCompression compress;
#endif

	/// Degree of the basis function of the problem.
	int degree;
	
	/// Dimension of the geometry.
	int dim;
      };
      
      
      namespace detail
      {

	/// Adds a new entry to a ParaView animation file.
	int updateAnimationFile(AdaptInfo *adaptInfo,
				std::string valueFilename,
				std::vector< std::pair<double, std::string> > *paraViewAnimationFrames,
				std::string animationFilename);

	/// Writes a pvtu file, which contains the links to all the rank files.
	void writeParallelFile(std::string name, int nRanks,
			      std::string fnPrefix, std::string fnPostfix,
			      std::vector<std::string> &componentNames);

	/// Writes the connectivity for the case dim = 2 and degree = 2 to an output file.
	template<typename S, typename OutputStream>
	void writeConnectivity_dim2_degree2(DataCollector<S>* dataCollector, OutputStream &file);

	/// Writes the connectivity for the case dim = 2 and degree = 3 to an output file.
	template<typename S, typename OutputStream>
	void writeConnectivity_dim2_degree3(DataCollector<S>* dataCollector, OutputStream &file);

	/// Writes the connectivity for the case dim = 2 and degree = 4 to an output file.
	template<typename S, typename OutputStream>
	void writeConnectivity_dim2_degree4(DataCollector<S>* dataCollector, OutputStream &file);
	
	/// Writes a world coordinate to a given file.
	template<typename OutputStream>
	void writeCoord(OutputStream &file, WorldVector<double> coord)
	{
	  file << std::scientific;
	  file.precision(15);
	  
	  for (int i = 0; i < Global::getGeo(WORLD); i++)
	    file << " " << coord[i];
	  for (int i = Global::getGeo(WORLD); i < 3; i++)
	    file << " "<<0.0;

	  file << "\n";
	}

      } // end namespace detail
    } // end namespace VtkWriter
} } // end namespace io, AMDiS

#include "VtkWriter.hh"

#endif // AMDIS_VTKWRITER_DETAIL_H
