/******************************************************************************
 *
 * AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: 
 * Simon Vey, Thomas Witkowski, Andreas Naumann, Simon Praetorius, et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * This file is part of AMDiS
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/


/** \file PMtl4Solver.h */

#ifndef AMDIS_PMTL4_SOLVER_H
#define AMDIS_PMTL4_SOLVER_H

#ifdef HAVE_PARALLEL_MTL4

#include "solver/MTL4SolverBase.h"
#include "parallel/ParallelSolver.h"

namespace AMDiS {

  namespace Parallel {
    

    /** \ingroup Solver
      * 
      * \brief
      * Wrapper class for different PMTL4 solvers. These solvers
      * are parametrized by Matrix- and VectorType. The different
      * algorithms, like krylov subspace methods or other external
      * solvers where PMTL4 provides an interface, can be assigned
      * by different Runner objects.
      **/
    template< typename MatrixType, typename VectorType, typename Runner >
    class PMTL4Solver : public ParallelSolver, protected MTL4SolverBase<MatrixType, VectorType, Runner>
    {
    public:
      typedef PMTL4Solver<MatrixType, VectorType, Runner> self;
      typedef MTL4SolverBase<MatrixType, VectorType, Runner> details;
      
      using details::solve;
    
      /// Creator class used in the LinearSolverMap.
      class Creator : public LinearSolverCreator
      {
      public:
	virtual ~Creator() {}
	
	/// Returns a new MTL4Solver object.
	LinearSolver* create() 
	{ 
	  return new self(this->name); 
	}
      };
    
      /// Constructor
      PMTL4Solver(std::string n) 
      : ParallelSolver(n, false),
	details(this)
      {}
      
      
      /// Implementation of \ref LinearSolver::getRunner()
      virtual OEMRunner* getRunner()
      {
	return details::getRunner();
      }
      
      
      /// Implementation of \ref LinearSolver::getLeftPrecon()
      virtual OEMPreconditioner* getLeftPrecon()
      {
	return details::getLeftPrecon();
      }
      
      
      /// Implementation of \ref LinearSolver::getRightPrecon()
      virtual OEMPreconditioner* getRightPrecon()
      {
	return details::getRightPrecon();
      }
      
    protected:
      /// Implementation of \ref LinearSolver::solveLinearSystem()
      virtual int solveLinearSystem(const SolverMatrix<Matrix<DOFMatrix*> >& A,
				    SystemVector& x,
				    SystemVector& b,
				    bool createMatrixData,
				    bool storeMatrixData)
      {
	ParallelMapper mapper(*ParallelSolver::getDofMapping());
	return details::solve(A, x, b, mapper, createMatrixData, storeMatrixData);
      }
    };
    
  } // end namespace Parallel
} // end namespace AMDiS

#endif // HAVE_PARALLEL_MTL4

#endif // AMDIS_PMTL4_SOLVER_H

