/******************************************************************************
 *
 * AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: 
 * Simon Vey, Thomas Witkowski, Andreas Naumann, Simon Praetorius, et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * This file is part of AMDiS
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/


/** \file ITL_Runner.h */


#ifndef AMDIS_ITL_RUNNER_H
#define AMDIS_ITL_RUNNER_H

#include "solver/MTL4Solver.h"
#include "solver/ITL_Preconditioner.h"

#include <boost/numeric/itl/itl.hpp>
#include <boost/numeric/mtl/mtl.hpp>

namespace AMDiS {

  template< typename MatrixType, typename VectorType >
  struct PreconPair
  { 
    /// Pointer to the left preconditioner
    ITL_BasePreconditioner< MatrixType, VectorType >* l;

    /// Pointer to the right preconditioner
    ITL_BasePreconditioner< MatrixType, VectorType >* r;
    
    PreconPair() : l(nullptr), r(nullptr) {}
  };


  template< typename ITLSolver, typename MatrixType, typename VectorType >
  struct ITL_Runner : public MTL4Runner< MatrixType, VectorType >
  {       
    typedef MTL4Runner< MatrixType, VectorType > super;
    
    ITL_Runner(LinearSolver* oemPtr)
    : oem(*oemPtr),
      solver(oem.getName())
    {
      setPrecon(preconPair);
    }
    
    
    ~ITL_Runner()
    {
      if (preconPair.l != nullptr) {
	preconPair.l->exit();
	delete preconPair.l;
	preconPair.l = nullptr;
      }

      if (preconPair.r != nullptr) {
	preconPair.r->exit();
	delete preconPair.r;
	preconPair.r = nullptr;
      }
    }
    
    
    /// Implementation of \ref MTL4Runner::init()
    void init(const typename super::BlockMatrix& A, const MatrixType& fullMatrix) override
    {
      preconPair.l->init(A, fullMatrix);
      preconPair.r->init(A, fullMatrix);
    }
    
    
    /// Implementation of \ref MTL4Runner::solve()
    int solve(const MatrixType& A , VectorType& x, const VectorType& b) override
    { FUNCNAME("ITL_Runner::solve()");
    
      TEST_EXIT(preconPair.l != nullptr)("there is no left preconditioner\n");
      TEST_EXIT(preconPair.r != nullptr)("there is no right preconditioner\n");
            
      VectorType r(A * x - b); 
      int error = 0;
      if (oem.getInfo() == 0) {
	itl::basic_iteration<typename MatrixType::value_type> 
	  iter(r, oem.getMaxIterations(), oem.getRelative(), oem.getTolerance());
	
	error = solver(A, x, b, *(preconPair.l), *(preconPair.r), iter);
	oem.setErrorCode(error);
	oem.setIterations(iter.iterations());
	oem.setResidual(iter.resid());
      } else {
	itl::cyclic_iteration<typename MatrixType::value_type> 
	  iter(r, oem.getMaxIterations(), oem.getRelative(), oem.getTolerance(), 
	       oem.getPrint_cycle());
	
	error = solver(A, x, b, *(preconPair.l), *(preconPair.r), iter);
	oem.setErrorCode(error);
	oem.setIterations(iter.iterations());
	oem.setResidual(iter.resid());
      }
      
      return error;
    }

    
    /// Implementation of \ref MTL4Runner::adjoint_solve()
    int adjoint_solve(const MatrixType& A , VectorType& x, const VectorType& b) override
    { FUNCNAME("ITL_Runner::adjoint_solve()");
    
      TEST_EXIT(preconPair.l != nullptr)("there is no left preconditioner\n");
      TEST_EXIT(preconPair.r != nullptr)("there is no right preconditioner\n");
            
      mtl::matrix::transposed_view<const MatrixType> B(A);
      VectorType r(B * x - b); 
      int error = 0;
      if (oem.getInfo() == 0) {
	itl::basic_iteration<typename MatrixType::value_type> 
	  iter(r, oem.getMaxIterations(), oem.getRelative(), oem.getTolerance());
	
	error = solver(B, x, b, *(preconPair.l), *(preconPair.r), iter);
	oem.setErrorCode(error);
	oem.setIterations(iter.iterations());
	oem.setResidual(iter.resid());
      } else {
	itl::cyclic_iteration<typename MatrixType::value_type> 
	  iter(r, oem.getMaxIterations(), oem.getRelative(), oem.getTolerance(), 
	       oem.getPrint_cycle());
	
	error = solver(B, x, b, *(preconPair.l), *(preconPair.r), iter);
	oem.setErrorCode(error);
	oem.setIterations(iter.iterations());
	oem.setResidual(iter.resid());
      }
      
      return error;
    }
      
      
    /// Implementation of \ref OEMRunner::exit()
    void exit() override
    {
      preconPair.l->exit();
      preconPair.r->exit();
    }
    
    
    /// Implementation of \ref OEMRunner::getLeftPrecon()
    OEMPreconditioner* getLeftPrecon() override
    {
      return preconPair.l;
    }    
    
    
    /// Implementation of \ref OEMRunner::getRightPrecon()
    OEMPreconditioner* getRightPrecon() override
    {
      return preconPair.r;
    }
    
  protected:
    LinearSolver& oem;
    ITLSolver solver;
    
    void setPrecon(PreconPair<MatrixType, VectorType>& pair)
    { FUNCNAME("ITL_Runner::setPrecon()");
    
      /// Creator for the left preconditioner
      CreatorInterfaceName< ITL_BasePreconditioner<MatrixType, VectorType> >* leftCreator(nullptr);

      /// Creator for the right preconditioner
      CreatorInterfaceName< ITL_BasePreconditioner<MatrixType, VectorType> >* rightCreator(nullptr);

      std::string preconType("no");
      std::string initFileStr = oem.getName() + "->left precon";
      Parameters::get(initFileStr, preconType);
      leftCreator = dynamic_cast<CreatorInterfaceName< ITL_BasePreconditioner<MatrixType, VectorType> >*>(
	CreatorMap<ITL_BasePreconditioner<MatrixType, VectorType> >::getCreator(preconType, initFileStr) );
      TEST_EXIT(leftCreator != nullptr)
	("there is no creator for the given left preconditioner\n");
      leftCreator->setName(initFileStr);

      preconType = "no";
      initFileStr = oem.getName() + "->right precon";
      Parameters::get(initFileStr, preconType);
      rightCreator = dynamic_cast<CreatorInterfaceName< ITL_BasePreconditioner<MatrixType, VectorType> >*>(
	CreatorMap<ITL_BasePreconditioner<MatrixType, VectorType> >::getCreator(preconType, initFileStr) );
      TEST_EXIT(rightCreator != nullptr)
	("there is no creator for the given right preconditioner\n");
      rightCreator->setName(initFileStr);

      pair.l = leftCreator->create();
      pair.r = rightCreator->create();
    }
    
  private:
    PreconPair<MatrixType, VectorType> preconPair;
  };

} // namespace AMDiS

#endif // AMDIS_ITL_RUNNER_H
