/******************************************************************************
 *
 * AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: 
 * Simon Vey, Thomas Witkowski, Andreas Naumann, Simon Praetorius, et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * This file is part of AMDiS
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/


/** \file MTL4Solver.h */

#ifndef AMDIS_MTL4SOLVER_H
#define AMDIS_MTL4SOLVER_H

#include "solver/LinearSolver.h"
#include "solver/MTL4SolverBase.h"
#include "Timer.h"
#include <iostream>
#include <boost/mpl/bool.hpp>
#include <boost/numeric/mtl/utility/is_distributed.hpp>

#if defined(HAVE_PARALLEL_DOMAIN_AMDIS) && defined(HAVE_PARALLEL_MTL4)
#include <boost/numeric/mtl/par/distribution.hpp>
#endif

namespace AMDiS {
  
  template< typename MatrixType, typename VectorType >
  struct MTL4Runner : public OEMRunner
  {
    typedef SolverMatrix<Matrix<DOFMatrix*> > BlockMatrix;
    
    virtual void init(const BlockMatrix& A, const MatrixType& fullMatrix) = 0;      
    virtual int solve(const MatrixType& A, VectorType& x, const VectorType& b) = 0;  
    virtual int adjoint_solve(const MatrixType& A, VectorType& x, const VectorType& b) {
      FUNCNAME("MTL4Runner::adjoint_solve()");
      ERROR_EXIT("Must be implemented in derived class!\n");
      return 0;
    };
  };
  

  /** \ingroup Solver
   * 
   * \brief
   * Wrapper class for different MTL4 solvers. These solvers
   * are parametrized by Matrix- and VectorType. The different
   * algorithms, like krylov subspace methods or other external
   * solvers where MTL4 provides an interface, can be assigned
   * by different Runner objects.
   **/
  template< typename MatrixType, typename VectorType, typename Runner >
  class MTL4Solver : public LinearSolver, protected MTL4SolverBase<MatrixType, VectorType, Runner>
  {    
    
  public:
    typedef MTL4Solver<MatrixType, VectorType, Runner> 		self;
    typedef MTL4SolverBase<MatrixType, VectorType, Runner> 	details;
    
    /// Creator class used in the LinearSolverMap.
    class Creator : public LinearSolverCreator
    {
    public:
      virtual ~Creator() {}
      
      /// Returns a new MTL4Solver object.
      LinearSolver* create() 
      { 
	return new self(this->name); 
      }
    };
    
    /// Constructor
    MTL4Solver(std::string n)
    : LinearSolver(n),
      details(this)
    {}
    
    
    /// Implementation of \ref LinearSolver::getRunner()
    OEMRunner* getRunner() override
    {
      return details::getRunner();
    }
    
    
    /// Implementation of \ref LinearSolver::getLeftPrecon()
    OEMPreconditioner* getLeftPrecon() override
    {
      return details::getLeftPrecon();
    }
    
    
    /// Implementation of \ref LinearSolver::getRightPrecon()
    OEMPreconditioner* getRightPrecon() override
    {
      return details::getRightPrecon();
    }
    
  protected:
    /// Implementation of \ref LinearSolver::solveLinearSystem()
    int solveLinearSystem(const SolverMatrix<Matrix<DOFMatrix*> >& A,
				  SystemVector& x,
				  SystemVector& b,
				  bool createMatrixData,
				  bool storeMatrixData) override
    {    
      VectorialMapper mapper(A);
      return details::solve(A, x, b, mapper, createMatrixData, storeMatrixData);
    }
  };
}

#endif // AMDIS_MTL4SOLVER_H
