/******************************************************************************
 *
 * AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: 
 * Simon Vey, Thomas Witkowski, Andreas Naumann, Simon Praetorius, et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * This file is part of AMDiS
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/


/** \file Mapper.h */

#ifndef AMDIS_MAPPER_H
#define AMDIS_MAPPER_H

#include "solver/SolverMatrix.h"
#include "MTL4Types.h"
#include <vector>

namespace AMDiS {
  
  /**
   * \brief BaseClass for all mapper types
   * 
   * A mapper assigned a local row/column index of a cell of
   * a block matrix to a global matrix index.
   * Call \ref setRow and \ref setCol first, to select a block
   * in the block matrix. Then get with \ref row, respective \ref col,
   * the global matrix index to the assigned local index.
   **/  
  template< typename size_type >
  struct MapperBase
  {
    /// set the current block row
    inline virtual void setRow( unsigned int r) = 0;
    
    /// set the current block columns
    inline virtual void setCol( unsigned int c) = 0;
    
    /// return global matrix row, for local row in the current matrix block
    inline virtual size_type row(size_type r) const = 0;
    
    /// return global matrix column, for local column in the current matrix block
    inline virtual size_type col(size_type c) const = 0;
    
    /// return overall number of rows
    inline virtual size_type getNumRows() const = 0;
    
    /// return overall number of columns
    inline virtual size_type getNumCols() const = 0;
    
    /// return number of components/blocks
    inline virtual unsigned int getNumComponents() const = 0;
    
    inline size_type row(unsigned int block_r, unsigned int block_c, size_type r)
    {
      setRow(block_r);
      setCol(block_c);
      return row(r);
    }
        
    inline size_type col(unsigned int block_r, unsigned int block_c, size_type c)
    {
      setRow(block_r);
      setCol(block_c);
      return col(c);
    }
  };

  
  /// Mapper implementation for non-parallel block matrices
  struct VectorialMapper : public MapperBase< unsigned int >
  {
    /// Constructor for block-matrices
    VectorialMapper(const SolverMatrix<Matrix<DOFMatrix* > >& sm )
    : nComp(sm.getOriginalMat()->getSize()), 
      rowOffset(0), colOffset(0), nrow(0), ncol(0), sizes(nComp)
    {
      const Matrix<DOFMatrix* >& orMat(*sm.getOriginalMat());
      const int ns = orMat.getSize();
      for (int i= 0; i < ns; i++) {
	sizes[i] = orMat[i][i]->getFeSpace()->getAdmin()->getUsedSize();
	nrow += sizes[i];
      }
      ncol = nrow;
    }
    
    /// Constructor for single matrix
    VectorialMapper(const DOFMatrix* sm )
    : nComp(1), 
      rowOffset(0), colOffset(0), 
      nrow(0), ncol(0), 
      sizes(nComp)
    {
      sizes[0] = sm->getFeSpace()->getAdmin()->getUsedSize();
      nrow += sizes[0];
      ncol = nrow;
    }

    /// Constructor for system with equal components
    VectorialMapper(unsigned int nComp, int nDOFperComp)
    : nComp(nComp), 
      rowOffset(0), colOffset(0), 
      nrow(nComp*nDOFperComp), ncol(nrow),
      sizes(nComp)
    {
       for (unsigned int i = 0; i < nComp; ++i)
         sizes[i] = nDOFperComp;
    }
      
    /// calculate row offset for row component \param r
    inline void setRow( unsigned int r)
    { 
      assert( r <= sizes.size() );
      rowOffset = sum(r); 
    }

    /// calculate column offset for col component \param c
    inline void setCol( unsigned int c)
    {
      assert( c <= sizes.size() ); 
      colOffset = sum(c);
    }

    inline unsigned int row(unsigned int r) const { return r + rowOffset; }
    inline unsigned int col(unsigned int c) const { return c + colOffset; }

    inline unsigned int getNumRows() const { return nrow; }
    inline unsigned int getNumCols() const { return ncol; }

    inline unsigned int getNumComponents() const { return nComp; }

  private: // methods

    ///compute the sum of sizes from [0, end)
    inline unsigned int sum(unsigned int end) const 
    {
      unsigned int ret(0);
      for (unsigned int i(0); i < end; ++i)
	ret += sizes[i];
      return ret;
    }
    
  private: // variables

    unsigned int nComp;
    unsigned int rowOffset;
    unsigned int colOffset;
    unsigned int nrow;
    unsigned int ncol;
    
    std::vector< unsigned int > sizes;
  };
}

#endif // AMDIS_MAPPER_H
