/******************************************************************************
 *
 * AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: 
 * Simon Vey, Thomas Witkowski, Andreas Naumann, Simon Praetorius, et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * This file is part of AMDiS
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/


/** \file MatrixStreams.h */

#ifndef AMDIS_MATRIXSTREAMS_H
#define AMDIS_MATRIXSTREAMS_H

#include "DOFMatrix.h"
#include "solver/SolverMatrix.h"
#include "SystemVector.h"
#include "Collection.h"

#ifdef HAVE_SEQ_PETSC
#include "solver/PetscTypes.h"
#include <petsc.h>
#include <petscmat.h> 
#include <petscvec.h>
#endif

namespace AMDiS {
  
  using namespace mtl::operations;
  
  template< typename VecT, typename CurMap>
  struct VecMap {
    VecT& vec;
    CurMap& mapper;
    VecMap(VecT& vec, CurMap& mapper):
      vec(vec),mapper(mapper) {}
  };

  template< typename MatT, typename CurMap>
  struct MatMap {
    MatT& mat;
    CurMap& mapper;
    MatMap(MatT& mat, CurMap& m):
      mat(mat), mapper(m) {}
  };

  template< typename MTLMatrix, typename Mapper >
  void operator<<(MTLMatrix& matrix, MatMap<const SolverMatrix<Matrix<DOFMatrix* > >, Mapper >& Asolver)
  {
    const Matrix< DOFMatrix* >& A = *(Asolver.mat.getOriginalMat());
    int ns = A.getSize();

    Mapper& mapper(Asolver.mapper);
    set_to_zero(matrix);

    int nnz = 0;
    for (int rb = 0; rb < ns; ++rb)
      for (int cb = 0; cb < ns; ++cb)
        if (A[rb][cb])
	  nnz += A[rb][cb]->getBaseMatrix().nnz();	  
      
    {
      typedef mtl::matrix::mapped_inserter< typename Collection< MTLMatrix >::Inserter, Mapper > Inserter;
      Inserter ins(matrix, mapper, int(1.2 * nnz / matrix.dim1()));
      for (int rb = 0; rb < ns; ++rb) {
        mapper.setRow(rb);
        for (int cb = 0; cb < ns; ++cb) {
	  mapper.setCol(cb);
          if (A[rb][cb])
	    ins << A[rb][cb]->getBaseMatrix();
        }
      }
    }
  }

  template< typename Vector, typename CurMap >
  void operator<<(Vector& dest, VecMap<DOFVector<double >, CurMap >& rhs)
  {
    DOFVector<double>::Iterator it_x(&rhs.vec, USED_DOFS);
    int counter(0);
    typedef typename Collection< Vector >::Inserter Inserter ;

    Inserter swapIns(dest);
    mtl::vector::mapped_inserter< Inserter, CurMap > ins(swapIns, rhs.mapper);

    for (it_x.reset(); !it_x.end(); ++it_x) {
      ins[counter] << *it_x;
      counter++;
    }
  }

  template< typename Vector, typename CurMap >
  inline void operator>>(const Vector& dest, VecMap<DOFVector< double >, CurMap >& rhs)
  {
    DOFVector<double>::Iterator it_x(&rhs.vec, USED_DOFS);
    int counter(0);
    {
      mtl::vector::extracter< Vector > extracter(dest);
      for (it_x.reset(); !it_x.end(); ++it_x) {
	extracter[rhs.mapper.row(counter)] >> *it_x ;
	counter++;
      }
    }
  }

  template< typename Vector, typename CurMap >
  void operator<<(Vector& dest, VecMap<SystemVector, CurMap>& rhs)
  {
    int ns = rhs.vec.getSize();  // Number of systems.

    // Copy vectors
    typedef typename Collection< Vector >::Inserter Inserter;
    Inserter swapInserter(dest);
    mtl::vector::mapped_inserter< Inserter, CurMap > ins(swapInserter, rhs.mapper);
    for (int i = 0; i < ns; i++) {
      DOFVector<double>::Iterator it_source(rhs.vec.getDOFVector(i), USED_DOFS);
      int counter(0);
      rhs.mapper.setRow(i);
      for (it_source.reset(); !it_source.end(); ++it_source) {
        ins[counter] << *it_source;
        counter++;
      }
    }
  }

  template< typename Vector, typename CurMap >
  void operator>>(const Vector& dest, VecMap<SystemVector, CurMap>& rhs) 
  {
    int ns = rhs.vec.getSize();  // Number of systems.

    // Copy vectors
    for (int i = 0; i < ns; i++) {
      DOFVector< double >& dofvec(*(rhs.vec.getDOFVector(i)));
      rhs.mapper.setRow(i);
      VecMap< DOFVector< double >, CurMap > swap(dofvec, rhs.mapper);
      dest >> swap;
    }
  }
  
#ifdef HAVE_SEQ_PETSC
  
  /// fill PETSc matrix from DOFMatrix
  inline void operator<<(Mat& mat, const DOFMatrix& rhs)
  {
    bool initMatrix = false;
    if (mat == PETSC_NULL) {
      std::vector<PetscInt> nnz(rhs.getSize());
      for (size_t k = 0; k < static_cast<size_t>(rhs.getSize()); k++)
	nnz[k] = rhs.getBaseMatrix().nnz_local(k);	  
      
      MatCreateSeqAIJ(PETSC_COMM_SELF, num_rows(rhs.getBaseMatrix()), num_cols(rhs.getBaseMatrix()), 0, &(nnz[0]), &mat);
      initMatrix = true;
    }
    
    std::vector<PetscInt> indices;
    for (size_t i = 0; i < rhs.getBaseMatrix().ref_minor().size(); i++)
      indices.push_back(rhs.getBaseMatrix().ref_minor()[i]);

    for (PetscInt i = 0; i < static_cast<PetscInt>(num_rows(rhs.getBaseMatrix())); i++) {
      if  (rhs.getBaseMatrix().nnz_local(i) > 0)
        MatSetValues(mat, 1, &i, rhs.getBaseMatrix().nnz_local(i), 
		   &(indices[rhs.getBaseMatrix().ref_major()[i]]), 
		   &(rhs.getBaseMatrix().value_from_offset(rhs.getBaseMatrix().ref_major()[i])), ADD_VALUES);	
    }
    
    if (initMatrix) {      
	MatAssemblyBegin(mat, MAT_FINAL_ASSEMBLY);
	MatAssemblyEnd(mat, MAT_FINAL_ASSEMBLY);
    }
  }
    

  /// create and fill \ref PetscMatrix
  inline void operator<<(PetscMatrix& mat, const SolverMatrix<Matrix<DOFMatrix*> >& Asolver)
  {
	  const Matrix< DOFMatrix* >& A = *(Asolver.getOriginalMat());
	  if(mat.isNested) {

		  std::vector<PetscInt> nnz;      
		  mat.nestMat.resize(A.getNumRows() * A.getNumCols());

		  for (size_t i = 0; i < static_cast<size_t>(A.getNumRows()); i++) {
			  for (size_t j = 0; j < static_cast<size_t>(A.getNumCols()); j++) {
				  size_t idx = i * A.getNumCols() + j;

				  if (A[i][j] == nullptr 
						  || num_rows(A[i][j]->getBaseMatrix()) == 0 
						  || num_cols(A[i][j]->getBaseMatrix()) == 0
						  || A[i][j]->getBaseMatrix().nnz() == 0) {
					  mat.nestMat[idx] = PETSC_NULL;
					  continue;
				  }

				  mat.nestMat[idx] << *(A[i][j]);
			  }
		  }

		  // create nested matrix from a vector of block matrices
		  MatCreateNest(PETSC_COMM_SELF, A.getNumRows(), PETSC_NULL, A.getNumCols(), PETSC_NULL, &(mat.nestMat[0]), &mat.matrix);
	  }else {
		  bool initMatrix = false;
		  unsigned nRows(0);
		  unsigned nEntries(0);
		  std::vector< int > nRowsBlock(A.getNumRows());
		  for(int i(0); i < A.getNumRows(); ++i) {
			  int j(0);
			  for( ; j<A.getNumCols() && A[i][j] == NULL; ++j) ;
			  if( j == A.getNumCols() ) {
				  std::stringstream ss; ss << "ERROR: solver matrix has empty row " << i << "\n";
				  throw std::runtime_error(ss.str());
			  }
			  nRowsBlock[i]=num_rows(A[i][j]->getBaseMatrix());
			  nRows+=nRowsBlock[i];
		  }
		  std::vector<PetscInt> nnz(nRows);
		  //initialize the nnz (could be avoided, but gets complicated
		  for (int i(0); i < nRows; ++i) nnz[i] = 0;

		  //build a list of pairs (col, value) for each row
		  std::vector< std::list< std::pair<int, double > > > rowData(nRows);
		  //use the standard mapper to build global structure
		  VectorialMapper mapper(Asolver);
		  PetscInt maxNnz(0);
		  for (int i(0); i < A.getNumRows(); ++i) {
			  mapper.setRow(i);
			  //compute the number of nonzeros per global row
			  for (int j(0); j < A.getNumCols(); ++j) {
				  if (A[i][j] != NULL) {
					  mapper.setCol(j);
					  const DOFMatrix::base_matrix_type& mtlMat(A[i][j]->getBaseMatrix());
					  for (unsigned k(0);  k < nRowsBlock[i]; ++k) {
						  unsigned curRow(mapper.row(k));
						  nnz[curRow] += mtlMat.nnz_local(k);
						  maxNnz = std::max(maxNnz, nnz[curRow]);
						  unsigned curPos(mtlMat.ref_major()[k]);
						  for(unsigned l(0); l < mtlMat.nnz_local(k); ++l, ++curPos) {
							  rowData[curRow].push_back(std::make_pair(mapper.col(mtlMat.ref_minor()[curPos]), mtlMat.data[curPos]));
						  }
					  }
				  }
			  }

		  }
		  if (mat.matrix == PETSC_NULL) {
			  MatCreateSeqAIJ(PETSC_COMM_SELF, nRows, nRows, 0, &(nnz[0]), &mat.matrix);
			  initMatrix = true;
		  }
		  //reduce mallocs..
		  std::vector< PetscInt > colIndices(maxNnz); 
		  std::vector< PetscScalar > values(maxNnz);
		  std::list< std::pair< int, double > >::iterator it;
		  unsigned j;
		  for (PetscInt i(0); i < nRows; ++i) {
			  j=0; it = rowData[i].begin();
			  for (; it != rowData[i].end(); ++it, ++j) {
				  colIndices[j] = it->first;
				  values[j] = it->second;
			  }
			  MatSetValues(mat.matrix, 1, &i, nnz[i], &colIndices[0], &values[0], INSERT_VALUES);
		  }
		  if (initMatrix) {      
			  MatAssemblyBegin(mat.matrix, MAT_FINAL_ASSEMBLY);
			  MatAssemblyEnd(mat.matrix, MAT_FINAL_ASSEMBLY);
		  }
	  }
  }
  
  
  /// fill PETSc vector from DOFVector
  inline void operator<<(Vec& petscVec, /* const */DOFVector<double>& vec)
  {
    // Traverse all used DOFs in the dof vector.
    DOFVector<double>::Iterator dofIt(&vec, USED_DOFS);
    PetscInt index = 0;
    for (dofIt.reset(); !dofIt.end(); ++dofIt, ++index) {
      double value = *dofIt;
      VecSetValue(petscVec, index, value, ADD_VALUES);      
    }
  }
    
    
  /// DOFVector from PETSc vector
  inline void operator>>(const Vec& petscVec, DOFVector<double>& vec)
  {
    // Traverse all used DOFs in the dof vector.
    DOFVector<double>::Iterator dofIt(&vec, USED_DOFS);
    PetscInt index = 0;
    for (dofIt.reset(); !dofIt.end(); ++dofIt, ++index) {
      double value = 0.0;
      VecGetValues(petscVec, 1, &index, &value);
      *dofIt = value;
    }
  }
  
  
  /// fill PETSc vector from DOFVector
  inline void operator<<(Vec& petscVec, /* const */SystemVector& vec)
  {
#ifdef VecType // PETSc uses MACROS instead of typedefs in Versions 3.3x
    const VecType vecType;
#else
    VecType vecType;
#endif
    VecGetType(petscVec, &vecType);
    
    if (strcmp(vecType, VECNEST) == 0) {
      for (size_t i = 0; i < static_cast<size_t>(vec.getSize()); i++) {  
	Vec v;
	VecNestGetSubVec(petscVec, i, &v);
	v << *(vec.getDOFVector(i));
      }
    } else {
      PetscInt index = 0;
      for (size_t i = 0; i < static_cast<size_t>(vec.getSize()); i++) {
	DOFVector<double>::Iterator dofIt(vec.getDOFVector(i), USED_DOFS);
	for (dofIt.reset(); !dofIt.end(); ++dofIt, ++index) {
	  double value = *dofIt;
	  VecSetValue(petscVec, index, value, ADD_VALUES);      
	}
      }
    }
  }
    
    
  /// fill SystemVector from PETSc vector
  inline void operator>>(const Vec& petscVec, SystemVector& vec)
  {
#ifdef VecType // PETSc uses MACROS instead of typedefs in Versions 3.3x
    const VecType vecType;
#else
    VecType vecType;
#endif
    VecGetType(petscVec, &vecType);
    
    if (strcmp(vecType, VECNEST) == 0) {
      for (size_t i = 0; i < static_cast<size_t>(vec.getSize()); i++) {  
	Vec v;
	VecNestGetSubVec(petscVec, i, &v);
	v >> *(vec.getDOFVector(i));
      }
    } else {
      PetscInt n = 0;
      for (size_t i = 0; i < static_cast<size_t>(vec.getSize()); i++)
	n += vec.getDOFVector(i)->getUsedSize();
      
      PetscInt N = 0;
      VecGetSize(petscVec, &N);      
      assert(n == N);
      
      PetscInt index = 0;
      for (size_t i = 0; i < static_cast<size_t>(vec.getSize()); i++) {
	DOFVector<double>::Iterator dofIt(vec.getDOFVector(i), USED_DOFS);
	for (dofIt.reset(); !dofIt.end(); ++dofIt, ++index) {
	  double value = 0.0;
	  VecGetValues(petscVec, 1, &index, &value);
	  *dofIt = value;
	}
      }
    }
  }
  

  /// create and fill \ref PetscVector
  inline void operator<<(PetscVector& petscVec, /* const */SystemVector& rhs)
  {
    size_t nComponents = rhs.getSize();
    
    // only create nested vector if requested, i.e. flag createNestedVector is set to true
    if (petscVec.isNested) {
      petscVec.nestVec.resize(nComponents);

      for (size_t i = 0; i < nComponents; i++) {
	VecCreateSeq(PETSC_COMM_SELF,rhs.getDOFVector(i)->getUsedSize(), &(petscVec.nestVec[i]));

	petscVec.nestVec[i] << *(rhs.getDOFVector(i));
	
	VecAssemblyBegin(petscVec.nestVec[i]);
	VecAssemblyEnd(petscVec.nestVec[i]);
      }

      // create nested vector from vector of block vectors
      VecCreateNest(PETSC_COMM_SELF, nComponents, PETSC_NULL, &(petscVec.nestVec[0]), &(petscVec.vector));
    } else {
      PetscInt n = 0;
      for (size_t i = 0; i < nComponents; i++)
	n += rhs.getDOFVector(i)->getUsedSize();
      
      VecCreateSeq(PETSC_COMM_SELF, n, &(petscVec.vector));
      petscVec.vector << rhs;
    }
  }
  
#endif
}
#endif // AMDIS_MATRIXSTREAMS_H
