/******************************************************************************
 *
 * AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: 
 * Simon Vey, Thomas Witkowski, Andreas Naumann, Simon Praetorius, et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * This file is part of AMDiS
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/
#ifndef MTL4SOLVER_BASE_H
#define MTL4SOLVER_BASE_H

#include "MatrixStreams.h"
#include "Timer.h"
#include <iostream>
#include <boost/mpl/bool.hpp>
#include <boost/numeric/mtl/utility/is_distributed.hpp>

#if defined(HAVE_PARALLEL_DOMAIN_AMDIS) && defined(HAVE_PARALLEL_MTL4)
#include <boost/numeric/mtl/par/distribution.hpp>
#endif

namespace AMDiS {
  namespace details {

  template< typename MatrixType, typename VectorType, typename Runner >
  class MTL4Solver
  {
  public:
    /// Constructor
    MTL4Solver(LinearSolver* oem)
    : runner(oem),
      matrix(0,0)
    {}
    
  protected:    
    Runner runner;

    /// init systemmatrix depending on Mapper parameters.
    template< typename Mapper, bool IsDist>
    void init(MatrixType& m, Mapper& mapper, boost::mpl::bool_<IsDist>)
    {
      m.change_dim(mapper.getNumRows(), mapper.getNumCols());
    }

    
#if defined(HAVE_PARALLEL_DOMAIN_AMDIS) && defined(HAVE_PARALLEL_MTL4)
    /// init systemmatrix depending on Mapper parameters,
    /// specialized for distributed matrices
    void init(MatrixType& m, ParallelMapper& mapper, boost::mpl::true_)
    {
      mtl::par::block_distribution dist(mapper.getNumRows());
      dist.setup_from_local_size(mapper.getMap().getLocalDofs());
      m.change_dim(0, 0);
      m.init_distribution(dist, dist, mapper.getNumRows(), mapper.getNumRows());
    }
#endif


    /// init MTL-vector depending on Mapper parameters.
    template< bool IsDist>
    void init(VectorType& v, boost::mpl::bool_<IsDist>)
    {
      v.change_dim(num_rows(matrix));
    }
    

    /// init MTL-vector depending on Mapper parameters,
    /// specialized for distributed matrices
    void init(VectorType& v, boost::mpl::true_)
    {
      v.init_distribution(row_distribution(matrix), num_rows(matrix));
    }
    

    /// Realization of solve(A, x, b) to solve the linear system \f$ Ax=b \f$.
    /// 1. transfer matrix and vectors to MTL datatypes
    /// 2. call Runner::solve(A, x, b)
    /// 3. transfer solution back to AMDiS \ref SystemVector
    template< typename MatrixT, typename VectorT, typename Mapper >
    int solve(const MatrixT& A, VectorT& x, VectorT& b, Mapper& mapper,
	      bool createMatrixData,
	      bool storeMatrixData) 
    { FUNCNAME("MTL4Solver::solve()");

      #ifdef HAVE_PARALLEL_DOMAIN_AMDIS
      MPI::COMM_WORLD.Barrier();
      #endif
      Timer t;
      if (createMatrixData) {
	init(matrix, mapper, mtl::traits::is_distributed<MatrixType>());

        set_to_zero(matrix);
        MatMap< const MatrixT, Mapper > matMap(A, mapper);
        matrix << matMap;
	mtl::io::matrix_market_ostream out("matrix.mtx");
	out << matrix;
	#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
	  mtl::io::matrix_market_ostream out2("matrix-p"+boost::lexical_cast<std::string>(MPI::COMM_WORLD.Get_rank())+".mtx");
	#else
	  mtl::io::matrix_market_ostream out2("matrix-full.mtx");
	#endif
	out2 << A.getSubMatrix(0,0);
	MSG("matrix.nnz=%d\n", matrix.nnz());
  
	runner.init(A, matrix);
      }

      VectorType mtl_x;
      init(mtl_x, mtl::traits::is_distributed<VectorType>());
      set_to_zero(mtl_x);
      VecMap< VectorT, Mapper > xVecMap(x, mapper);
      mtl_x << xVecMap;

      VectorType mtl_b;
      init(mtl_b, mtl::traits::is_distributed<VectorType>());
      set_to_zero(mtl_b);
      VecMap< VectorT, Mapper> bVecMap(b, mapper);
      mtl_b << bVecMap;
   
      MSG("fill MTL4 matrix needed %.5f seconds\n", t.elapsed());

      t.reset();
      int error = runner.solve(matrix ,mtl_x, mtl_b);
      MSG("solve MTL4 matrix needed %.5f seconds\n", t.elapsed());

      VectorType r(mtl_b); 
      r -= matrix * mtl_x; 
      double residual = two_norm(r);
      MSG("MTL4Solver: ||b-Ax||= %e\n", residual);

      mtl_x >> xVecMap;

      runner.exit();

      return error;
    }
    
    OEMRunner* getRunner()
    {
      return &runner;
    }
    
    OEMPreconditioner* getLeftPrecon()
    {
      return runner.getLeftPrecon();
    }
    
    OEMPreconditioner* getRightPrecon()
    {
      return runner.getRightPrecon();
    }
    
  private:
    MatrixType matrix;
  };
}
#endif
