/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/

#ifndef EXTENSIONS_META_TOOLS_H
#define EXTENSIONS_META_TOOLS_H

// #include "AMDiS.h"
#include <boost/mpl/if.hpp>
#include <boost/mpl/arithmetic.hpp>
#include <boost/mpl/comparison.hpp>
#include <boost/mpl/logical.hpp>
#include <boost/utility/enable_if.hpp>

#define GET_LOOP_INDEX(VAR, SEQ, IDX) static const long VAR = boost::mpl::at_c<SEQ,IDX>::type::value;

// namespace boost {
//   namespace fusion {
//     template<int I> inline double at_c(WorldVector<double> &x) { return x[I]; }
//   }
// }

namespace tools {
  
  /**
    * \defgroup binomial_coefficients binomial coefficients
    * @{
    */
  namespace details {

    template <int N, int K, bool N_pos = true, bool K_pos = true>
    struct Binomial
    {
      BOOST_STATIC_CONSTANT(int, value = (details::Binomial<N-1, K-1, (N-1>0), (K-1>0) >::value
					+ details::Binomial<N-1, K,   (N-1>0), (K>0)   >::value) );
    };
    
    template <int N, bool N_pos>
    struct Binomial<N, N, N_pos, true> // K > 0, N > 0, K==N
    {
      BOOST_STATIC_CONSTANT(int, value = 1);
    };
    
    template <int N, int K, bool N_pos>
    struct Binomial<N, K, N_pos, false> // K <= 0
    {
      BOOST_STATIC_CONSTANT(int, value = 1);
    };
    
    template <int N, int K>
    struct Binomial<N, K, false, true> // K > 0, N <= 0
    {
      BOOST_STATIC_CONSTANT(int, value = 0);
    };
  }
  
  /// \brief
  /// interface to calc binomial coefficients (N, K) at compile-time for N,K >= 0
  ///
  template <int N, int K>
  struct Binomial
  {
    BOOST_STATIC_CONSTANT(int, value = (details::Binomial<N, K, (N>0), (K>0)>::value) );
  };
  /**@}*/

// ______________________________________________________________________________________________________
  
  /**
    * \defgroup math_basics some mathematical basics
    * @{
    */
  
  /// \brief
  /// check divisibility
  ///
  template <int u, int v>
  struct Divisible
  {
    BOOST_STATIC_CONSTANT(bool, value = u % v == 0 ? true : false );
  };
    
  /// \brief
  /// check divide by zero condition
  ///
  template <int u>
  struct Divisible<u, 0>
  {
    BOOST_STATIC_CONSTANT(bool, value = false ); // better: static assert
  };
  
  /// \brief
  /// check number is even or not
  ///
  template <int u>
  struct IsEven
  {
    BOOST_STATIC_CONSTANT(bool, value = (Divisible<u, 2>::value) );
  };
    
  /// \brief
  /// check number is odd or not
  ///
  template <int u>
  struct IsOdd
  {
    BOOST_STATIC_CONSTANT(bool, value = (-Divisible<u, 2>::value) );
  };
  
  /// \brief
  /// absolute value calculation
  ///
  template <int N>
  struct Abs
  {
    BOOST_STATIC_CONSTANT(bool, value = (N >= 0) ? N : -N );
  };

// ______________________________________________________________________________________________________
//  greatest common devisor and least common multiple

  /// \brief
  /// calculate gcd of template parameters u and v
  ///
  template <int u, int v>
  struct Gcd
  {
    BOOST_STATIC_CONSTANT(int, value = (Gcd<v, u % v>::value) );
  };
    
  template <int u>
  struct Gcd<u, 0>
  {
    BOOST_STATIC_CONSTANT(int, value = u );
  };
    
  template <>
  struct Gcd<0, 0>
  {
    BOOST_STATIC_CONSTANT(int, value = -1 );
  };
  
  /// \brief
  /// calculate the least common multiple of template parameters u and v
  ///
  template <int u, int v>
  struct Lcm
  {
    BOOST_STATIC_CONSTANT(int, value = (Abs<u * v>::value / Gcd<u, v>::value) );
  };

// ______________________________________________________________________________________________________
// powers of values
 
  /// \brief
  /// calculate the power of template parameters a<sup>p</sup>
  ///
  template <int a, int p>
  struct Pow
  {
    BOOST_STATIC_CONSTANT(int, value = (a * Pow<a, p-1>::value) );
  };
  
  template <int a>
  struct Pow<a, 1>
  {
    BOOST_STATIC_CONSTANT(int, value = a );
  };
  
  template <int a>
  struct Pow<a, 0>
  {
    BOOST_STATIC_CONSTANT(int, value = 1 );
  };
  
  template <int n>
  struct Sqr
  {
    BOOST_STATIC_CONSTANT(int, value = (Pow<n, 2>::value) );
  };

// ______________________________________________________________________________________________________
// roots of values
 
  /// \brief
  /// template to compute sqrt(a) via iteration
  ///
  template <int a, int I=1>
  struct Sqrt {
      // instantiate next step or result type as branch
      typedef typename boost::mpl::if_c< (I*I < a),
					Sqrt<a, I+1>,
					boost::mpl::int_<I>
				      >::type impl;

      // use the result of branch type
      BOOST_STATIC_CONSTANT(int, value = impl::value );
  };
  
  // partial specialization to end the iteration
  template<int a>
  struct Sqrt<a, a> {
    BOOST_STATIC_CONSTANT(int, value = a );
  };
  template<int I>
  struct Sqrt<0, I> {
    BOOST_STATIC_CONSTANT(int, value = 0 );
  };
  
  /// \brief
  /// template to compute the Nth root of a via iteration
  ///
#ifndef _WIN32
  template <int a, int N, int I=1>
  struct Root {
    // instantiate next step or result type as branch
    typedef typename boost::mpl::if_c<Pow<I, N>::value < a,
				      Root<a, N, I+1>,
				      boost::mpl::int_<I>
				    >::type impl;

    // use the result of branch type
    BOOST_STATIC_CONSTANT(int, value = impl::value );
  };
  
  // partial specialization to end the iteration
  template<int a, int N>
  struct Root<a, N, a> {
    BOOST_STATIC_CONSTANT(int, value = a );
  };
  template<int N, int I>
  struct Root<0, N, I> {
    BOOST_STATIC_CONSTANT(int, value = 0 );
  };
  /**@}*/
#endif
// ______________________________________________________________________________________________________
// monomials and polynomials

  template<int N>
  struct Monomial {
    typedef double result_type;
    result_type operator()(double x) { return x*Monomial<N-1>()(x); }
  };
  
  template<>
  struct Monomial<1> {
    typedef double result_type;
    result_type operator()(double x) { return x; }
  };
  
  template<>
  struct Monomial<0> {
    typedef double result_type;
    result_type operator()(double x) { return 1.0; }
  };

  template<typename Coefficients, int I, int N>
  struct Horner {
    typedef double result_type;
    Horner(Coefficients& coefficients_) : coefficients(coefficients_) {}
    result_type operator()(double x) {
      return boost::fusion::at_c<I>(coefficients) + x*(Horner<Coefficients, I+1, N>(coefficients)(x));
    }
  private:
    Coefficients& coefficients;
  };

  template<typename Coefficients, int N>
  struct Horner<Coefficients, N, N> {
    typedef double result_type;
    Horner(Coefficients& coefficients_) : coefficients(coefficients_) {}
    result_type operator()(double x) {
      return boost::fusion::at_c<N>(coefficients);
    }
  private:
    Coefficients& coefficients;
  };

  template<typename Coefficients>
  struct Polynomial {
    typedef double result_type;
    Polynomial(Coefficients& coefficients_) : coefficients(coefficients_) {}
    result_type operator()(double x) {
      return Horner<Coefficients, 0, boost::fusion::result_of::size<Coefficients>::value - 1>(coefficients)(x);
    }
  private:
    Coefficients& coefficients;
  };

// ______________________________________________________________________________________________________
// Bernstein polynomials and Bezier-Surfaces

  template<int I, int order>
  struct BernsteinPolynomial 
  {
    static inline double value(double x) {
      return Binomial<order, I>::value
	      * std::pow(x, I) * std::pow(1.0 - x, order - I);
    }
  };
  
  template<int I, int order, bool valid = true>
  struct BernsteinPolynomial_
  {
    static inline double value(double x) {
      return (1.0 - x) * BernsteinPolynomial_<I, order - 1, I >= 0 && I <= (order-1)>::value(x)
	+ x * BernsteinPolynomial_<I - 1, order - 1, (I-1) >= 0 && (I-1) <= (order-1)>::value(x);
    }
  };
  
  template<>
  struct BernsteinPolynomial_<0, 0, true>
  {
    static inline double value(double x) {
      return 1.0;
    }
  };
  
  template<int I, int order>
  struct BernsteinPolynomial_<I, order, false>
  {
    static inline double value(double x) {
      return 0.0;
    }
  };
  
  template<int I, int J, int order>
  struct BernsteinPolynomial2
  {
    static inline double value(double x, double y) {
      return BernsteinPolynomial<I, order>::value(x) * BernsteinPolynomial<J, order>::value(x);
    }
  };
  
  template<int I, int J, int K, int order>
  struct BernsteinPolynomial3
  {
    static inline double value(double x, double y, double z) {
      return BernsteinPolynomial<I, order>::value(x) * BernsteinPolynomial<J, order>::value(x) * BernsteinPolynomial<K, order>::value(z);
    }
  };

// ______________________________________________________________________________________________________
// Bezier-Polynomial
#ifndef _WIN32
  namespace details {
    
    // sum_{I=0}^N B_I^N(x)*c_I
    template<int I, int N>
    struct BezierPolynomial
    {
      template<typename Coefficients>
      static double value(Coefficients& coefficients, double x) {
	return boost::fusion::at_c<I>(coefficients)*BernsteinPolynomial<I, N>::value(x) + details::BezierPolynomial<I+1, N>::value(coefficients, x);
      }
    };
    
    template<int N>
    struct BezierPolynomial<N, N>
    {
      template<typename Coefficients>
      static double value(Coefficients& coefficients, double x) {
	return boost::fusion::at_c<N>(coefficients)*BernsteinPolynomial<N,N>::value(x);
      }
    };
  }
  
  template<typename Coefficients, int DOW=1>
  struct BezierPolynomial
  {
    BOOST_STATIC_CONSTANT(int, N = (boost::mpl::minus< boost::mpl::int_< Root< boost::fusion::result_of::size<Coefficients>::value, 
									       DOW >::value >, 
						      boost::mpl::int_<1> >::type::value) ); // N = rootN(coeff.size, DOW)-1
    
    BezierPolynomial(Coefficients& coefficients_) : coefficients(coefficients_) {}
    double value(double x) {
      return details::BezierPolynomial<0, N>::value(coefficients, x);
    }
  private:
    Coefficients& coefficients;
  };

// ______________________________________________________________________________________________________

  
  namespace details {
    
    template<int IJ, int N, bool finished=false> // IJ = I + J*N, max(IJ) = N*N - 2
    struct BezierPolynomial2
    {
      template<typename Coefficients>
      static double value(Coefficients& coefficients, double x, double y) {
	return boost::fusion::at_c<N>(coefficients) * BernsteinPolynomial2<IJ%N, IJ/N, N>::value(x, y)
	  + details::BezierPolynomial2<IJ + 1, N, IJ == (N+1)*(N+1)>::value(coefficients, x, y);
      }
    };
    
    template<int IJ, int N>
    struct BezierPolynomial2<IJ, N, true>
    {
      template<typename Coefficients>
      static double value(Coefficients& coefficients, double x, double y) {
	return 0.0;
      }
    };
  }
    
  /// \brief
  /// interface for bezier-surfaces of order (N,N)
  ///
  template<typename Coefficients>
  struct BezierPolynomial2
  {
    BOOST_STATIC_CONSTANT(int, N = (boost::mpl::minus< 
				      boost::mpl::int_<
					Root<
					  boost::fusion::result_of::size<Coefficients>::value, 
					  2
					>::value
				      >, 
				      boost::mpl::int_<1>
				    >::type::value) ); // N = rootN(coeff.size, 2)-1
    
    BezierPolynomial2(Coefficients& coefficients_) : coefficients(coefficients_) {} 
    double value(double x, double y) {
      return details::BezierPolynomial2<0, N, N == 0>::value(coefficients, x, y);
    }
  private:
    Coefficients& coefficients; // p in points is element of result_type
  };
#endif
// ______________________________________________________________________________________________________
// for loops: for<i, n [,{incr,decr}]>, for<mpl::range_c>, for<mpl::vector_c>

  template<int I> struct incr { BOOST_STATIC_CONSTANT(int, value = I+1); };
  template<int I> struct decr { BOOST_STATIC_CONSTANT(int, value = I-1); };

  template<int I, int N, template<int> class op, bool finished, template<int> class F>
  struct FOR_LOOP {
    static void loop() {
      F<I>::eval(); // your implementation here
      FOR_LOOP<op<I>::value, N, op, op<I>::value == N, F>::loop();
    }
  };

  // initial/break condition for 'for-loop': e.g. I==N
  template<int I, int N, template<int> class op, template<int> class F> 
  struct FOR_LOOP<I, N, op, true, F> 
  {
    static void loop() {}
  };

  // 'for-loop', for I<N
  template<int I, int N, bool plus, bool finished, template<int> class F> 
  struct FOR_LOOP_CHECK 
  {
    static void loop() { FOR_LOOP<I, N, incr, finished, F>::loop(); }
  };

  // 'for-loop', for I>N
  template<int I, int N, bool finished, template<int> class F> 
  struct FOR_LOOP_CHECK<I,N,false,finished, F> 
  {
    static void loop() { FOR_LOOP<I, N, decr, finished, F>::loop(); }
  };

  /// \brief
  /// interface for classical for loop: for j=I..N
  /// the index will be incremented, if I&lt;N and decremented otherwise
  ///
  template<int I, int N, template<int> class F> 
  struct FOR 
  { 
    static void loop() { FOR_LOOP_CHECK<I, N, I < N, I == N, F>::loop(); } 
  };

  // for-loop over sequence of values
  template<typename Seq, long I, long N, template<int> class F>
  struct FOR_SUBSEQ
  {
    static void loop() {
      GET_LOOP_INDEX (i, Seq, I);
      F<i>::eval(); // your implementation here
      FOR_SUBSEQ<Seq, I+1, N, F>::loop();
    }
  };

  // initial/break condition for 'for-loop' over sequences: I==N
  template<typename Seq, long N, template<int> class F> 
  struct FOR_SUBSEQ<Seq,N,N,F> 
  { 
    static void loop() {} 
  };

  /// \brief
  /// interface for loop over sequence of values
  /// for i in {seq[0]..seq[end]}
  ///
  template<typename Seq, template<int> class F> 
  struct FOR_SEQ 
  { 
    static void loop() { FOR_SUBSEQ<Seq, 0, boost::mpl::size<Seq>::value, F>::loop(); } 
  };

//____________________________________________________________________________
// for-each loops

  template<int I> struct for_each
  {
    template<typename Container1, typename Container2, typename Binary_Functor>
    static void eval(Container1& vec1, Container2& vec2, Binary_Functor& f);
    
    template<typename Container1, typename Container2, typename T, typename Tertiary_Functor>
    static void accumulate(Container1& vec1, Container2& vec2, T& value, Tertiary_Functor& f);
  };

  template<> struct for_each<0>
  {
    template<typename Container1, typename Container2, typename Binary_Functor>
    static void eval(Container1& vec1, Container2& vec2, Binary_Functor& f);

    template<typename Container1, typename Container2, typename T, typename Tertiary_Functor>
    static void accumulate(Container1& vec1, Container2& vec2, T& value, Tertiary_Functor& f);
  };

  template<int I>
  template<typename Container1, typename Container2, typename Binary_Functor>
  void for_each<I>::eval(Container1& vec1, Container2& vec2, Binary_Functor& f) {
    f(boost::fusion::at_c<I>(vec1), boost::fusion::at_c<I>(vec2));
    for_each<I-1>::eval(vec1,vec2,f);
  }

  template<int I>
  template<typename Container1, typename Container2, typename T, typename Tertiary_Functor>
  void for_each<I>::accumulate(Container1& vec1, Container2& vec2, T& value, Tertiary_Functor& f) {
    f(boost::fusion::at_c<I>(vec1), boost::fusion::at_c<I>(vec2), value);
    for_each<I-1>::accumulate(vec1,vec2,value,f);
  }

  template<typename Container1, typename Container2, typename Binary_Functor>
  void for_each<0>::eval(Container1& vec1, Container2& vec2, Binary_Functor& f) {
    f(boost::fusion::at_c<0>(vec1), boost::fusion::at_c<0>(vec2));
  }

  template<typename Container1, typename Container2, typename T, typename Tertiary_Functor>
  void for_each<0>::accumulate(Container1& vec1, Container2& vec2, T& value, Tertiary_Functor& f) {
    f(boost::fusion::at_c<0>(vec1), boost::fusion::at_c<0>(vec2),value);
  }
  
} // namespace tools

#endif // EXTENSIONS_META_TOOLS_H
