/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/


#ifndef EXTENSIONS_P_OPERATORS_FOT_H
#define EXTENSIONS_P_OPERATORS_FOT_H

#include "AMDiS.h"

using namespace std;
using namespace AMDiS;


/// < factor*phase*grad(u) , grad(psi) >
class Phase_FOT : public FirstOrderTerm 
{
public:
/// Constructor.
Phase_FOT(DOFVectorBase<double>* phaseDV, double fac=1.0);

/// Implementation of \ref OperatorTerm::initElement().
void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
    Quadrature *quad = NULL);

/// Implementation of \ref OperatorTerm::initElement() for multilpe meshes.
void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
    SubAssembler* subAssembler,
    Quadrature *quad = NULL);

void getLb(const ElInfo *elInfo,
    vector<mtl::dense_vector<double> >& result) const;

void eval(int nPoints,
    const mtl::dense_vector<double>&,
    const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
    const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
    mtl::dense_vector<double>& result,
    double opFactor);

virtual WorldVector<double> f(const int iq) const;
void setFactor(const double fac_) { fac=fac_; }

protected:
DOFVectorBase<double> *phaseDV;
mtl::dense_vector<double> phase;
double fac;
};

/// < phase*(vecs*grad(u)) , psi > or < phase*vecs*u , grad(psi) > where vecs in R^n
class WorldVecAndVecFct_FOT : public FirstOrderTerm
{
public:
WorldVecAndVecFct_FOT(WorldVector<DOFVectorBase<double>*> vecs,
	DOFVectorBase<double>* phaseDV_, double fac=1.0);

void initElement(const ElInfo*, 
	SubAssembler* ,
	Quadrature *quad= NULL);
void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
	SubAssembler*,
	Quadrature *quad = NULL);

void getLb(const ElInfo *elInfo,
		vector<mtl::dense_vector<double> >& result) const;

void eval(int nPoints,
		const mtl::dense_vector<double>&,
		const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
		const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
		mtl::dense_vector<double>& result,
		double opFactor);

protected:
DOFVectorBase<double> *vec0DV, *vec1DV, *vec2DV, *phaseDV;
mtl::dense_vector<double> phase, vec0, vec1, vec2;
double fac;
int numVecs;
};



/* -------------------------------------------------------------- */

/// < vecs*grad(u) , psi > or < vecs*u , grad(psi) > where vecs in R^n
class WorldVec_FOT : public FirstOrderTerm
{
public:
WorldVec_FOT(WorldVector<DOFVector<double>*> vecs,double fac=1.0);

void initElement(const ElInfo*, 
	SubAssembler* ,
	Quadrature *quad= NULL);
void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
	SubAssembler*,
	Quadrature *quad = NULL);

void getLb(const ElInfo *elInfo,
		vector<mtl::dense_vector<double> >& result) const;

void eval(int nPoints,
		const mtl::dense_vector<double>&,
		const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
		const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
		mtl::dense_vector<double>& result,
		double opFactor);

protected:
DOFVectorBase<double> *vec0DV, *vec1DV, *vec2DV;
mtl::dense_vector<double> vec0, vec1, vec2;
double fac;
int numVecs;
};

/* -------------------------------------------------------------- */

/// First order term: \f$ < \vec{v}(\vec{x}) \cdot \nabla u(\vec{x}) , \theta > or < \vec{v}(\vec{x}) \cdot u(\vec{x}) , \nabla \theta >\f$.
class WorldVector_FOT : public FirstOrderTerm
{
public:
  /// Constructor.
  WorldVector_FOT(DOFVector<WorldVector<double> > *dv, double fac_=1.0);

  /// Implementation of \ref OperatorTerm::initElement().
  void initElement(const ElInfo* elInfo,
		   SubAssembler* subAssembler,
		   Quadrature *quad = NULL);
  
  void initElement(const ElInfo* largeElInfo,
		   const ElInfo* smallElInfo,
		   SubAssembler* subAssembler,
		   Quadrature *quad = NULL);

  /// Implements FirstOrderTerm::getLb().
  void getLb(const ElInfo *elInfo,
	     vector<mtl::dense_vector<double> >& Lb) const;

  /// Implements FirstOrderTerm::eval().
  void eval(int nPoints,
	    const mtl::dense_vector<double>&,
	    const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	    const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	    mtl::dense_vector<double>& result,
	    double factor);

protected:
  DOFVector<WorldVector<double> >* vec;

  /// Gradient of v at quadrature points.
  mtl::dense_vector<WorldVector<double> > vecAtQPs;

  double fac;
};

/// < phase*(vecs*grad(u)) , psi > or < phase*vecs*u , grad(psi) > where vecs in R^n
class WorldVecPhase_FOT : public FirstOrderTerm
{
public:
WorldVecPhase_FOT(DOFVectorBase<double>* phaseDV_, WorldVector<DOFVector<double>*> vecs,
    double fac=1.0);

void initElement(const ElInfo*, 
    SubAssembler* ,
    Quadrature *quad= NULL);
void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
    SubAssembler*,
    Quadrature *quad = NULL);

void getLb(const ElInfo *elInfo,
        vector<mtl::dense_vector<double> >& result) const;

void eval(int nPoints,
        const mtl::dense_vector<double>&,
        const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
        const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
        mtl::dense_vector<double>& result,
        double opFactor);

protected:
DOFVectorBase<double> *vec0DV, *vec1DV, *vec2DV, *phaseDV;
mtl::dense_vector<double> phase, vec0, vec1, vec2;
double fac;
int numVecs;
};



/* -------------------------------------------------------------- */

/// < f(vec0)*vecs*grad(u) , psi > or < f(vec0)*vecs*u , grad(psi) > where vecs in R^n
struct VecAndWorldVec_FOT : FirstOrderTerm
{
  VecAndWorldVec_FOT(DOFVector<double>* vec0, 
					 WorldVector<DOFVector<double>*> vecs, 
					 AbstractFunction<double, double>* fct, 
					 double fac = 1.0);

  void initElement(const ElInfo*, 
				   SubAssembler*,
				   Quadrature *quad= NULL);
  
  void initElement(const ElInfo* largeElInfo, 
				   const ElInfo* smallElInfo,
				   SubAssembler*,
				   Quadrature *quad = NULL);

  void getLb(const ElInfo *elInfo,
			 vector<mtl::dense_vector<double> >& result) const;

  void eval(int nPoints,
			const mtl::dense_vector<double>&,
			const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
			const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
			mtl::dense_vector<double>& result,
			double opFactor);

private:
  DOFVector<double> *vDV, *vec0DV, *vec1DV, *vec2DV;
  mtl::dense_vector<double> v, vec0, vec1, vec2;
  AbstractFunction<double, double>* fct;
  double fac;
  int numVecs;
};

/* -------------------------------------------------------------- */

/// < factor*d_i(u) , psi > or < factor*u , d_i(psi) >
class PartialDerivative_FOT : public FirstOrderTerm
{
public:
PartialDerivative_FOT(int component_, double fac_=1.0);

void initElement(const ElInfo*, 
	SubAssembler*,
	Quadrature *quad= NULL);
void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
	SubAssembler*,
	Quadrature *quad = NULL);

void getLb(const ElInfo *elInfo,
		vector<mtl::dense_vector<double> >& result) const;

void eval(int nPoints,
		const mtl::dense_vector<double>&,
		const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
		const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
		mtl::dense_vector<double>& result,
		double opFactor);

protected:
int component;
double fac;
};

/* -------------------------------------------------------------- */

/// < factor*v1*d_i(u) , psi > or < factor*v1*u , d_i(psi) >
class VecAndPartialDerivative_FOT : public FirstOrderTerm
{
public:

VecAndPartialDerivative_FOT(DOFVectorBase<double> *dv1, int component_, 
  double fac_=1.0);

void initElement(const ElInfo* elInfo,
  SubAssembler* subAssembler,
  Quadrature *quad = NULL);
void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
  SubAssembler* subAssembler,
  Quadrature *quad = NULL);

void getLb(const ElInfo *elInfo, vector<mtl::dense_vector<double> >& result) const;

void eval(int nPoints,
  const mtl::dense_vector<double>&,
  const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
  const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
  mtl::dense_vector<double>& result,
  double opFactor);

protected:
DOFVectorBase<double>* vec1DV;
mtl::dense_vector<double> vec1;
double fac;
int component;
};

/* -------------------------------------------------------------- */

/// < d_i(u) * factor*v1*d_j(v2) , psi > or < u * factor*v1*d_j(v2) , d_i(psi) >
struct VecAndPartialDerivativeIJ_FOT : FirstOrderTerm
{
  VecAndPartialDerivativeIJ_FOT(DOFVectorBase<double> *dv1, 
				DOFVectorBase<double> *dv2,
				int i_,
				int j_,
				double fac_ = 1.0);

  VecAndPartialDerivativeIJ_FOT(DOFVectorBase<double> *dv1, 
				DOFVectorBase<double> *dv2,
				int i_,
				int j_,
				BinaryAbstractFunction<double, double, double>* fct, // = f(v1, d_j(v2))
				double fac_ = 1.0);

  void initElement(const ElInfo* elInfo,
	SubAssembler* subAssembler,
	Quadrature *quad = NULL);

  void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
	SubAssembler* subAssembler,
	Quadrature *quad = NULL);

  void getLb(const ElInfo *elInfo, vector<mtl::dense_vector<double> >& result) const;

  void eval(int nPoints,
	const mtl::dense_vector<double>&,
	const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	mtl::dense_vector<double>& result,
	double opFactor);

private:
  DOFVectorBase<double>* vec1DV;
  DOFVectorBase<double>* vec2DV;
  mtl::dense_vector<double> vec1;
  mtl::dense_vector<WorldVector<double> > grad;
  BinaryAbstractFunction<double, double, double>* fct;
  double fac;
  int xi, xj;
};


/* -------------------------------------------------------------- */

/// < factor*v1*v2*d_i(u) , psi > or < factor*v1*v2*u , d_i(psi) >
struct Vec2AndPartialDerivative_FOT : public FirstOrderTerm
{
  Vec2AndPartialDerivative_FOT(DOFVectorBase<double> *dv1, 
			       DOFVectorBase<double> *dv2,
			       int component_, 
			       double fac_=1.0);
  
  Vec2AndPartialDerivative_FOT(DOFVectorBase<double> *dv1, 
			       DOFVectorBase<double> *dv2,
			       int component_,
			       BinaryAbstractFunction<double, double, double>* fct, // = f(v1, v2)
			       double fac_=1.0);

  void initElement(const ElInfo* elInfo,
	  SubAssembler* subAssembler,
	  Quadrature *quad = NULL);
  
  void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
	  SubAssembler* subAssembler,
	  Quadrature *quad = NULL);

  void getLb(const ElInfo *elInfo, vector<mtl::dense_vector<double> >& result) const;

  void eval(int nPoints,
	  const mtl::dense_vector<double>&,
	  const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	  const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	  mtl::dense_vector<double>& result,
	  double opFactor);

private:
  DOFVectorBase<double>* vec1DV;
  DOFVectorBase<double>* vec2DV;
  mtl::dense_vector<double> vec1,vec2;
  BinaryAbstractFunction<double, double, double>* fct;
  double fac;
  int component;
};

/* -------------------------------------------------------------- */

/**
  * \ingroup Assembler
  *
  * \brief
  * First order term: \f$ b(v(\vec{x}), \nabla w_1(\vec{x}), \nabla w_2(\vec{x})) \cdot \nabla u(\vec{x})\f$.
  */
class VecGrad2_FOT : public FirstOrderTerm
{
public:
  /// Constructor
  VecGrad2_FOT(DOFVectorBase<double> *dv, DOFVectorBase<double> *dw1, DOFVectorBase<double> *dw2,
	      TertiaryAbstractFunction<WorldVector<double>, double, WorldVector<double>, WorldVector<double> > *vecFct_);

  /// Implementation of \ref OperatorTerm::initElement().
  void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
		    Quadrature *quad = NULL);
  
  /// Implements FirstOrderTerm::getLb().
  void getLb(const ElInfo *elInfo,
	      vector<mtl::dense_vector<double> >& Lb) const; 
  
  /// Implements FirstOrderTerm::eval().
  void eval(int nPoints,
	    const mtl::dense_vector<double>& uhAtQP,
	    const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	    const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	    mtl::dense_vector<double>& result,
	    double factor);
  
protected:
  /// DOFVector to be evaluated at quadrature points.
  DOFVectorBase<double>* vec0;
  
  /// DOFVector to be evaluated at quadrature points.
  DOFVectorBase<double>* vec1;

  /// DOFVector to be evaluated at quadrature points.
  DOFVectorBase<double>* vec2;
  
  mtl::dense_vector<double> vec0AtQPs;
  
  /// Gradient v at quadrature points.
  mtl::dense_vector<WorldVector<double> > grad1AtQPs;

  /// Gradient v at quadrature points.
  mtl::dense_vector<WorldVector<double> > grad2AtQPs;

  /// Function for b.
  TertiaryAbstractFunction<WorldVector<double>, double, WorldVector<double>, WorldVector<double> > *vecFct;
};


#endif // EXTENSIONS_P_OPERATORS_FOT_H
