/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/


#ifndef EXTENSIONS_P_OPERATORS_SOT_H
#define EXTENSIONS_P_OPERATORS_SOT_H

#include "AMDiS.h"

using namespace std;
using namespace AMDiS;

/// < factor*phase*grad(u) , grad(psi) >
class Phase_SOT : public SecondOrderTerm 
{
public:
/// Constructor.
Phase_SOT(DOFVectorBase<double>* phaseDV, double fac=1.0);

/// Implementation of \ref OperatorTerm::initElement().
void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
		Quadrature *quad = NULL);

/// Implementation of \ref OperatorTerm::initElement() for multilpe meshes.
void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
		SubAssembler* subAssembler,
		Quadrature *quad = NULL);

/// Implements SecondOrderTerm::getLALt().
inline void getLALt(const ElInfo *elInfo, std::vector<mtl::dense2D<double> > &LALt) const;    

/// Implenetation of SecondOrderTerm::eval().
inline void eval(int nPoints,
		const mtl::dense_vector<double>&,
		const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
		const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
		mtl::dense_vector<double>& result,
		double opFactor);

/// Implenetation of SecondOrderTerm::weakEval().
void weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
		std::vector<WorldVector<double> > &result);

virtual double f(const int iq) const;
void setFactor(const double fac_) { fac=fac_; }

protected:
DOFVectorBase<double> *phaseDV;
mtl::dense_vector<double> phase;
double fac;
};

/* -------------------------------------------------------------- */

/// < factor*phase*A*grad(u) , grad(psi) > ... 'I' kann ueberladen werden
class MatrixPhase_SOT : public SecondOrderTerm
{
public:
/// Constructor
MatrixPhase_SOT(DOFVectorBase<double>* phaseDV, int comp, double fac=1.0);

/// Implementation of \ref OperatorTerm::initElement().
void initElement(const ElInfo* elInfo, 
        SubAssembler* subAssembler,
        Quadrature *quad=NULL);
void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
        SubAssembler* subAssembler,
        Quadrature *quad=NULL);
inline void lDlt(const DimVec<WorldVector<double> >& Lambda, 
            const WorldVector<double>& vec,
            mtl::dense2D<double>& LALt,
            double factor) const;
void getLALt(const ElInfo *elInfo,
        std::vector<mtl::dense2D<double> > &LALt) const; 

void eval(int nPoints,
        const mtl::dense_vector<double>& uhAtQP,
        const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
        const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
        mtl::dense_vector<double>& result,
        double opFactor);

void weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
        std::vector<WorldVector<double> > &result);

inline WorldVector<double> f(const int iq) const;
void setFactor(const double fac_) { fac=fac_; }

protected:
DOFVectorBase<double> *phaseDV;
mtl::dense_vector<double> phase;
int comp;
double fac;
bool symmetric;
};

/* -------------------------------------------------------------- */

/// < factor*()_ij*grad(u) , grad(psi) >
class MatrixIJ_SOT : public SecondOrderTerm
{
public:
/// Constructor
MatrixIJ_SOT(int row, int col, double fac=1.0);

/// Implementation of \ref OperatorTerm::initElement().
void initElement(const ElInfo* elInfo, 
        SubAssembler* subAssembler,
        Quadrature *quad=NULL) {}
void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
        SubAssembler* subAssembler,
        Quadrature *quad=NULL) {}
void getLALt(const ElInfo *elInfo,
        std::vector<mtl::dense2D<double> > &LALt) const; 
void eval(int nPoints,
        const mtl::dense_vector<double>& uhAtQP,
        const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
        const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
        mtl::dense_vector<double>& result,
        double opFactor);
void weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
        std::vector<WorldVector<double> > &result);
void setFactor(const double fac_) { fac=fac_; }

protected:
int row,col;
double fac;
bool symmetric;
};

/* -------------------------------------------------------------- */

/// < factor*phase*A*grad(u) , grad(psi) > ... 'I' kann ueberladen werden
class MatrixIJPhase_SOT : public SecondOrderTerm
{
public:
/// Constructor
MatrixIJPhase_SOT(DOFVectorBase<double>* phaseDV, int row, int col, double fac=1.0);

/// Implementation of \ref OperatorTerm::initElement().
void initElement(const ElInfo* elInfo, 
        SubAssembler* subAssembler,
        Quadrature *quad=NULL);
void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
        SubAssembler* subAssembler,
        Quadrature *quad=NULL);
void getLALt(const ElInfo *elInfo,
        std::vector<mtl::dense2D<double> > &LALt) const; 
void eval(int nPoints,
        const mtl::dense_vector<double>& uhAtQP,
        const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
        const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
        mtl::dense_vector<double>& result,
        double opFactor);
void weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
        std::vector<WorldVector<double> > &result);
void setFactor(const double fac_) { fac=fac_; }

protected:
DOFVectorBase<double> *phaseDV;
mtl::dense_vector<double> phase;
int row,col;
double fac;
bool symmetric;
};


#endif // EXTENSIONS_P_OPERATORS_SOT_H
