/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/


#ifndef EXTENSIONS_P_OPERATORS_ZOT_H
#define EXTENSIONS_P_OPERATORS_ZOT_H

#include "AMDiS.h"

using namespace std;
using namespace AMDiS;

/// < factor*phase*u , psi >
class Phase_ZOT : public ZeroOrderTerm 
{ 
public: 
Phase_ZOT(DOFVectorBase<double>* phaseDV, double fac=1.0);

/// Implementation of \ref OperatorTerm::initElement().
void initElement(const ElInfo* elInfo, 
		SubAssembler* subAssembler,
		Quadrature *quad=NULL);
void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
		SubAssembler* subAssembler,
		Quadrature *quad=NULL);

void getC(const ElInfo *, int nPoints, ElementVector &C);

void eval(int nPoints,
		const mtl::dense_vector<double>&,
		const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
		const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
		mtl::dense_vector<double>& result,
		double opFactor);

virtual double f(const int iq) const;
void setFactor(const double fac_) { fac=fac_; }

protected:
DOFVectorBase<double> *phaseDV;
mtl::dense_vector<double> phase;
double fac;
}; 

/* -------------------------------------------------------------- */

/// < factor*(1-phase)*u , psi > or < factor*v[i]*(1-phase)*u , psi > where v in R^n
class PhaseInverse_ZOT : public ZeroOrderTerm 
{ 
public: 
PhaseInverse_ZOT(DOFVectorBase<double>* phaseDV, double fac=1.0);
PhaseInverse_ZOT(DOFVectorBase<double>* phaseDV, double fac, WorldVector<double>* facVec, int component);

/// Implementation of \ref OperatorTerm::initElement().
void initElement(const ElInfo* elInfo, 
		SubAssembler* subAssembler,
		Quadrature *quad=NULL);
void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
		SubAssembler* subAssembler,
		Quadrature *quad=NULL);

void getC(const ElInfo *, int nPoints, ElementVector &C);

void eval(int nPoints,
		const mtl::dense_vector<double>&,
		const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
		const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
		mtl::dense_vector<double>& result,
		double opFactor);

double f(const int iq) const;
void setFactor(const double fac_) { fac=fac_; }

protected:
DOFVectorBase<double> *phaseDV;
mtl::dense_vector<double> phase;
double fac;
int component;
WorldVector<double>* facVec;
}; 

/* -------------------------------------------------------------- */

/// < factor*d_i(v)*u , psi > or < factor*b[j]*d_i(v)*u , psi >
class PartialDerivative_ZOT : public ZeroOrderTerm
{
public:
PartialDerivative_ZOT(DOFVectorBase<double>* vecDV_, int component_, double fac_=1.0);
PartialDerivative_ZOT(DOFVectorBase<double>* vecDV_, int component_, double fac_, WorldVector<double>* vec, int component2_);

void initElement(const ElInfo*, 
	SubAssembler*,
	Quadrature *quad= NULL);
void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
	SubAssembler*,
	Quadrature *quad = NULL);

void getC(const ElInfo *, int nPoints, ElementVector &C);

void eval(int nPoints,
		const mtl::dense_vector<double>&,
		const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
		const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
		mtl::dense_vector<double>& result,
		double opFactor);

protected:
DOFVectorBase<double> *vecDV;
mtl::dense_vector<WorldVector<double> > grad;
double fac;
WorldVector<double>* facVec;
int component, component2;
};

/* -------------------------------------------------------------- */

/// < factor*v*d_i(w)*u , psi >
struct VecAndPartialDerivative_ZOT : ZeroOrderTerm
{
  VecAndPartialDerivative_ZOT(DOFVectorBase<double>* vecDV_, 
			      DOFVectorBase<double>* gradDV_, 
			      int component_, 
			      double fac_ = 1.0);
  
  VecAndPartialDerivative_ZOT(DOFVectorBase<double>* vecDV_, 
			      DOFVectorBase<double>* gradDV_, 
			      int component_, 
			      AbstractFunction<double, double>* fct, // f(v)
			      double fac_ = 1.0);

  void initElement(const ElInfo*, 
	  SubAssembler*,
	  Quadrature *quad= NULL);
  
  void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
	  SubAssembler*,
	  Quadrature *quad = NULL);

  void getC(const ElInfo *, int nPoints, ElementVector &C);

  void eval(int nPoints,
		  const mtl::dense_vector<double>&,
		  const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
		  const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
		  mtl::dense_vector<double>& result,
		  double opFactor);

private:
  DOFVectorBase<double> *vecDV, *gradDV;
  int component;
  AbstractFunction<double, double>* fct;
  double fac;
  
  mtl::dense_vector<double> vec;
  mtl::dense_vector<WorldVector<double> > grad;
};

/* -------------------------------------------------------------- */

/// < factor*v1*v2*d_i(w)*u , psi >
struct Vec2AndPartialDerivative_ZOT : ZeroOrderTerm
{
  Vec2AndPartialDerivative_ZOT(DOFVectorBase<double>* vec1DV_, 
							  DOFVectorBase<double>* vec2DV_, 
							  DOFVectorBase<double>* gradDV_, 
							  int component_, 
							  double fac_ = 1.0);
  
  Vec2AndPartialDerivative_ZOT(DOFVectorBase<double>* vec1DV_, 
							  DOFVectorBase<double>* vec2DV_, 
							  DOFVectorBase<double>* gradDV_, 
							  int component_, 
							  BinaryAbstractFunction<double, double, double>* fct,
							  double fac_ = 1.0);

  void initElement(const ElInfo*,
	  SubAssembler*,
	  Quadrature *quad= NULL);

  void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
	  SubAssembler*,
	  Quadrature *quad = NULL);

  void getC(const ElInfo *, int nPoints, ElementVector &C);

  void eval(int nPoints,
		  const mtl::dense_vector<double>&,
		  const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
		  const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
		  mtl::dense_vector<double>& result,
		  double opFactor);

private:
  DOFVectorBase<double> *vec1DV, *vec2DV, *gradDV;
  mtl::dense_vector<double> vec1, vec2;
  BinaryAbstractFunction<double, double, double>* fct;
  double fac;
  mtl::dense_vector<WorldVector<double> > grad;
  int component;
};

#endif // EXTENSIONS_P_OPERATORS_ZOT_H
