/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/


#ifndef EXTENSIONS_PHASE_FIELD_CONVERT_H
#define EXTENSIONS_PHASE_FIELD_CONVERT_H

#include "Helpers.h"

using namespace AMDiS;

/// \brief
/// Converts an AbstractFunction <i>dist</i>, that describes a signed distance function, to
/// a phasefield function <i>p</i>, by \f$ \frac{1}{2}(1 - tanh(s \cdot dist(x) / \epsilon))\f$
///
struct SignedDistFctToPhaseField : AbstractFunction<double, WorldVector<double> >
{
  SignedDistFctToPhaseField(double epsilon_, 
			    AbstractFunction<double,WorldVector<double> > *dist_, 
			    double scalingFactor_ = 1.0/sqrt(2.0))
  : AbstractFunction<double, WorldVector<double> >(6), 
    epsilon(epsilon_), 
    dist(dist_),
    scalingFactor(scalingFactor_) {};

  double operator()(const WorldVector<double> &x) const
  {
    return 0.5 * (1.0 - tanh(scalingFactor * (*dist)(x) / epsilon));
  }
        
private:

  double epsilon;
  AbstractFunction<double,WorldVector<double> > *dist;
  double scalingFactor;
};


/// \brief
/// Converts a DOFVector, that describes a signed distance function, to
/// a phasefield function <i>p</i>, by \f$ \frac{1}{2}(1 - tanh(s \cdot dist / \epsilon))\f$.
/// You have to use <i>transformDOF</i> to apply this function to the DOFVector <i>dist</i>.
///
struct SignedDistToPhaseField : AbstractFunction<double, double>
{
  SignedDistToPhaseField(double epsilon_ = -1.0, double scalingFactor_ = 1.0/sqrt(2.0))
  : AbstractFunction<double, double>(6),
    epsilon(epsilon_),
    scalingFactor(scalingFactor_)
  {
    if (epsilon < 0.0)
      Parameters::get("mesh->refinement->epsilon", epsilon);
  }
  
  double operator()(const double &dist) const
  {
    return 0.5 * (1.0 - tanh(scalingFactor * dist / epsilon));
  }
  
private:
  double epsilon;
  double scalingFactor;
};


/// \brief
/// Converts a DOFVector, that describes a signed distance function, to
/// a phasefield function <i>p</i> with values in [-1,1], by \f$ - tanh(s \cdot dist / \epsilon)\f$.
/// You have to use <i>transformDOF</i> to apply this function to the DOFVector <i>dist</i>.
///
struct SignedDistToCh : AbstractFunction<double, double>
{
  SignedDistToCh(double epsilon_ = -1.0, double scalingFactor_ = 1.0/sqrt(2.0))
  : AbstractFunction<double, double>(6),
    epsilon(epsilon_),
    scalingFactor(scalingFactor_)
  {
    if (epsilon < 0.0)
      Parameters::get("mesh->refinement->epsilon", epsilon);
  }
  
  double operator()(const double &dist) const
  {
    return -tanh(scalingFactor * dist / epsilon);
  }
  
private:
  double epsilon;
  double scalingFactor;
};


/// \brief
/// Converts a vector of AbstractFunctions <i>{dist_i}</i>, that describe signed distance functions, to
/// a phasefield function <i>p</i>, by \f$ \frac{1}{2}(1 - tanh(s \cdot \min_i(dist_i(x)) / \epsilon))\f$.
/// The minimum of all distance function describes the union of the areas of negative values, of the 
/// distance functions.
///
struct SignedDistFctListToPhaseField : AbstractFunction<double, WorldVector<double> >
{
  SignedDistFctListToPhaseField(double epsilon_, 
				std::vector<AbstractFunction<double,WorldVector<double> >*> dist_, 
				double scalingFactor_ = 1.0/sqrt(2.0))
  : AbstractFunction<double, WorldVector<double> >(6), 
    epsilon(epsilon_), 
    dist(dist_),
    scalingFactor(scalingFactor_) {};

  double operator()(const WorldVector<double> &x) const
  {
    double d = 1.e12;
    for (size_t i = 0; i < dist.size(); ++i) {
      d = std::min((*(dist[i]))(x), d);
    }
    return 0.5 * (1.0 - tanh(scalingFactor * d / epsilon));
  }
        
private:

  double epsilon;
  std::vector<AbstractFunction<double,WorldVector<double> >*> dist;
  double scalingFactor;
};


/// \brief
/// Calculates the maximum of vector of distance function. This describes the intersections of the areas
/// of negative values, of the distance functions.
///
struct SignedDistList : AbstractFunction<double, WorldVector<double> >
{
  SignedDistList(std::vector<AbstractFunction<double,WorldVector<double> >*> dist_)
  : AbstractFunction<double, WorldVector<double> >(1), 
    dist(dist_) {};

  double operator()(const WorldVector<double> &x) const
  {
    d.setX(x);
    return std::for_each(dist.begin(), dist.end(), d);
  }

private:  
  struct MyMax {
    mutable double val;
    mutable WorldVector< double > x;
    MyMax()
    : val(-1.0e12) {}
    
    void operator()(const AbstractFunction<double,WorldVector<double> >* v) const
    {
      val = max((*v)(x), val);
    }
    
    void setX(const WorldVector<double> &x_) const
    {
      x= x_;
    }
    
    inline operator double() { return val; }
  } d;
  
  std::vector<AbstractFunction<double,WorldVector<double> >*> dist;
};


/// \brief
/// Converts a DOFVector, that describes a phasefield function <i>phi</i>, to
/// a signed distance function <i>dist</i>, by \f$ atanh(-\phi)\cdot\epsilon/s \f$.
/// You have to use <i>transformDOF</i> to apply this function to the DOFVector <i>phi</i>.
/// the phasefield values are cutted to allow the atanh calculation, by
/// \f$ \phi:=\max(-1 + 10^{-10}, min(1-10^{-10}, \phi) ) \f$
///
struct PhaseFieldToSignedDist : AbstractFunction<double, double>
{
  PhaseFieldToSignedDist(double epsilon_= -1.0, double scalingFactor_ = 1.0/sqrt(2.0))
  : AbstractFunction<double, double>(6), 
    epsilon(epsilon_),
    scalingFactor(scalingFactor_)
  {
    if (epsilon < 0.0)
      Parameters::get("mesh->refinement->epsilon", epsilon);
  }
  
  double operator()(const double &phi) const
  {
    double z = std::max(-1.0 + 1.e-12, std::min(1.0 - 1.e-12, -phi));
    return (epsilon/scalingFactor) * Helpers::atanh(z);
  }
  
private:
  double epsilon;
  double scalingFactor;
};


/// \brief
/// Converts a DOFVector, that describes a phasefield function <i>c</i> with
/// values in [-1, 1], to a phasefield function with values in [0,1], by 
/// \f$ \frac{1}{2}(c + 1) \f$.
/// You have to use <i>transformDOF</i> to apply this function to the DOFVector <i>phi</i>.
///
struct ChToPhaseField : AbstractFunction<double, double>
{
  ChToPhaseField() : AbstractFunction<double, double>(1) {};
  double operator()(const double &c) const
  {
    return std::max(0.0, std::min(1.0, 0.5 * (c + 1.0)));
  }
};


/// \brief
/// Converts a DOFVector, that describes a phasefield function <i>phi</i> with
/// values in [0, 1], to a phasefield function with values in [-1,1], by 
/// \f$ 2\cdot\phi-1 \f$.
/// You have to use <i>transformDOF</i> to apply this function to the DOFVector <i>phi</i>.
///
struct PhaseFieldToCh : AbstractFunction<double, double>
{
  PhaseFieldToCh() : AbstractFunction<double, double>(1) {};
  double operator()(const double &c) const
  {
    return std::max(-1.0, std::min(1.0, 2.0 * c - 1.0));
  }
};

#endif // EXTENSIONS_PHASE_FIELD_CONVERT_H
