/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/


#ifndef EXTENSIONS_SINGULAR_DIRICHLET_BC2_H
#define EXTENSIONS_SINGULAR_DIRICHLET_BC2_H

#include "AMDiS.h"
#include "boost/type_traits.hpp"
#include <boost/mpl/or.hpp>

namespace AMDiS {

/**
 * data structure that holds DOF-Index and double value for singular dirichlet boundary condition.
 **/
struct SingularDirichletBC {
  SingularDirichletBC(int i_, int j_, DegreeOfFreedom idx_, double value_) :
    row(i_), col(j_), idx(idx_), value(value_) {}

  int row, col;
  DegreeOfFreedom idx;
  double value;
};

/**
 * data structure that holds all periodic associations.
 **/
struct ManualPeriodicBC {
  ManualPeriodicBC(size_t row_, std::vector<std::pair<DegreeOfFreedom, DegreeOfFreedom> > &indices_) :
    row(row_), indices(indices_) {}

  size_t row;
  std::vector<std::pair<DegreeOfFreedom, DegreeOfFreedom> > indices;
};

namespace details {
    
  inline bool dofOnFace(int dim, int face, Element* el, ElementDofIterator& elDofIter)
  {
    GeoIndex boundaryPos = (dim == 1 ? VERTEX : (dim == 2 ? EDGE : FACE));
    
    // all DOFs on the face
    if (elDofIter.getPosIndex() == boundaryPos && 
	elDofIter.getCurrentElementPos() == face)
      return true;
    
    // Vertex DOFs on the face. Since position-number of the face
    // is associated with the vertex opposite to the face all vertices
    // with position != face number can be collected
    if (elDofIter.getPosIndex() == VERTEX &&
	elDofIter.getCurrentElementPos() != face)
      return true;
    
    // in 3d the DOFs on the edges on the face must be collected
    if (dim == 3) {
      if (elDofIter.getPosIndex() == EDGE) {
	int localEdge = elDofIter.getCurrentElementPos();
	if (el->getEdgeOfFace(face,0) == localEdge ||
	    el->getEdgeOfFace(face,1) == localEdge ||
	    el->getEdgeOfFace(face,2) == localEdge)
	  return true;
      }
    }
    
    return false;
  }
  
  template<typename PosType>
  void getIndicesAux(PosType &pos, 
		  BoundaryType b, 
		  const FiniteElemSpace* 
		  feSpace, std::map<DegreeOfFreedom, double> &indices, 
		  boost::true_type is_abstract_fct);
  
  // template specialization for AbstractFunctions
  template<>
  void getIndicesAux<AbstractFunction<double, WorldVector<double> > >(
		  AbstractFunction<double, WorldVector<double> > &pos, 
		  BoundaryType b, 
		  const FiniteElemSpace* 
		  feSpace, std::map<DegreeOfFreedom, double> &indices, 
		  boost::true_type is_abstract_fct);
  template<>
  void getIndicesAux<AbstractFunction<bool, WorldVector<double> > >(
		  AbstractFunction<bool, WorldVector<double> > &pos, 
		  BoundaryType b, 
		  const FiniteElemSpace* 
		  feSpace, std::map<DegreeOfFreedom, double> &indices, 
		  boost::true_type is_abstract_fct);
  
  template<typename PosType>
  void getIndicesAux(PosType &pos, 
		  BoundaryType b, 
		  const FiniteElemSpace* feSpace, 
		  std::map<DegreeOfFreedom, double> &indices, 
		  boost::false_type is_not_abstract_fct);
  
  template<typename PosType>
  void getIndices(PosType &pos, 
		  BoundaryType b, 
		  const FiniteElemSpace* feSpace, 
		  std::map<DegreeOfFreedom, double> &indices)
  {
    details::getIndicesAux(pos, b, feSpace, indices, 
	typename boost::conditional<
	  boost::mpl::or_<
	    boost::is_base_of<AbstractFunction<double, WorldVector<double> >, PosType>,
	    boost::is_base_of<AbstractFunction<bool, WorldVector<double> >, PosType>
	  >::value,
	  boost::true_type,
	  boost::false_type
	>::type());
  }

  template<typename ValueType>
  void getValuesAux(const ValueType &val, 
		 const FiniteElemSpace* feSpace, 
		 std::map<DegreeOfFreedom, double> &indices, 
		 boost::true_type is_abstract_fct,
		 int depth = 0)
  {
    TEST_EXIT(depth == 0)("Recursion bricht nicht ab!\n");
    details::getValuesAux(dynamic_cast<const AbstractFunction<double, WorldVector<double> >&>(val), feSpace, indices, is_abstract_fct, depth+1);
  }
  // template specialization for AbstractFunctions
  template<>
  void getValuesAux<AbstractFunction<double, WorldVector<double> > >(
		  const AbstractFunction<double, WorldVector<double> >& val, 
		  const FiniteElemSpace *feSpace, 
		  std::map<DegreeOfFreedom, double> &indicesValues,
		  boost::true_type is_abstract_fct,
		  int depth);
    
  template<typename ValueType>
  void getValuesAux(const ValueType &val, 
		 const FiniteElemSpace* feSpace, 
		 std::map<DegreeOfFreedom, double> &indices, 
		 boost::false_type is_not_abstract_fct,
		 int depth = 0);

  template<typename ValueType>
  void getValues(const ValueType &val, 
		 const FiniteElemSpace* feSpace, 
		 std::map<DegreeOfFreedom, double> &indices)
  {
    details::getValuesAux(val, feSpace, indices, typename boost::is_base_of<AbstractFunction<double, WorldVector<double> >, ValueType>::type());
  }

  /**
   * traverses the mesh and stores all DegreeOfFreedom, that lie on the boundary with index 'nr'
   * and have coordinates that give true in the 'meshIndicator', into a vector of 'indices'.
   **/
  inline void getBoundaryIndices(const FiniteElemSpace* feSpace,
			  BoundaryType nr,
			  AbstractFunction<bool, WorldVector<double> >* meshIndicator,
			  std::map<DegreeOfFreedom, double> &indices)
  {
    Mesh* mesh = feSpace->getMesh();
    int dim = mesh->getDim();
    const BasisFunction *basFcts = feSpace->getBasisFcts();
    int numBasFcts = basFcts->getNumber();
    std::vector<DegreeOfFreedom> localIndices(numBasFcts);

    ElementDofIterator elDofIter(feSpace, true);    

    WorldVector<double> coord;
    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(mesh, -1,
      Mesh::CALL_LEAF_EL | Mesh::FILL_BOUND | Mesh::FILL_COORDS);
    while (elInfo) {
      for (int face = 0; face < dim + 1; face++) {
	if ((nr && elInfo->getBoundary(face) == nr) || 
	    (!nr && elInfo->getBoundary(face) != 0)) {
	  if (!meshIndicator) {
	    elDofIter.reset(elInfo->getElement());
	    do {
	      if (dofOnFace(dim, face, elInfo->getElement(), elDofIter))
		indices[elDofIter.getDof()] = 0.0;
	    } while (elDofIter.next()); 
	  } else {
	    basFcts->getLocalIndices(elInfo->getElement(), feSpace->getAdmin(), localIndices);
	    for (int i = 0; i < numBasFcts; i++) {
	      elInfo->coordToWorld(*(basFcts->getCoords(i)), coord);
	      if ((*meshIndicator)(coord))
		indices[localIndices[i]] = 0.0;
	    }
	  }
	  break;
	}
      }

      elInfo = stack.traverseNext(elInfo);
    }
  }

  inline void getImplicitIndices(const FiniteElemSpace* feSpace,
			  AbstractFunction<double, WorldVector<double> >* signedDistFct,
			  std::map<DegreeOfFreedom, double> &indices)
  {
    const BasisFunction *vecBasisFcts = feSpace->getBasisFcts();
    int nVecBasisFcts = vecBasisFcts->getNumber();

    std::vector<DegreeOfFreedom> vecLocalIndices(nVecBasisFcts);
    
    DOFVector<short int> visited(feSpace, "visited"); visited.set(0);
    WorldVector<double> coord;
    TraverseStack stack;
    Flag traverseFlag = Mesh::CALL_LEAF_EL | Mesh::FILL_COORDS;
    ElInfo *elInfo = stack.traverseFirst(feSpace->getMesh(), -1, traverseFlag);

    while (elInfo) {
      Element *el = elInfo->getElement();
      vecBasisFcts->getLocalIndices(el, feSpace->getAdmin(), vecLocalIndices);

      for (int i = 0; i < nVecBasisFcts; i++) {
	if ( visited[vecLocalIndices[i]] == 0 ) {
	  elInfo->coordToWorld(*(vecBasisFcts->getCoords(i)), coord);
	  if ( (*signedDistFct)(coord) <= DBL_TOL ) { // eventuell nur Koordinaten an Gitterknoten bekommen!
	    indices[vecLocalIndices[i]] = 0.0;
	  }
	  visited[vecLocalIndices[i]] = 1;
	}
      }
      elInfo = stack.traverseNext(elInfo); 
    }
  }
  
  inline void getImplicitIndices(const FiniteElemSpace* feSpace,
			  DOFVector<double>* signedDistDOF,
			  std::map<DegreeOfFreedom, double> &indices)
  {
   
    const BasisFunction *vecBasisFcts = feSpace->getBasisFcts();
    int nVecBasisFcts = vecBasisFcts->getNumber();

    std::vector<DegreeOfFreedom> vecLocalIndices(nVecBasisFcts);
    ElementVector signedDistLocalCoeffs(nVecBasisFcts);

    DimVec<double> *baryCoords = NULL;
    DOFVector<short int> visited(feSpace, "visited"); visited.set(0);
    TraverseStack stack;
    Flag traverseFlag = Mesh::CALL_LEAF_EL;
    ElInfo *elInfo = stack.traverseFirst(feSpace->getMesh(), -1, traverseFlag);

    while (elInfo) {
      Element *el = elInfo->getElement();

      vecBasisFcts->getLocalIndices(el, feSpace->getAdmin(), vecLocalIndices);
      signedDistDOF->getLocalVector(el, signedDistLocalCoeffs);

      for (int i = 0; i < nVecBasisFcts; i++) {
	if ( visited[vecLocalIndices[i]] == 0 ) {
	  baryCoords = vecBasisFcts->getCoords(i);

	  double dist = vecBasisFcts->evalUh(*baryCoords, signedDistLocalCoeffs);
	  if (dist <= DBL_TOL) {
	    indices[vecLocalIndices[i]] = 0.0;
	  }
	  visited[vecLocalIndices[i]] = 1;
	}
      }
      elInfo = stack.traverseNext(elInfo); 
    }
  }
  
  inline void getPeriodicAssociation(const FiniteElemSpace* feSpace,
				     std::map<DegreeOfFreedom, double> indices,
				     AbstractFunction<WorldVector<double>, WorldVector<double> >* periodicMap,
				     std::vector<std::pair<DegreeOfFreedom, DegreeOfFreedom> > &association)
  {
    association.clear();

    DOFVector<WorldVector<double> > coords(feSpace, "coords");
    feSpace->getMesh()->getDofIndexCoords(coords);
    std::map<DegreeOfFreedom, double>::iterator it;
    
    DegreeOfFreedom idx_;
    WorldVector<double> p;
    for (it = indices.begin(); it != indices.end(); it++)
    {
      p = (*periodicMap)(coords[it->first]);
      TEST_EXIT(coords.getDofIdxAtPoint(p,idx_))("periodic association not found!\n");
      association.push_back(std::make_pair(std::min(it->first, idx_), std::max(it->first, idx_)));
    }
    
#if DEBUG != 0
    MSG("Nr of associations: %d\n", association.size());
#endif
  }

  
} } // end namespace

// #include "SingularDirichletBC2.hh"

namespace AMDiS {
  
struct DirichletBcDataBase
{
  DirichletBcDataBase(int r, int c) : row(r), col(c) {}
  virtual void addToList(const FiniteElemSpace *feSpace, std::vector<SingularDirichletBC> &list) = 0;

  int row,col;
};

/**
 * data structure that holds all possible variation of description of coords and value
 * of (singular/implicit) dirichlet boundary condition.
 **/
template<typename PosType, typename ValueType>
struct DirichletBcData : DirichletBcDataBase {
  
  DirichletBcData(int i_, int j_, PosType &pos_, const ValueType &val_)
  : DirichletBcDataBase(i_, j_), boundary_nr(0), pos(pos_), val(val_) {}
    
  DirichletBcData(int i_, int j_, BoundaryType nr_, PosType &meshIndicator_, const ValueType &val_)
  : DirichletBcDataBase(i_, j_), boundary_nr(nr_), pos(meshIndicator_), val(val_) {}
    
  void addToList(const FiniteElemSpace *feSpace, std::vector<SingularDirichletBC> &list) override
  {
    std::map<DegreeOfFreedom, double> indicesValues;
    
    details::getIndices(pos, boundary_nr, feSpace, indicesValues);
    details::getValues(val, feSpace, indicesValues);
    
    std::map<DegreeOfFreedom, double>::iterator it;
    for (it = indicesValues.begin(); it != indicesValues.end(); it++) {
      list.push_back(SingularDirichletBC(row, col, it->first, it->second));
    }
    
#if DEBUG != 0
    MSG("Dirichle BC at %d DOFs added.\n",indicesValues.size());
#endif
  }

private:
  BoundaryType boundary_nr;
  PosType& pos;
  const ValueType& val;
};


struct PeriodicBcData {

  // pos = WorldVector
  PeriodicBcData(int row_, BoundaryType nr_, AbstractFunction<bool, WorldVector<double> >* meshIndicator_,
		 AbstractFunction<WorldVector<double>, WorldVector<double> >* periodicMap_)
  : row(row_), nr(nr_), meshIndicator(meshIndicator_), periodicMap(periodicMap_) {}

  PeriodicBcData(int row_, AbstractFunction<bool, WorldVector<double> >* meshIndicator_,
		 AbstractFunction<WorldVector<double>, WorldVector<double> >* periodicMap_)
  : row(row_), nr(0), meshIndicator(meshIndicator_), periodicMap(periodicMap_) {}

  void addToList(const FiniteElemSpace *feSpace, std::vector<ManualPeriodicBC> &list)
  {
    std::map<DegreeOfFreedom, double> indices;
    details::getBoundaryIndices(feSpace, nr, meshIndicator, indices);
    
    std::vector<std::pair<DegreeOfFreedom, DegreeOfFreedom> > associations;
    details::getPeriodicAssociation(feSpace, indices, periodicMap, associations);
    list.push_back(ManualPeriodicBC(row, associations));
  }
  int row;

protected:


  BoundaryType nr;
  AbstractFunction<bool, WorldVector<double> >* meshIndicator;
  AbstractFunction<WorldVector<double>, WorldVector<double> >* periodicMap;
};

} // end namespace 

#endif // EXTENSIONS_SINGULAR_DIRICHLET_BC2_H
