/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/


/** \file VtuReader.h */

#ifndef AMDIS_VTUREADER_H
#define AMDIS_VTUREADER_H

#include <cstring>
#include "DOFVector.h"
#include "Mesh.h"
#include "boost/filesystem.hpp"
#include "pugixml.hpp"
#include "kdtree_nanoflann.h"
#include "VectorOperations.h"

namespace AMDiS {

  namespace VtuReader {
    
    /// Copies the values of a VTU-file to a DOF vector.
    template<typename T>
    inline void readValue(std::string filename,
			  Mesh *mesh,
			  DOFVector<T> *dofVector,
			  std::string componentName);

    /// Copies the values of a VTU-file to a vector of DOF vectors.
    template<typename T>
    inline void readValue(std::string filename,
			  Mesh *mesh,
			  std::vector<DOFVector<T>*> dofVectors,
			  std::vector<std::string> componentNames);


    namespace details {

      template<typename T>
      struct LessPairEps
      {
	LessPairEps() : eps(1.e-7) {}
	bool operator()(const std::pair<WorldVector<double>, T>& lhs, const std::pair<WorldVector<double>, T>& rhs) const
	{
	  return (lhs.first.getSize()==1 ? vertexi_compare1(lhs.first, rhs.first) :
		    (lhs.first.getSize()==2 ? vertexi_compare2(lhs.first, rhs.first) :
		      (lhs.first.getSize()==3 ? vertexi_compare3(lhs.first, rhs.first) : false)));
	}
	bool operator()(const std::pair<WorldVector<double>, T>& lhs, const WorldVector<double>& rhs) const
	{
	  return (lhs.first.getSize()==1 ? vertexi_compare1(lhs.first, rhs) :
		    (lhs.first.getSize()==2 ? vertexi_compare2(lhs.first, rhs) :
		      (lhs.first.getSize()==3 ? vertexi_compare3(lhs.first, rhs) : false)));
	}
	bool operator()(const WorldVector<double>& lhs, const std::pair<WorldVector<double>, T>& rhs) const
	{
	  return (lhs.getSize()==1 ? vertexi_compare1(lhs, rhs.first) :
		    (lhs.getSize()==2 ? vertexi_compare2(lhs, rhs.first) :
		      (lhs.getSize()==3 ? vertexi_compare3(lhs, rhs.first) : false)));
	}
	bool operator()(const WorldVector<double>& lhs, const WorldVector<double>& rhs) const
	{
	  return (lhs.getSize()==1 ? vertexi_compare1(lhs, rhs) :
		    (lhs.getSize()==2 ? vertexi_compare2(lhs, rhs) :
		      (lhs.getSize()==3 ? vertexi_compare3(lhs, rhs) : false)));
	}
      private:
	double eps;

	//returns true if vertex lhs is "smaller" than vertex rhs
	//----------------------------------------< vertexi_compare1 >
	bool vertexi_compare1(const WorldVector<double> lhs, const WorldVector<double> rhs) const {
	  return (lhs[0]-rhs[0]>eps);
	}

	//----------------------------------------< vertexi_compare2 >
	bool vertexi_compare2(const WorldVector<double> lhs, const WorldVector<double> rhs) const {
	  return (lhs[0]-rhs[0]>eps)
		|| (abs(lhs[0]-rhs[0])<=eps && lhs[1]-rhs[1]>eps);
	}

	//----------------------------------------< vertexi_compare3 >
	bool vertexi_compare3(const WorldVector<double> lhs, const WorldVector<double> rhs) const {
	  return (lhs[0]-rhs[0]>eps)
		|| (abs(lhs[0]-rhs[0])<=eps && lhs[1]-rhs[1]>eps)
		|| (abs(lhs[0]-rhs[0])<=eps && abs(lhs[1]-rhs[1])<=eps
		&& lhs[2]-rhs[2]>eps);
	}
      };

      inline void valueVector2type(std::vector<double> p, double& value)
      {
	if (p.size() == 0)
	  throw(std::runtime_error("Not enough data for assignment!\n"));
	value = p[0];
      }

      inline void valueVector2type(std::vector<double> p, WorldVector<double>& value)
      {
	if (static_cast<int>(p.size()) != Global::getGeo(WORLD))
	  throw(std::runtime_error("Not enough data for assignment!\n"));
	for (int i = 0; i < Global::getGeo(WORLD); i++)
	  value[i] = p[i];
      }

      inline void valueVector2type(std::vector<double> p, std::vector<double>& value)
      {
	if (p.size() == 0)
	  throw(std::runtime_error("Not enough data for assignment!\n"));
	for (size_t i = 0; i < p.size(); i++)
	  value.push_back(p[i]);
      }

      inline void string2pointList(std::string& input, std::vector<WorldVector<double> >& pointList)
      {
	int dow = Global::getGeo(WORLD);
	int dowExplicit = 3;

	std::vector<std::string> valueStringList;
	std::stringstream ss(input);
	std::string buf;
	while (ss >> buf)
	  valueStringList.push_back(buf);

	pointList.clear();

	std::vector<std::string>::iterator it;
	size_t j = 0;
	for (it = valueStringList.begin(); it != valueStringList.end();j++) {
	  WorldVector<double> p;
	  for (int i = 0; i < dowExplicit && it != valueStringList.end(); i++, it++)
	    if (i < dow)
	      p[i] = boost::lexical_cast<double>(*it);
	  pointList.push_back(p);
	}
      }

      template<typename T>
      inline void string2valueList(std::string& input, std::vector<T>& valueList, int numComponent = 1, int numComponentMax = -1)
      {
	if (numComponentMax < 0)
	  numComponentMax = numComponent;

	std::vector<std::string> valueStringList;
	std::stringstream ss(input);
	std::string buf;
	while (ss >> buf)
	  valueStringList.push_back(buf);

	valueList.clear();

	std::vector<std::string>::iterator it;
	for (it = valueStringList.begin(); it != valueStringList.end();) {
	  std::vector<double> p;
	  for (int i = 0; i < numComponentMax && it != valueStringList.end(); i++, it++)
	    if (i < numComponent)
	      p.push_back(boost::lexical_cast<double>(*it));
	  T value; valueVector2type(p, value);
	  valueList.push_back(value);
	}
      }

      template<typename T>
      inline void string2dataMap(std::string& strPoints, std::string& strValues, std::map<WorldVector<double>, T, LessPairEps<T> >& dataList, int numComponent = 1, int numComponentMax = -1)
      {
	int dow = Global::getGeo(WORLD);
	int dowExplicit = 3;
	if (numComponentMax < 0)
	  numComponentMax = numComponent;
	std::string buf;

	std::vector<std::string> pointStringList;
	std::stringstream ss1(strPoints);
	while (ss1 >> buf)
	  pointStringList.push_back(buf);

	std::vector<std::string> valueStringList;
	std::stringstream ss2(strValues);
	while (ss2 >> buf)
	  valueStringList.push_back(buf);

	if (pointStringList.size()/dowExplicit != valueStringList.size()/numComponentMax)
	  throw(std::runtime_error("Coordinates and values do not match! nPoints="
				    + boost::lexical_cast<std::string>(pointStringList.size()/dowExplicit)
				    + ", nValues="
				    + boost::lexical_cast<std::string>(valueStringList.size()/numComponentMax)));

	std::vector<std::string>::iterator it1, it2;
	it1 = pointStringList.begin();
	it2 = valueStringList.begin();
	for (; it1 != pointStringList.end() && it2 != valueStringList.end();) {
	  WorldVector<double> p;
	  for (int i = 0; i < dowExplicit && it1 != pointStringList.end(); i++, it1++)
	    if (i < dow)
	      p[i] = boost::lexical_cast<double>(*it1);

	  std::vector<double> v;
	  for (int i = 0; i < numComponentMax && it2 != valueStringList.end(); i++, it2++)
	    if (i < numComponent)
	      v.push_back(boost::lexical_cast<double>(*it2));
	  T value; valueVector2type(v, value);

	  dataList[p] = value;
	}
      }

      // find point in mesh using KD-tree structure
      inline size_t getNearestIndex(experimental::KD_Tree& tree, WorldVector<double>& x)
      {
	// do a knn search
	const size_t num_nnp = 1;
	std::vector<size_t>   ret_indexes(num_nnp);
	std::vector<double> out_dists_sqr(num_nnp);

	nanoflann::KNNResultSet<double> resultSet(num_nnp);
	resultSet.init(&ret_indexes[0], &out_dists_sqr[0] );

	tree.index->findNeighbors(resultSet, x.begin(), nanoflann::SearchParams(10));
	return ret_indexes[0];
      }

    } // end namespace details
    
  } // end namespace VtuReader
  
} // end namespace AMDiS

#include "VtuReader.hh"

#endif // AMDIS_VTUREADER_H
