/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/
#include "CahnHilliardNavierStokes_TwoPhase.h"
#include "Views.h"
#include "SignedDistFunctors.h"
#include "PhaseFieldConvert.h"
#include "POperators.h"

using namespace AMDiS;

CahnHilliardNavierStokes_TwoPhase::CahnHilliardNavierStokes_TwoPhase(const std::string &name_) :
  super(name_),
  useMobility(false),
  useConservationForm(false),
  doubleWell(1),
  gamma(1.0),
  eps(0.1),
  minusEps(-0.1),
  epsInv(10.0),
  minusEpsInv(-10.0),
  epsSqr(0.01),
  minusEpsSqr(-0.01),
  laplaceType(0),
  nonLinTerm(3),
  oldTimestep(0.0),
  viscosity1(1.0),
  viscosity2(1.0),
  density1(1.0),
  density2(1.0),
  c(0.0),
  sigma(1.0),
  surfaceTension(1.0),
  reinit(NULL),
  velocity(NULL),
  fileWriter(NULL)
{
  // parameters for CH
  Parameters::get(name + "->use mobility", useMobility); // dt(c) = div(M(c)*grad(c)) + ...
  Parameters::get(name + "->gamma", gamma); // M(c) = gamma * Double-Well
  Parameters::get(name + "->epsilon", eps); // interface width
  
  // type of double well: 0= [0,1], 1= [-1,1]
  Parameters::get(name + "->double-well type", doubleWell); 

  // parameters for navier-stokes
  Initfile::get(name + "->viscosity1", viscosity1);
  Initfile::get(name + "->viscosity2", viscosity2);
  Initfile::get(name + "->density1", density1);
  Initfile::get(name + "->density2", density2);
  
  // type of laplace operator: 	0... div(nu*grad(u)), 
  //							1... div(0.5*nu*(grad(u)+grad(u)^T))
  Initfile::get(name + "->laplace operator", laplaceType); 
  
  // type of non-linear term: 	0... u^old*grad(u_i^old), 
  //							1... u*grad(u_i^old), 
  //							2... u^old*grad(u_i)
  //							3... u*grad(u_i^old) + u^old*grad(u_i) - u^old*grad(u_i^old)
  Initfile::get(name + "->non-linear term", nonLinTerm); 
  
  // Parameters for CH-Coupling
  // 0... u*grad(c), 1... -div(u*c)
  Initfile::get(name + "->use conservation form", useConservationForm);

  // Parameters for NS-Coupling
  // cahn-hiliard-force: sigma*mu*grad(c)
  Initfile::get(name + "->sigma", sigma); // surface tension
  surfaceTension = sigma*3.0/(2.0*sqrt(2.0)) / eps;

  MSG("surface tension: %e\n", surfaceTension);
    
  force.set(0.0);
  Initfile::get(name + "->force", force);
  
  // transformation of the parameters
  minusEps = -eps;
  epsInv = 1.0/eps;
  minusEpsInv = -epsInv;
  epsSqr = sqr(eps);
  minusEpsSqr = -epsSqr;
}


CahnHilliardNavierStokes_TwoPhase::~CahnHilliardNavierStokes_TwoPhase() 
{ FUNCNAME("CahnHilliardNavierStokes_TwoPhase::~CahnHilliardNavierStokes_TwoPhase()");

  if (reinit != NULL) {
    delete reinit;
    reinit = NULL;
  }
  if (velocity != NULL) {
    delete velocity;
    velocity = NULL;
  }
  delete fileWriter;
  fileWriter = NULL;
}


void CahnHilliardNavierStokes_TwoPhase::initData()
{ FUNCNAME("CahnHilliardNavierStokes_TwoPhase::initData()");

  dim = getMesh()->getDim();
  // create instance redistancing class
  reinit = new HL_SignedDistTraverse("reinit", dim);
  
  if (velocity == NULL)
    velocity = new DOFVector<WorldVector<double> >(getFeSpace(2), "velocity");
  fileWriter = new FileVectorWriter(name + "->velocity->output", getFeSpace(2)->getMesh(), velocity);

  super::initData();
}


void CahnHilliardNavierStokes_TwoPhase::transferInitialSolution(AdaptInfo *adaptInfo)
{ FUNCNAME("CahnHilliardNavierStokes_TwoPhase::transferInitialSolution()");

  calcVelocity();
  
  for (int i = 0; i < 2+dow; i++)
    prob->setExactSolution(prob->getSolution()->getDOFVector(i), i);

  fileWriter->writeFiles(adaptInfo, false);
  writeFiles(adaptInfo, false);

  // initial parameters for detecting mesh changes
  oldMeshChangeIdx= getMesh()->getChangeIndex();
}


void CahnHilliardNavierStokes_TwoPhase::fillOperators()
{ FUNCNAME("CahnHilliardNavierStokes_TwoPhase::fillOperators()");
  MSG("CahnHilliardNavierStokes_TwoPhase::fillOperators()\n");
  
  // variable order:
  // (c, mu, u0, u1 [, u2], p)
  
  WorldVector<DOFVector<double>* > vel;
  for (size_t i = 0; i < dow; i++)
    vel[i] = prob->getSolution()->getDOFVector(2+i);


  // dt(c) = laplace(mu) - u*grad(c)
  // -----------------------------------
  /// < phi*c/tau , psi >
  Operator *opChMnew = new Operator(prob->getFeSpace(0),prob->getFeSpace(0));
  opChMnew->addZeroOrderTerm(new Simple_ZOT);
  prob->addMatrixOperator(*opChMnew, 0, 0, getInvTau()); 
  
  /// < phi*c^old/tau , psi >
  Operator *opChMold = new Operator(prob->getFeSpace(0),prob->getFeSpace(0));
  opChMold->addZeroOrderTerm(new VecAtQP_ZOT(prob->getSolution()->getDOFVector(0)));
  prob->addVectorOperator(*opChMold,0, getInvTau());   
  
  /// < phi*grad(mu) , grad(psi) >
  // div(M(c)grad(mu)), with M(c)=gamma/4*(c^2-1)^2
  Operator *opChLM = new Operator(prob->getFeSpace(0),prob->getFeSpace(1));
  if (useMobility) {
    if (doubleWell == 0)
      opChLM->addSecondOrderTerm(new VecAtQP_SOT(
	prob->getSolution()->getDOFVector(0),
	new MobilityCH0(gamma)));
    else
      opChLM->addSecondOrderTerm(new VecAtQP_SOT(
	prob->getSolution()->getDOFVector(0),
	new MobilityCH1(gamma)));
  } else
    opChLM->addSecondOrderTerm(new Simple_SOT(gamma));
  prob->addMatrixOperator(*opChLM, 0, 1);
  
  /// < u_old * grad(c) - u_old * grad(c_old), psi >
  Operator *uGradC = new Operator(prob->getFeSpace(0), prob->getFeSpace(0));
  if (useConservationForm)
    uGradC->addTerm(new WorldVec_FOT(vel, -1.0), GRD_PSI);
  else
    uGradC->addTerm(new WorldVec_FOT(vel), GRD_PHI);
  prob->addMatrixOperator(*uGradC, 0, 0);
  uGradC->setUhOld(prob->getSolution()->getDOFVector(0));
  prob->addVectorOperator(*uGradC, 0);
  
  /// < u * grad(c_old) , psi >
  for (size_t i = 0; i < dow; i++) {
    Operator *uGradC2 = new Operator(prob->getFeSpace(0), prob->getFeSpace(0));
    if (useConservationForm)
      uGradC2->addTerm(new VecAndPartialDerivative_FOT(prob->getSolution()->getDOFVector(0), i, -1.0), GRD_PSI);
    else {
      uGradC2->addTerm(new PartialDerivative_ZOT(prob->getSolution()->getDOFVector(0), i));
    }
    prob->addMatrixOperator(*uGradC2, 0, 2+i);
  }
  

  // mu + eps^2*laplace(c) + c - 3*(c_old^2)*c = -2*c_old^3 [+ BC]
  // ----------------------------------------------------------------------
  /// < -3*phi*c*c_old^2 , psi >
  // -3*c_old^2 * c
  Operator *opChWImpl = new Operator(prob->getFeSpace(1),prob->getFeSpace(0));
  opChWImpl->addZeroOrderTerm(new VecAtQP_ZOT(
    prob->getSolution()->getDOFVector(0),
    (doubleWell == 0 ? static_cast<AbstractFunction<double, double>*>(new DoubleWell0Diff(-1.0))
		     : static_cast<AbstractFunction<double, double>*>(new DoubleWell1Diff(-1.0)))));
  prob->addMatrixOperator(*opChWImpl, 1, 0);  
  
  /// < -2*phi*c_old^3 , psi >
  // -2*c_old^3 + 3/2*c_old^2
  Operator *opChWExpl = new Operator(prob->getFeSpace(1),prob->getFeSpace(0));
  opChWExpl->addZeroOrderTerm(new VecAtQP_ZOT(
    prob->getSolution()->getDOFVector(0),
    new Pow3Functor(-2.0)));
  if (doubleWell == 0) {
    opChWExpl->addZeroOrderTerm(new VecAtQP_ZOT(
      prob->getSolution()->getDOFVector(0),
      new Pow2Functor(3.0/2.0)));
  }
  prob->addVectorOperator(*opChWExpl, 1);            
  
  /// < -eps^2*phi*grad(c) , grad(psi) >
  Operator *opChL = new Operator(prob->getFeSpace(1),prob->getFeSpace(0));
  opChL->addSecondOrderTerm(new Simple_SOT);
  prob->addMatrixOperator(*opChL, 1, 0, &minusEpsSqr);  
  
  /// < phi*mu , psi >
  Operator *opChM = new Operator(prob->getFeSpace(0),prob->getFeSpace(0));
  opChM->addZeroOrderTerm(new Simple_ZOT);
  prob->addMatrixOperator(*opChM, 1, 1);              

  // -----------------------------------------------------------------------------------
  for (size_t i = 0; i < dow; ++i) {
    if (abs(density1 - density2) > FLT_TOL)
      addMaterialDerivativeNonConst(i);
    else
      addMaterialDerivativeConst(i);

    /// Diffusion-Operator (including Stress-Tensor for space-dependent viscosity)
    if (abs(viscosity1 - viscosity2) > FLT_TOL)
      addLaplaceTermNonConst(i);
    else
      addLaplaceTermConst(i);
  
    /// < p , d_i(psi) >
    Operator *opGradP = new Operator(getFeSpace(2+i),getFeSpace(2+dow));
    opGradP->addTerm(new PartialDerivative_FOT(i, -1.0), GRD_PSI);
    prob->addMatrixOperator(*opGradP, 2+i, 2+dow);
  
    /// external force, i.e. gravitational force
    if (norm(force) > DBL_TOL) {
      Operator *opForce = new Operator(getFeSpace(2+i), getFeSpace(2+i));
      opForce->addTerm(new Simple_ZOT(force[i]));
      prob->addVectorOperator(*opForce, 2+i);
    }
    
    // forces by Cahn-Hilliard terms
    // -----------------------------
    /// < mu_old * grad(c) , psi >
    Operator *opMuGradC = new Operator(getFeSpace(2+i),getFeSpace(0));
    opMuGradC->addTerm(new VecAndPartialDerivative_FOT(
      prob->getSolution()->getDOFVector(1), i, -1.0), GRD_PHI);
    prob->addMatrixOperator(*opMuGradC, 2+i, 0, &surfaceTension, &surfaceTension);
    
	/// < mu * grad(c_old) , psi >
    Operator *opMuGradC2 = new Operator(getFeSpace(2+i),getFeSpace(1));
    opMuGradC2->addTerm(new PartialDerivative_ZOT(
      prob->getSolution()->getDOFVector(0), i, -1.0));
    prob->addMatrixOperator(*opMuGradC2, 2+i, 1, &surfaceTension, &surfaceTension);
	
	/// < mu_old * grad(c_old) , psi >
    opMuGradC2->setUhOld(prob->getSolution()->getDOFVector(1));
    prob->addVectorOperator(*opMuGradC2, 2+i, &surfaceTension, &surfaceTension);
  }

  /// div(u) = 0
  for (size_t i = 0; i < dow; ++i) {
    /// < d_i(u'_i) , psi >
    Operator *opDivU = new Operator(getFeSpace(2+dow),getFeSpace(2+i));
    opDivU->addTerm(new PartialDerivative_FOT(i), GRD_PHI);
    prob->addMatrixOperator(*opDivU, 2+dow, 2+i);
  }
}


void CahnHilliardNavierStokes_TwoPhase::addMaterialDerivativeConst(int i)
{  
  WorldVector<DOFVector<double>* > vel;
  for (size_t j = 0; j < dow; j++)
    vel[j] = prob->getSolution()->getDOFVector(2+j);

  /// < (1/tau)*u'_i , psi >
  Operator *opTime = new Operator(getFeSpace(2+i), getFeSpace(2+i));
  opTime->addTerm(new Simple_ZOT(density1));
  prob->addMatrixOperator(*opTime, 2+i, 2+i, getInvTau(), getInvTau());
  /// < (1/tau)*u_i^old , psi >
  Operator *opTimeOld = new Operator(getFeSpace(2+i), getFeSpace(2+i));
  opTimeOld->addTerm(new Phase_ZOT(prob->getSolution()->getDOFVector(2+i), density1));
  prob->addVectorOperator(*opTimeOld, 2+i, getInvTau(), getInvTau());

  /// < u^old*grad(u_i^old) , psi >
  if (nonLinTerm == 0 || nonLinTerm == 3) {
	Operator *opUGradU0 = new Operator(getFeSpace(2+i), getFeSpace(2+i));
	opUGradU0->addTerm(new WorldVec_FOT(vel, (nonLinTerm == 0 ? -1.0 : 1.0)*density1), 
						GRD_PHI);
	opUGradU0->setUhOld(prob->getSolution()->getDOFVector(2+i));
    prob->addVectorOperator(*opUGradU0, 2+i);
  }
  if (nonLinTerm == 1 || nonLinTerm == 3) {
    /// < u*grad(u_i^old) , psi >
    for (size_t j = 0; j < dow; ++j) {
      Operator *opUGradU1 = new Operator(getFeSpace(2+i),getFeSpace(2+i));
      opUGradU1->addTerm(new PartialDerivative_ZOT(
	prob->getSolution()->getDOFVector(2+i), j, density1));
      prob->addMatrixOperator(*opUGradU1, 2+i, 2+j);
    }
  } 
  if (nonLinTerm == 2 || nonLinTerm == 3) {
    /// < u^old*grad(u_i) , psi >
    for(size_t j = 0; j < dow; ++j) {
      Operator *opUGradU2 = new Operator(getFeSpace(2+i),getFeSpace(2+i));
      opUGradU2->addTerm(new VecAndPartialDerivative_FOT(
	prob->getSolution()->getDOFVector(2+j), j, density1), GRD_PHI);
      prob->addMatrixOperator(*opUGradU2, 2+i, 2+i);
    }
  }
}


void CahnHilliardNavierStokes_TwoPhase::addMaterialDerivativeNonConst(int i)
{  
  WorldVector<DOFVector<double>* > vel;
  for (size_t j = 0; j < dow; j++)
    vel[j] = prob->getSolution()->getDOFVector(2+j);

  /// < (1/tau)*u'_i , psi >
  Operator *opTime = new Operator(getFeSpace(2+i), getFeSpace(2+i));
  opTime->addTerm(new VecAtQP_ZOT(
    prob->getSolution()->getDOFVector(0),
    new LinearInterpolation(density1, density2)));
  prob->addMatrixOperator(*opTime, 2+i, 2+i, getInvTau(), getInvTau());
  /// < (1/tau)*u_i^old , psi >
  opTime->setUhOld(prob->getSolution()->getDOFVector(2+i));
  prob->addVectorOperator(*opTime, 2+i, getInvTau(), getInvTau());
  
  /// < (rho(c^old) + c^old*rho'(c^old)) * u^old*grad(u_i^old) , psi >
  for(size_t j = 0; j < dow; ++j) {
    Operator *opCUGradU = new Operator(getFeSpace(2+i),getFeSpace(2+i));
    opCUGradU->addTerm(new Vec2AndPartialDerivative_FOT(
      prob->getSolution()->getDOFVector(0),
      prob->getSolution()->getDOFVector(2+j),
      j,
      new TaylorExpansionFactor(density1, density2)), GRD_PHI);
    opCUGradU->setUhOld(prob->getSolution()->getDOFVector(2+i));
    prob->addVectorOperator(*opCUGradU, 2+i);
  }
  
  /// < rho(c^old)*u*grad(u_i^old) , psi >
  for (size_t j = 0; j < dow; ++j) {
    Operator *opUGradU1 = new Operator(getFeSpace(2+i),getFeSpace(2+j));
    opUGradU1->addTerm(new VecAndPartialDerivative_ZOT(
      prob->getSolution()->getDOFVector(0),
      prob->getSolution()->getDOFVector(2+i),
      j,
      new LinearInterpolation(density1, density2)));
    prob->addMatrixOperator(*opUGradU1, 2+i, 2+j);
  }    
  
  /// < rho(c^old)*u^old*grad(u_i) , psi >
  for(size_t j = 0; j < dow; ++j) {
    Operator *opUGradU2 = new Operator(getFeSpace(2+i),getFeSpace(2+i));
    opUGradU2->addTerm(new Vec2AndPartialDerivative_FOT(
      prob->getSolution()->getDOFVector(0),
      prob->getSolution()->getDOFVector(2+j),
      j,
      new LinearInterpolationFactor(density1, density2)), GRD_PHI);
    prob->addMatrixOperator(*opUGradU2, 2+i, 2+i);
  }

  /// < c * rho'(c^old) * u^old*grad(u_i^old) , psi >
  for(size_t j = 0; j < dow; ++j) {
    Operator *opCUGradU2 = new Operator(getFeSpace(2+i),getFeSpace(0));
    opCUGradU2->addTerm(new Vec2AndPartialDerivative_ZOT(
    prob->getSolution()->getDOFVector(0),
    prob->getSolution()->getDOFVector(2+j),
    prob->getSolution()->getDOFVector(2+i),
    j,
    new LinearInterpolationDiffFactor(density1, density2)));
    prob->addMatrixOperator(*opCUGradU2, 2+i, 0);
  }
}


void CahnHilliardNavierStokes_TwoPhase::addLaplaceTermConst(int i)
{ FUNCNAME("CahnHilliardNavierStokes_TwoPhase::addLaplaceTerm()");

  /// < alpha*[grad(u)+grad(u)^t] , grad(psi) >
  if (laplaceType == 1) {
    for (size_t j = 0; j < dow; ++j) {
      Operator *opLaplaceUi1 = new Operator(getFeSpace(2+i), getFeSpace(2+j));
      opLaplaceUi1->addTerm(new MatrixIJ_SOT(1-i, 1-j, viscosity1));
      prob->addMatrixOperator(*opLaplaceUi1, 2+i, 2+j);
    }
  }
  
  /// < alpha*grad(u'_i) , grad(psi) >
  Operator *opLaplaceUi = new Operator(getFeSpace(2+i), getFeSpace(2+i));
    opLaplaceUi->addTerm(new Simple_SOT(viscosity1));
  prob->addMatrixOperator(*opLaplaceUi, 2+i, 2+i);
}


void CahnHilliardNavierStokes_TwoPhase::addLaplaceTermNonConst(int i)
{ FUNCNAME("CahnHilliardNavierStokes_TwoPhase::addLaplaceTerm()");

  /// < alpha*[grad(u)+grad(u)^t] , grad(psi) >
  if (laplaceType == 1) {
    for (size_t j = 0; j < dow; ++j) {
      Operator *opLaplaceUi1 = new Operator(getFeSpace(2+i), getFeSpace(2+j));
      opLaplaceUi1->addTerm(new VecAtQP_IJ_SOT(
		  prob->getSolution()->getDOFVector(0),
		  new LinearInterpolation(viscosity1, viscosity2),
		  1-i, 1-j));
      prob->addMatrixOperator(*opLaplaceUi1, 2+i, 2+j);
    }
  }
  
  /// < alpha*grad(u'_i) , grad(psi) >
  Operator *opLaplaceUi = new Operator(getFeSpace(2+i), getFeSpace(2+i));
    opLaplaceUi->addTerm(new VecAtQP_SOT(
	  prob->getSolution()->getDOFVector(0),
	  new LinearInterpolation(viscosity1, viscosity2)));
  prob->addMatrixOperator(*opLaplaceUi, 2+i, 2+i);
}


void CahnHilliardNavierStokes_TwoPhase::closeTimestep(AdaptInfo *adaptInfo)
{ FUNCNAME("CahnHilliardNavierStokes_TwoPhase::closeTimestep()");

  calcVelocity();
  fileWriter->writeFiles(adaptInfo, false);
  writeFiles(adaptInfo, false);
}
