/** \file CouplingBaseProblem.h */

#ifndef COUPLING_BASE_PROBLEM_H
#define COUPLING_BASE_PROBLEM_H

#include "AMDiS.h"

// coupling structures
#include "CouplingIterationInterface.h"
#include "CouplingTimeInterface.h"
#include "CouplingProblemStat.h"
#include "BaseProblem.h"
#include "ProblemInstat.h"

#include <boost/preprocessor/repetition/enum_params.hpp>
#include <boost/preprocessor/repetition/enum_params_with_a_default.hpp>
#include <boost/preprocessor/punctuation/comma_if.hpp>
#include <boost/preprocessor/repetition/repeat.hpp>
#include <boost/preprocessor/arithmetic/inc.hpp>

#include "boost/tuple/tuple.hpp"
#include "boost/static_assert.hpp"

#include "GenericLoops.h"

#define MAX_NUM_COUPLED_PROBLEMS 10

namespace AMDiS {

namespace detail {
  
  /// for generic loops this struct can be passed
  struct AddProblem {
    template<typename CouplingProblemStatType, typename BaseProblemType>
    static void call(CouplingProblemStatType& couplingProblemStat, BaseProblemType& baseProblem) {
      for (size_t j = 0; j < baseProblem.getNumProblems(); j++)
	couplingProblemStat.addProblem(baseProblem.getProblem(j));
    }
  };
    
  /// for generic loops this struct can be passed
  struct AddTimeInterface {
    template<typename CouplingProblemStatType, typename BaseProblemType>
    static void call(CouplingProblemStatType& couplingProblemStat, BaseProblemType& baseProblem) {
      couplingProblemStat.addTimeInterface(&baseProblem);
    }
  };
    
  /// for generic loops this struct can be passed
  struct AddIterationInterface {
    template<typename CouplingProblemStatType, typename BaseProblemType>
    static void call(CouplingProblemStatType& couplingProblemStat, BaseProblemType& baseProblem) {
      couplingProblemStat.addIterationInterface(&baseProblem);
    }
  };
    
  /// Functor for generic loops. Method initData() is called for each element in a sequence.
  struct InitData { template<typename B> static void call(B& b) { b.initData(); } };
    
  /// Functor for generic loops. Method initData() is called for each element in a sequence.
  struct FinalizeData { template<typename B> static void call(B& b) { b.finalizeData(); } };
    
  /// Functor for generic loops. Method fillOperators() is called for each element in a sequence.
  struct FillOperators { template<typename B> static void call(B& b) { b.fillOperators(); } };
    
  /// Functor for generic loops. Method fillBoundaryConditions() is called for each element in a sequence.
  struct FillBoundaryConditions { template<typename B> static void call(B& b) { b.fillBoundaryConditions(); } };
  
} // end namespace detail

/**
  * \ingroup Problem 
  *
  * \brief Structur to couple BaseProblems of variouse types
  */
template<class ProblemType=ProblemStat, 
	 BOOST_PP_ENUM_PARAMS_WITH_A_DEFAULT(MAX_NUM_COUPLED_PROBLEMS, class BaseProblemType, void) >
class CouplingBaseProblem : public CouplingIterationInterface,
			   public CouplingTimeInterface,
			   public detail::CouplingProblemStat<ProblemType>
{
public:
  typedef detail::CouplingProblemStat<ProblemType> CProblemStat;
  typedef typename boost::tuples::detail::tie_mapper
    < BOOST_PP_ENUM_PARAMS(MAX_NUM_COUPLED_PROBLEMS, BaseProblemType) >::type BaseProblemsTupleType;

  #define BASEPROBLEM_ARGS(z, n, text) BOOST_PP_COMMA_IF(n) _BaseProblemType ## n &prob ## n
  #define BASEPROBLEM_TYPES(z, n, text) class _BaseProblemType ## n
    
  // constructors created by C-Macros for each number of arguments, using BOOST_PP-Macros
  
  #define COUPLED_BASEPROBLEM_CONSTRUCTOR(z, n, text) \
      template<BOOST_PP_ENUM(BOOST_PP_INC(n), BASEPROBLEM_TYPES, BaseProblemType)> \
      CouplingBaseProblem(std::string name_, \
			 BOOST_PP_REPEAT_ ## z(BOOST_PP_INC(n), BASEPROBLEM_ARGS, nil) ) \
      : CProblemStat(name_),\
	baseProblems( BOOST_PP_ENUM_PARAMS_Z(z, BOOST_PP_INC(n), prob) ), name(name_) \
      { dow = Global::getGeo(WORLD); } \
  /****************************************************************************/
  
  #define BASEPROBLEM_ARGS2(z, n, text) BOOST_PP_COMMA_IF(n) _BaseProblemType ## n *prob ## n
  
  #define COUPLED_BASEPROBLEM_CONSTRUCTOR2(z, n, text) \
      template<BOOST_PP_ENUM(BOOST_PP_INC(n), BASEPROBLEM_TYPES, BaseProblemType)> \
      CouplingBaseProblem(std::string name_, \
			 BOOST_PP_REPEAT_ ## z(BOOST_PP_INC(n), BASEPROBLEM_ARGS2, nil) ) \
      : CProblemStat(name_),\
	baseProblems( BOOST_PP_ENUM_PARAMS_Z(z, BOOST_PP_INC(n), *prob) ), name(name_) \
      { dow = Global::getGeo(WORLD); } \
  /****************************************************************************/
  BOOST_PP_REPEAT(MAX_NUM_COUPLED_PROBLEMS, COUPLED_BASEPROBLEM_CONSTRUCTOR, nil)

  BOOST_PP_REPEAT(MAX_NUM_COUPLED_PROBLEMS, COUPLED_BASEPROBLEM_CONSTRUCTOR2, nil)
  
  ~CouplingBaseProblem() { }  
  
  void initialize(Flag initFlag,
		  ProblemStatSeq *adoptProblem = NULL,
		  Flag adoptFlag = INIT_NOTHING) override
  {  
    tools::FOR_EACH< detail::AddProblem >::loop2(*this, baseProblems);
    tools::FOR_EACH< detail::AddIterationInterface >::loop2(*this, baseProblems);
    tools::FOR_EACH< detail::AddTimeInterface >::loop2(*this, baseProblems);

    // initialize all ProblemStat
    CProblemStat::initialize(initFlag, adoptProblem, adoptFlag);
    
    // set mesh dimension
    dim = CProblemStat::getMesh(0)->getDim();
  }  
  
  virtual void initData() {}
  virtual void finalizeData() {}  
  
  virtual void initTimeInterface()
  {
    tools::FOR_EACH< detail::InitData >::loop(baseProblems);
    initData();
    
    tools::FOR_EACH< detail::FillOperators >::loop(baseProblems);
    fillCouplingOperators();
    
    tools::FOR_EACH< detail::FillBoundaryConditions >::loop(baseProblems);
    fillCouplingBoundaryConditions();
    
    tools::FOR_EACH< detail::FinalizeData >::loop(baseProblems);
    finalizeData();
  }
  
  
  virtual void fillCouplingOperators() {}
  virtual void fillCouplingBoundaryConditions() {}
  
  
  /// get the j-th solution-vector of the i-th problem
  template<int i>
  DOFVector<double> *getSolution(int j)
  { FUNCNAME("CouplingBaseProblem::getSolution<i>(j)");
    BOOST_STATIC_ASSERT_MSG(0 <= i && i < _LENGTH_<BaseProblemsTupleType>::value , "BaseProblem-index out of range");
    
    TEST_EXIT(0 <= j && j <= _GET_<i>(baseProblems).getNumComponents())("Indices out of range!\n");
    return _GET_<i>(baseProblems).getSolution()->getDOFVector(j);
  }
  
  
  /// pointer to the j-th feSpace of the i-th problem
  template<int i>
  inline const FiniteElemSpace* getFeSpace(int j=0) 
  { FUNCNAME("CouplingBaseProblem::getFeSpace<i>(j)");
    BOOST_STATIC_ASSERT_MSG(0 <= i && i < _LENGTH_<BaseProblemsTupleType>::value , "BaseProblem index out of range");
    
    TEST_EXIT(0 <= j && j <= _GET_<i>(baseProblems).getNumComponents())("Indices out of range!\n");
    return _GET_<i>(baseProblems).getFeSpace(j); 
  }
  
  std::string getName() { return name; }

protected:
  BaseProblemsTupleType baseProblems;

  unsigned dim;	// dimension of the meshes
  unsigned dow;	// dimension of the world
  
  std::string name;
};

} // end namespace AMDiS

#endif // COUPLING_BASE_PROBLEM_H
