/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/

#include "AMDiS.h"
#include "BaseProblem.h"
#include "POperators.h"
#include "ExtendedProblemStat.h"
#include "chns.h"
#include "Refinement.h"
#include "MeshFunction_Level.h"
#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
  #include "parallel/PetscSolverNSCH.h"
#endif



using namespace AMDiS;

class NavierStokesCahnHilliard : public BaseProblem<ProblemStat>
{
public: // definition of types

  typedef BaseProblem<ProblemStat> super;

public: // public methods

  NavierStokesCahnHilliard(const std::string &name_, bool createProblem = true);
  ~NavierStokesCahnHilliard();

  virtual void initData();
  
  /** initTimestep of super and
   * initialization of densityPhase and viscosityPhase
   **/
  virtual void initTimestep(AdaptInfo *adaptInfo);
  virtual void closeTimestep(AdaptInfo *adaptInfo);
  
  virtual void setTime(AdaptInfo* adaptInfo);
  
/*   * indicates the different phases. Will be initialized
   * in \ref initTimeInterface with const 1
   **/
  void setMultiPhase(DOFVector<double>* p) { multiPhase=p; }

  
  virtual void solveInitialProblem(AdaptInfo *adaptInfo);
  virtual void solveInitialProblem2(AdaptInfo *adaptInfo);
  virtual void transferInitialSolution(AdaptInfo* adaptInfo);

  double getEpsilon() { return eps; }
  int getDoubleWellType() { return doubleWell; }

protected: // protected methods

  virtual void fillOperators();
  virtual void addLaplaceTerm(int i);


protected: // protected variables
  ///viscosity of phase 1
  double viscosity1;
  
  ///viscosity of phase 2
  double viscosity2;
  
  ///density of phase 1
  double density1;
  
  ///density of phase 2
  double density2;
  
  double a, b, ab;
  
  double tau;

  /// phase dependent density
  DOFVector<double> *densityPhase;
  
  /// phase dependent viscosity
  DOFVector<double> *viscosityPhase;
  
  double delta;
  
  /// phase inticator
  DOFVector<double> *multiPhase;
  
  DOFVector<WorldVector<double> >* velocity;

  void calcVelocity()
  {
    if (dow == 1)
      transformDOF(prob->getSolution()->getDOFVector(0), velocity, new AMDiS::Vec1WorldVec<double>);
    else if (dow == 2)
      transformDOF(prob->getSolution()->getDOFVector(0), prob->getSolution()->getDOFVector(1), velocity, new AMDiS::Vec2WorldVec<double>);
    else if (dow == 3)
      transformDOF(prob->getSolution()->getDOFVector(0), prob->getSolution()->getDOFVector(1), prob->getSolution()->getDOFVector(2), velocity, new AMDiS::Vec3WorldVec<double>);
  }

  FileVectorWriter* fileWriter;
  
  bool useMobility;

  unsigned dow;		// dimension of the world
  unsigned dim;
  PhaseFieldRefinement* refFunction;
  RefinementLevelDOF *refinement;

  double sigma, minus1;		// coupling parameter to calculate the surface tension
  double surfaceTension;// := sigma/epsilon

  int doubleWell;

  double gamma;
  double eps;
  double minusEps;
  double epsInv;
  double minusEpsInv;
  double epsSqr;
  double minusEpsSqr;
  
  int nonLinTerm;
  double theta;
  double theta1;
  double minusTheta1;
  WorldVector<double> force;
};



class Force : public AbstractFunction<double, double>
{
public:
  Force(double c_, int degree) :
    AbstractFunction<double, double>(degree), c(c_) {};
    
  double operator()(const double &x) const {    
          return x*c;    
  }

private:
  double c;
};


class LinearInterpolation1 : public AbstractFunction<double, double>
{
public:

  LinearInterpolation1(double c1, double c2, int degree=1) :
    AbstractFunction<double, double>(degree)
     { a = (c1-c2)/2.0; b = (c1+c2)/2.0; cmin=std::min(c1,c2); cmax=std::max(c1,c2);}
    
  double operator()(const double &phase) const { 
    double result = b+a*phase;
    if (result<cmin) result = cmin; 
    if (result>cmax) result = cmax;
    return result;
  }

private:
  double a,b,cmin,cmax;
};

/** linear interpolation between two values (like density, viscosity)
 *  using a phase-field variable in [0,1]
 **/
class LinearInterpolation0 : public AbstractFunction<double, double>
{
public:

  LinearInterpolation0(double val1_, double val2_, int degree=1) :
    AbstractFunction<double, double>(degree),
    val1(val1_), val2(val2_) {}

  double operator()(const double &phase) const {    
    return phase*val1 + (1.0-phase)*val2;
  }

private:
  
  double val1;
  double val2;
};


