/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/
#include "NavierStokes_TH_MultiPhase.h"

using namespace AMDiS;

NavierStokes_TH_MultiPhase::NavierStokes_TH_MultiPhase(const std::string &name_, bool createProblem) :
  super(name_, createProblem),
  viscosity1(1.0),
  viscosity2(1.0),
  density1(1.0),
  density2(1.0),
  phaseFieldType(0),
  densityPhase(NULL),
  viscosityPhase(NULL),
  multiPhase(NULL)
{ FUNCNAME("NavierStokes_TH_MultiPhase::_constructor()");

  Initfile::get(name + "->viscosity1", viscosity1); // viscosity of fluid 1
  Initfile::get(name + "->viscosity2", viscosity2); // viscosity of fluid 2
  Initfile::get(name + "->density1", density1); // density of fluid 1
  Initfile::get(name + "->density2", density2); // density of fluid 2  

  phaseFieldType = 1;  
  Initfile::get(name + "->phase field type", phaseFieldType); // 0: [0,1], 1: [-1,1]

  if (viscosity1 <= 0.0 && viscosity > 0.0)
    viscosity1 = viscosity;
  
  TEST_EXIT(viscosity1 > 0.0)("Positive viscosity is necessary!\n");
}


NavierStokes_TH_MultiPhase::~NavierStokes_TH_MultiPhase()
{
  if (densityPhase) { 
    delete densityPhase; 
    densityPhase = NULL; 
  }
  if (viscosityPhase) { 
    delete viscosityPhase; 
    viscosityPhase = NULL; 
  }
}


void NavierStokes_TH_MultiPhase::initData()
{
  densityPhase = new DOFVector<double>(prob->getFeSpace(0), "densityPhase");
  viscosityPhase = new DOFVector<double>(prob->getFeSpace(0), "viscosityPhase");

  densityPhase->set(density1);
  viscosityPhase->set(viscosity1);

  super::initData();
}


void NavierStokes_TH_MultiPhase::initTimestep(AdaptInfo *adaptInfo)
{
  super::initTimestep(adaptInfo);
  if (multiPhase) {
    if (phaseFieldType == 0) {
      LinearInterpolation0 dLI(density1, density2);
      LinearInterpolation0 vLI(viscosity1, viscosity2);
      transformDOF(multiPhase, densityPhase, &dLI);
      transformDOF(multiPhase, viscosityPhase, &vLI);
    } else {
      LinearInterpolation1 dLI(density1, density2);
      LinearInterpolation1 vLI(viscosity1, viscosity2);
      transformDOF(multiPhase, densityPhase, &dLI);
      transformDOF(multiPhase, viscosityPhase, &vLI);
    }
  }
}


void NavierStokes_TH_MultiPhase::fillOperators()
{
  int degree_u = prob->getFeSpace(0)->getBasisFcts()->getDegree();
//   int degree_p = prob->getFeSpace(dow)->getBasisFcts()->getDegree();
  
  WorldVector<DOFVector<double>* > vel;
  for (unsigned i = 0; i < dow; i++){
    vel[i]= prob->getSolution()->getDOFVector(i);
  }
  
  // fill operators for prob
  for (size_t i = 0; i < dow; ++i) {
    /// < (1/tau)*rho*u'_i , psi >
    Operator *opTime = new Operator(getFeSpace(i), getFeSpace(i));
    if (density1 == density2 || !densityPhase)
      opTime->addTerm(new Simple_ZOT(density1));  
    else
      opTime->addTerm(new VecAtQP_ZOT(densityPhase));
    prob->addMatrixOperator(*opTime, i, i, getInvTau(), getInvTau());
    
    /// < (1/tau)*rho*u_i^old , psi >
    Operator *opTimeOld = new Operator(getFeSpace(i), getFeSpace(i));
    if (density1 == density2 || !densityPhase)
      opTimeOld->addTerm(new VecAtQP_ZOT(prob->getSolution()->getDOFVector(i), new AMDiS::Factor<double>(density1, degree_u)));  
    else
      opTimeOld->addTerm(new Vec2AtQP_ZOT(densityPhase, prob->getSolution()->getDOFVector(i)));
    prob->addVectorOperator(*opTimeOld, i, getInvTau(), getInvTau());
 
    /// < u^old*grad(u_i^old) , psi >
    Operator *opUGradU0 = new Operator(getFeSpace(i),getFeSpace(i));
    if (density1 == density2 || !densityPhase)
        opUGradU0->addTerm(new WorldVec_FOT(vel, density1), GRD_PHI);
    else
        opUGradU0->addTerm(new WorldVecPhase_FOT(densityPhase, vel, 1.0), GRD_PHI);
    opUGradU0->setUhOld(prob->getSolution()->getDOFVector(i));
    prob->addVectorOperator(*opUGradU0, i);


    /// < u'*grad(u_i^old) , psi >
    for (size_t j = 0; j < dow; ++j) {
      Operator *opUGradU1 = new Operator(getFeSpace(i),getFeSpace(j));
      if (density1 == density2 || !densityPhase)
	opUGradU1->addTerm(new PartialDerivative_ZOT(prob->getSolution()->getDOFVector(i), j, density1));
      else
	opUGradU1->addTerm(new VecAndPartialDerivative_ZOT(densityPhase, prob->getSolution()->getDOFVector(i), j));
      prob->addMatrixOperator(*opUGradU1, i, j);
    }
      
    /// < u^old*grad(u'_i) , psi >
    for (size_t j = 0; j < dow; ++j) {
      Operator *opUGradU2 = new Operator(getFeSpace(i),getFeSpace(i));
      if (density1 == density2 || !densityPhase)
	opUGradU2->addTerm(new VecAndPartialDerivative_FOT(prob->getSolution()->getDOFVector(j), j, density1), GRD_PHI);
      else
	opUGradU2->addTerm(new Vec2AndPartialDerivative_FOT(densityPhase, prob->getSolution()->getDOFVector(j), j), GRD_PHI);
      prob->addMatrixOperator(*opUGradU2, i, i);
    }

    /// Diffusion-Operator (including Stress-Tensor for space-dependent viscosity
    addLaplaceTerm(i);
  
    /// < p , d_i(psi) >
    Operator *opGradP = new Operator(getFeSpace(i),getFeSpace(dow));
    opGradP->addTerm(new PartialDerivative_FOT(i, -1.0), GRD_PSI);
    prob->addMatrixOperator(*opGradP, i, dow);
    
    /// external force, e.g. gravitational force
    if (std::abs(force[i]) > DBL_TOL) {
      Operator *opForce = new Operator(getFeSpace(i), getFeSpace(i));
      if (density1 == density2 || !densityPhase)
	opForce->addTerm(new Simple_ZOT(density1*force[i]));
      else
	opForce->addTerm(new VecAtQP_ZOT(densityPhase, NULL, force[i]));
      prob->addVectorOperator(*opForce, i);
    }
  }

  /// div(u) = 0
  for (size_t i = 0; i < dow; ++i) {
    /// < d_i(u'_i) , psi >
    Operator *opDivU = new Operator(getFeSpace(dow),getFeSpace(i));
    opDivU->addTerm(new PartialDerivative_FOT(i), GRD_PHI);
    prob->addMatrixOperator(*opDivU, dow, i);
  }
  
  Operator *opZero = new Operator(getFeSpace(dow),getFeSpace(dow));
  opZero->addTerm(new Simple_ZOT(0.0));
  prob->addMatrixOperator(*opZero, dow, dow);
}


void NavierStokes_TH_MultiPhase::addLaplaceTerm(int i)
{
  /// < alpha*[grad(u)+grad(u)^t] , grad(psi) >
  if (viscosity1 != viscosity2 && viscosityPhase) {
    for (unsigned j = 0; j < dow; ++j) {
      Operator *opLaplaceUi1 = new Operator(getFeSpace(i), getFeSpace(j));
      opLaplaceUi1->addTerm(new VecAtQP_IJ_SOT(viscosityPhase, NULL, j, i));
      prob->addMatrixOperator(*opLaplaceUi1, i, j);
    }
  }
  
  /// < alpha*grad(u'_i) , grad(psi) >
  Operator *opLaplaceUi = new Operator(getFeSpace(i), getFeSpace(i));
  if (viscosity1 == viscosity2 || !viscosityPhase) {
    opLaplaceUi->addTerm(new Simple_SOT(viscosity1));
  } else {
    opLaplaceUi->addTerm(new VecAtQP_SOT(viscosityPhase));
  }
  prob->addMatrixOperator(*opLaplaceUi, i, i);
}
