/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/
#include "NavierStokes_TaylorHood.h"

using namespace AMDiS;

NavierStokes_TaylorHood::NavierStokes_TaylorHood(const std::string &name_, bool createProblem) :
  super(name_, createProblem),
  velocity(NULL),
  laplaceType(0),
  nonLinTerm(2),
  oldTimestep(0.0),
  viscosity(1.0),
  density(1.0),
  minus1(-1.0),
  theta(0.5+1.e-2),
  theta1(1-theta),
  minusTheta1(-theta1),
  fileWriter(NULL)
{  
  // fluid viscosity
  Initfile::get(name + "->viscosity", viscosity);
  // fluid density
  Initfile::get(name + "->density", density);

  // type of laplace operator: 
  //   0... div(nu*grad(u)), 
  //   1... div(0.5*nu*(grad(u)+grad(u)^T))
  Initfile::get(name + "->laplace operator", laplaceType); 

  // type of non-linear term: 
  //   0... u^old*grad(u_i^old), 
  //   1... u*grad(u_i^old), 
  //   2... u^old*grad(u_i)
  //   3... (1)+(2)-(0)
  Initfile::get(name + "->non-linear term", nonLinTerm); 

  // theta-scheme parameter
  Initfile::get(name + "->theta", theta); 
  theta1 = 1.0-theta;
  minusTheta1 = -theta1;

  // volume force
  force.set(0.0);
  Initfile::get(name + "->force", force);

  oldSolution.resize(dow);
  for (size_t i = 0; i < dow; i++)
    oldSolution[i] = NULL;
}


NavierStokes_TaylorHood::~NavierStokes_TaylorHood() 
{ 
  if (velocity != NULL) {
    delete velocity;
    velocity = NULL;
  }
  
  for (size_t i = 0; i < dow; i++) {
    if (oldSolution[i] != NULL)
      delete oldSolution[i];
    oldSolution[i] = NULL;
  }
 
  if (fileWriter) { 
    delete fileWriter;
    fileWriter = NULL;
  }
}


void NavierStokes_TaylorHood::initData()
{ 
  if (velocity == NULL)
    velocity = new DOFVector<WorldVector<double> >(getFeSpace(0), "velocity");
  
  for (size_t i = 0; i < dow; i++)
    oldSolution[i] = new DOFVector<double>(getFeSpace(i), "old(v_"+Helpers::toString(i)+")");
  
  fileWriter = new FileVectorWriter(name + "->velocity->output", getFeSpace()->getMesh(), velocity);

  super::initData();
}


void NavierStokes_TaylorHood::transferInitialSolution(AdaptInfo *adaptInfo)
{ 
  calcVelocity();
  for (size_t i = 0; i < dow; i++)
    oldSolution[i]->copy(*prob->getSolution()->getDOFVector(i));
  
  super::transferInitialSolution(adaptInfo);
}


void NavierStokes_TaylorHood::fillOperators()
{ 
  WorldVector<DOFVector<double>* > vel;
  for (size_t k = 0; k < dow; k++)
    vel[k] = prob->getSolution()->getDOFVector(k);
  
  const FiniteElemSpace* feSpace_u = getFeSpace(0);
  const FiniteElemSpace* feSpace_p = getFeSpace(dow);
  
  // fill operators for prob
  for (size_t i = 0; i < dow; ++i) {
    /// < (1/tau)*u'_i , psi >
    Operator *opTime = new Operator(feSpace_u, feSpace_u);
    opTime->addTerm(new Simple_ZOT(density));
    prob->addMatrixOperator(*opTime, i, i, getInvTau(), getInvTau());
    
    /// < (1/tau)*u_i^old , psi >
    Operator *opTimeOld = new Operator(feSpace_u, feSpace_u);
    opTimeOld->addTerm(new Phase_ZOT(getOldSolution(i), density));
    prob->addVectorOperator(*opTimeOld, i, getInvTau(), getInvTau());

    /// < u^old*grad(u_i^old) , psi >
    if (nonLinTerm == 0 || nonLinTerm == 3) {
      Operator *opUGradU0 = new Operator(feSpace_u, feSpace_u);
      opUGradU0->addTerm(new WorldVec_FOT(vel, density), GRD_PHI);
      opUGradU0->setUhOld(prob->getSolution()->getDOFVector(i));
      if (nonLinTerm == 0)
        prob->addVectorOperator(*opUGradU0, i);
      else
        prob->addVectorOperator(*opUGradU0, i, &minus1, &minus1);
    }
    
    /// < u*grad(u_i^old) , psi >
    if (nonLinTerm == 1 || nonLinTerm == 3) {
      for (size_t k = 0; k < dow; ++k) {
        Operator *opUGradU1 = new Operator(feSpace_u, feSpace_u);
        opUGradU1->addTerm(new PartialDerivative_ZOT(
	  prob->getSolution()->getDOFVector(i), k, density));
        prob->addMatrixOperator(*opUGradU1, i, k);
      }
    }

    /// < u^old*grad(u_i) , psi >
    if (nonLinTerm == 2 || nonLinTerm == 3) {
      for (size_t k = 0; k < dow; ++k) {
        Operator *opUGradU2 = new Operator(feSpace_u, feSpace_u);
        opUGradU2->addTerm(new VecAndPartialDerivative_FOT(
	  vel[k], k, density), GRD_PHI);
        prob->addMatrixOperator(*opUGradU2, i, i);
      }
    }

    /// Diffusion-Operator
    addLaplaceTerm(i);
  
    /// < p , d_i(psi) >
    Operator *opGradP = new Operator(feSpace_u, feSpace_p);
    opGradP->addTerm(new PartialDerivative_FOT(i, -1.0), GRD_PSI);
    prob->addMatrixOperator(*opGradP, i, dow);
  
    /// external force, i.e. gravitational force
    if (std::abs(force[i]) > DBL_TOL) {
      Operator *opForce = new Operator(feSpace_u, feSpace_u);
      opForce->addTerm(new Simple_ZOT(density*force[i]));
      prob->addVectorOperator(*opForce, i);
    }
  }

  /// div(u) = 0
  for (size_t i = 0; i < dow; ++i) {
    /// < d_i(u'_i) , psi >
    Operator *opDivU = new Operator(feSpace_p, feSpace_u);
    opDivU->addTerm(new PartialDerivative_FOT(i), GRD_PHI);
    prob->addMatrixOperator(*opDivU, dow, i);
  }

  Operator *opNull = new Operator(feSpace_p, feSpace_p);
  opNull->addTerm(new Simple_ZOT(0.0));
  prob->addMatrixOperator(*opNull, dow, dow);
}


void NavierStokes_TaylorHood::addLaplaceTerm(int i)
{
  const FiniteElemSpace* feSpace_u = getFeSpace(0);
  
  /// < alpha*[grad(u)+grad(u)^t] , grad(psi) >
  if (laplaceType == 1) {
    for (size_t j = 0; j < dow; ++j) {
      Operator *opLaplaceUi1 = new Operator(feSpace_u, feSpace_u);
      opLaplaceUi1->addTerm(new MatrixIJ_SOT(j, i, viscosity));
      prob->addMatrixOperator(*opLaplaceUi1, i, j, &theta, &theta);

      if (std::abs(minusTheta1) > DBL_TOL) {
	opLaplaceUi1->setUhOld(prob->getSolution()->getDOFVector(j));
	prob->addVectorOperator(*opLaplaceUi1, i, &minusTheta1, &minusTheta1);
      }
    }
  }
  
  /// < alpha*grad(u'_i) , grad(psi) >
  Operator *opLaplaceUi = new Operator(feSpace_u, feSpace_u);
  opLaplaceUi->addTerm(new Simple_SOT(viscosity));
  prob->addMatrixOperator(*opLaplaceUi, i, i, &theta, &theta);

  if (std::abs(minusTheta1) > DBL_TOL) {
    opLaplaceUi->setUhOld(prob->getSolution()->getDOFVector(i));
    prob->addVectorOperator(*opLaplaceUi, i, &minusTheta1, &minusTheta1);
  }
}


void NavierStokes_TaylorHood::closeTimestep(AdaptInfo *adaptInfo)
{ 
  calcVelocity();
  for (size_t i = 0; i < dow; i++)
    oldSolution[i]->copy(*prob->getSolution()->getDOFVector(i));
  
  super::closeTimestep(adaptInfo);
}


void NavierStokes_TaylorHood::writeFiles(AdaptInfo *adaptInfo, bool force)
{ 
  super::writeFiles(adaptInfo, force);
  fileWriter->writeFiles(adaptInfo, false);
}

