/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/

#ifndef NAVIER_STOKES_TAYLOR_HOOD_H
#define NAVIER_STOKES_TAYLOR_HOOD_H

#include "AMDiS.h"
#include "BaseProblem.h"
#include "ExtendedProblemStat.h"


namespace detail 
{
  using namespace AMDiS;
  
  /** \ingroup NavierStokes_TaylorHood
  * \brief
  * Navier-Stokes problem, using Taylor Hood elements
  * and theta-scheme time-integration
  */
  
  template<typename ProblemStatType>
  class NavierStokes_TaylorHood : public BaseProblem<ProblemStatType>
  {
  public: // typedefs
    
    typedef NavierStokes_TaylorHood<ProblemStatType> self;
    typedef BaseProblem<ProblemStatType> super;
    
  public: // methods

    NavierStokes_TaylorHood(const std::string &name_, bool createProblem = true);

    ~NavierStokes_TaylorHood();

    void initData() override;

    void transferInitialSolution(AdaptInfo *adaptInfo) override;  

    void closeTimestep(AdaptInfo *adaptInfo) override;
    
    void writeFiles(AdaptInfo *adaptInfo, bool force = false) override;

    // === getting/setting methods ===

    DOFVector<WorldVector<double> >* getVelocity()
    { FUNCNAME_DBG("getVelocity()");
      TEST_EXIT_DBG(velocity != NULL)
	("velocity is NULL!");
      return velocity;
    }
    
    DOFVector<double> *getOldSolution(int i)
    { FUNCNAME_DBG("getOldSolution()");
      TEST_EXIT_DBG(oldSolution[i] != NULL)
	("Index out of range, or oldSolution not provided for this component!\n");
      return oldSolution[i];
    }
    
    double getDensity() { return density; }
    double getViscosity() { return viscosity; }

    void fillOperators() override;
    
    virtual void addLaplaceTerm(int i);

  protected: // variables

    DOFVector<WorldVector<double> >* velocity;

    void calcVelocity()
    {
      if (self::dow == 1)
	transformDOF(self::prob->getSolution()->getDOFVector(0), 
		    velocity, new AMDiS::Vec1WorldVec<double>);
      else if (self::dow == 2)
	transformDOF(self::prob->getSolution()->getDOFVector(0), 
		    self::prob->getSolution()->getDOFVector(1), 
		    velocity, new AMDiS::Vec2WorldVec<double>);
      else if (self::dow == 3)
	transformDOF(self::prob->getSolution()->getDOFVector(0), 
		    self::prob->getSolution()->getDOFVector(1), 
		    self::prob->getSolution()->getDOFVector(2), 
		    velocity, new AMDiS::Vec3WorldVec<double>);
    }

    int laplaceType;
    int nonLinTerm;

    double oldTimestep;
    double viscosity;
    double density;
    
    double minus1;
    double theta;
    double theta1;
    double minusTheta1;

    WorldVector<double> force;

    FileVectorWriter *fileWriter;
    std::vector<DOFVector<double>*> oldSolution;
  };

} // end namespace detail

#include "NavierStokes_TaylorHood.hh"

typedef ::detail::NavierStokes_TaylorHood<AMDiS::ProblemStat> NavierStokes_TaylorHood;



#endif // NAVIER_STOKES_TAYLOR_HOOD_H

