/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/
#include "PhaseFieldCrystal.h"

using namespace AMDiS;

PhaseFieldCrystal::PhaseFieldCrystal(const std::string &name_, bool createProblem) :
  super(name_, createProblem),
  useMobility(false),
  tempParameter(-0.6),
  r(-0.4),		// temperature deviation
  q(sqrt(3.0)/2.0),	// lattice constant
  density(-0.3),	// mean density
  two(2.0),
  minus2(-2.0)
{  
  Parameters::get(name + "->r",r);
  Parameters::get(name + "->density", density);
  Parameters::get(name + "->q", q);
  Parameters::get(name + "->use mobility", useMobility);
  tempParameter= -(1.0+r);
  
  q /= sqrt(3.0)/2.0;
  q2 = 1.0/sqr(q);
  twoQ2 = 2.0*q2;
}


void PhaseFieldCrystal::fillOperators()
{ 
  int degree = prob->getFeSpace(0)->getBasisFcts()->getDegree();
  
  // phi*rho
  Operator *opMnew = new Operator(prob->getFeSpace(0), prob->getFeSpace(0));
  ZeroOrderTerm *simple_zot = new Simple_ZOT; 
  operatorTermList.push_back(simple_zot);
  opMnew->addTerm(simple_zot);
  operatorList.push_back(opMnew);
  
  Operator *opMold = new Operator(prob->getFeSpace(0), prob->getFeSpace(0));
  ZeroOrderTerm *vecAtQP_zot = new VecAtQP_ZOT(prob->getSolution()->getDOFVector(0));
  operatorTermList.push_back(vecAtQP_zot);
  opMold->addTerm(vecAtQP_zot);
  operatorList.push_back(opMold);

  // -nabla*(phi*grad(rho))
  Operator *opL = new Operator(prob->getFeSpace(0), prob->getFeSpace(0));
  SecondOrderTerm *simple_sot = new Simple_SOT;
  operatorTermList.push_back(simple_sot);
  opL->addTerm(simple_sot);
  operatorList.push_back(opL);
  
  double M0 = 1.0;
  Parameters::get(name + "->M0", M0);
  Operator *opLM = new Operator(prob->getFeSpace(0), prob->getFeSpace(0));
  SecondOrderTerm *mobility_sot;
  if (useMobility) // non-constant mobility
    mobility_sot = new VecAtQP_SOT(
      prob->getSolution()->getDOFVector(0),
      new MobilityPfc(density, M0, degree));
  else // constant mobility
    mobility_sot = new Simple_SOT(M0);
  operatorTermList.push_back(mobility_sot);
  opLM->addTerm(mobility_sot);
  operatorList.push_back(opLM);
  
  // -2*rho_old^3
  Operator *opMPowExpl = new Operator(prob->getFeSpace(0), prob->getFeSpace(0));
  ZeroOrderTerm *pow3_zot = new VecAtQP_ZOT(prob->getSolution()->getDOFVector(0),new AMDiS::Pow<3>(-2.0, 3*degree));
  operatorTermList.push_back(pow3_zot);
  opMPowExpl->addTerm(pow3_zot);
  operatorList.push_back(opMPowExpl);
  
  // -3*rho_old^2
  Operator *opMPowImpl = new Operator(prob->getFeSpace(0), prob->getFeSpace(0));
  ZeroOrderTerm *pow2_zot = new VecAtQP_ZOT(prob->getSolution()->getDOFVector(0),new AMDiS::Pow<2>(-3.0, 2*degree));
  operatorTermList.push_back(pow2_zot);
  opMPowImpl->addTerm(pow2_zot);
  operatorList.push_back(opMPowImpl);
  
  // 0 = nu-laplace(rho)
  // -------------------
  prob->addMatrixOperator(opL, 0, 0, &q2, &q2); 					// -laplace(rho)
  prob->addMatrixOperator(opMnew, 0, 2); 					// nu
  
  // dt(rho) = laplace(mu) - u*grad(rho)
  // -----------------------------------
  prob->addMatrixOperator(opMnew, 1, 0, getInvTau(), getInvTau());
  prob->addMatrixOperator(opLM, 1, 1); 						// -laplace(mu)
  // . . . vectorOperators . . . . . . . . . . . . . . .
  prob->addVectorOperator(opMold, 1, getInvTau(), getInvTau());
  
  // mu-2*nu-laplace(nu)-(1+r)*rho = (rho_old^3) + ExtPot
  // ----------------------------------------------------------------------
  prob->addMatrixOperator(opMnew, 2, 0, &tempParameter, &tempParameter); 	// -phi*(1+r)*rho
  prob->addMatrixOperator(opMPowImpl, 2, 0); 					// -3*rho*rho_old^2
  prob->addMatrixOperator(opL, 2, 0, &twoQ2, &twoQ2); 				// -2*phi*laplace(rho) * psi
  prob->addMatrixOperator(opMnew, 2, 1); 					// phi*mu * psi
  prob->addMatrixOperator(opL, 2, 2, &q2, &q2); 					// phi*grad(nu) * grad(psi)
  // . . . vectorOperators . . . . . . . . . . . . . . .
  prob->addVectorOperator(opMPowExpl, 2); 					// -2*phi^old*rho_old^3
  
  // ===== ADD a zero-.operator ALL blocks ==============
  Operator *opZero = new Operator(prob->getFeSpace(0), prob->getFeSpace(0));
  ZeroOrderTerm *zero_zot = new Simple_ZOT(0.0); 
  operatorTermList.push_back(zero_zot);
  opZero->addTerm(zero_zot);
  operatorList.push_back(opZero);
  
  prob->addMatrixOperator(opZero, 0, 1);
  prob->addMatrixOperator(opZero, 1, 2);
}


void PhaseFieldCrystal::finalizeData()
{
  setAssembleMatrixOnlyOnce_butTimestepChange(0,0);
  setAssembleMatrixOnlyOnce_butTimestepChange(0,2);
  
  setAssembleMatrixOnlyOnce_butTimestepChange(1,0);
  if (!useMobility)
    setAssembleMatrixOnlyOnce_butTimestepChange(1,1);
    
  setAssembleMatrixOnlyOnce_butTimestepChange(2,1);
  setAssembleMatrixOnlyOnce_butTimestepChange(2,2);
}

  
