/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/

#ifndef PHASE_FIELD_CRYSTAL_PHASE_H
#define PHASE_FIELD_CRYSTAL_PHASE_H

#include "PhaseFieldCrystal.h"
#include "PhaseFieldConvert.h"
#include "POperators.h"

using namespace AMDiS;

/** Phase-field Crystal problem in diffuse domain
 */
class PhaseFieldCrystal_Phase : public PhaseFieldCrystal
{
public: // typedefs

  typedef PhaseFieldCrystal super;

public: // methods

  PhaseFieldCrystal_Phase(const std::string &name_);
  ~PhaseFieldCrystal_Phase();

  void initTimeInterface();

  DOFVector<double> *getPhase()
  {
    return phase;
  }

  void setPhase(DOFVector<double>* phase_)
  {
    if (phase == NULL)
      phase = new DOFVector<double>(prob->getFeSpace(0), "phase");
    if (phaseDisturbed == NULL)
      phaseDisturbed = new DOFVector<double>(prob->getFeSpace(0), "phaseDisturbed");

    phase->interpol(phase_);
    transformDOF(phase, 1.e-6, phaseDisturbed, new Max<double>);
  };

  void fillOperators();

private: // variables

  DOFVector<double> *phase;
  DOFVector<double> *phaseDisturbed;
};


///Abstract function to calculate the pure PFC-Energy
class EnergyPhase : public TertiaryAbstractFunction<double,double,double,double>
{
  public:
    EnergyPhase() : TertiaryAbstractFunction<double,double,double,double>(5) { }

    double operator()(const double &phase, const double &rho, const double &mu) const {
      return phase * (-0.25*sqr(sqr(rho)) + 0.5*rho*mu); }
};

/// AbstractFunction that defines the pfc-mobility, depending on the density an minimal density
class MobilityPhasePfc : public BinaryAbstractFunction<double,double,double>
{
  public:
    MobilityPhasePfc(double density_ = -0.3, double factor_ = 1.0, int degree=1) : 
      BinaryAbstractFunction<double,double,double>(2*degree + 1),
      density(density_),
      factor(factor_),
      delta(1.e-6) { }

    double operator()(const double &phase, const double &rho) const 
    {
      double mobility= abs(rho - 3.0*density)*factor;
      return std::max(phase*mobility, delta);
    }
  
  protected:
    double density;
    double factor;
    double delta;
};

#endif // PHASE_FIELD_CRYSTAL_PHASE_H
