/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/

#ifndef MOVING_MESH_H
#define MOVING_MESH_H

#include "BaseProblem.h"

// coupling structures
#include "CouplingIterationInterface.h"
#include "CouplingTimeInterface.h"
#include "CouplingProblemStat.h"

// structures for local refinement
// #include "Refinement.h"
// #include "MeshFunction_Level.h"

// #include "SignedDistFunctors.h"
// #include "PhaseFieldConvert.h"

using namespace AMDiS;


struct MinWrapper : AbstractFunction<double, WorldVector<double> >
{
  MinWrapper(AbstractFunction<double, WorldVector<double> >* dist1_, AbstractFunction<double, WorldVector<double> >* dist2_)
  : dist1(dist1_), dist2(dist2_) {}

  double operator()(const WorldVector<double>& x) const
  {
    return std::min((*dist1)(x), (*dist2)(x));
  }
private:
  AbstractFunction<double, WorldVector<double> >* dist1;
  AbstractFunction<double, WorldVector<double> >* dist2;
};


struct BeamDisplacement1d : AbstractFunction<double, double>
{
  BeamDisplacement1d(double P_, double L_, double EI_, double* time_ = NULL, double* tau_ = NULL) : P(P_), L(L_), EI(EI_), time(time_), tau(tau_) {}

  double operator()(const double& x) const
  {
    return (*time < DBL_TOL ? 0.0 : (sin(*time)-sin(*time - *tau))*P*sqr(x)*(3.0*L-x)/(6.0*EI));
  }

  void setTimePtr(double* time_)
  {
    time = time_;
  }

  void setTauPtr(double* tau_)
  {
    tau = tau_;
  }
  
private:
  double P;
  double L;
  double EI;
  double* time;
  double* tau;
};

struct BeamDisplacement : AbstractFunction<double, WorldVector<double> >
{
  BeamDisplacement(double P, double L, double EI, WorldVector<double> fixedPoint_, double* time = NULL, double* tau = NULL)
  : fct(new BeamDisplacement1d(P, L, EI, time, tau)), fixedPoint(fixedPoint_) {}
  
  double operator()(const WorldVector<double>& x) const
  {
    WorldVector<double> displacement; displacement[0] = 0.0;
    displacement[1] = (x[0] >= fixedPoint[0] ? (*fct)(x[0] - fixedPoint[0]) : 0.0);
    return displacement[1];
  }

  void setTimePtr(double* time_)
  {
    fct->setTimePtr(time_);
  }

  void setTauPtr(double* tau_)
  {
    fct->setTauPtr(tau_);
  }

private:
  BeamDisplacement1d* fct;
  WorldVector<double> fixedPoint;
};

/**
  * \ingroup Problem 
  *
  * \brief
  */
template<typename Elasticity_Type1 = StandardBaseProblem,
		 typename Elasticity_Type2 = StandardBaseProblem>
class MovingMesh : public CouplingIterationInterface,
                                 public CouplingTimeInterface,
                                 public CouplingProblemStat
{
public:

  MovingMesh(std::string name_, Elasticity_Type1 *elastProb_, Elasticity_Type2 *elastProb2_)
  : CouplingProblemStat(name_),
    elastProb(elastProb_),
    elastProb2(elastProb2_)
//     refFunction1(NULL),
//     refFunction2(NULL),
//     refinement1(NULL),
//     refinement2(NULL)
  {
    dow = Global::getGeo(WORLD);
  }


  ~MovingMesh() {
//     if (refFunction1 != NULL)
//       delete refFunction1;
//     if (refinement1 != NULL)
//       delete refinement1;
//     if (refFunction2 != NULL)
//       delete refFunction2;
//     if (refinement2 != NULL)
//       delete refinement2;

    for (int i = 0; i < Global::getGeo(WORLD); i++) {
      delete parametricCoords1[i];
      delete parametricCoords2[i];
	}

    delete parametric1;
    delete parametric2;
  }


  /**
   * Add the problems to the iterationInterface, timeInterface and couplingProblemStat.
   * As a consequence all problem can be initialized one after another and in the
   * adaption loop they are solved in rotation.
   * At the end the problems are filled with operators and coupling operators as well as
   * boundary conditions are added.
   *
   * In the adaption loop the problems are solved the same order as they are added to the
   * iterationInterface in this method. This order can be changed manually in the oneIteration
   * method.
   **/
  void initialize(AdaptInfo *adaptInfo)
  {
    for (size_t i = 0; i < elastProb->getNumProblems(); i++)
      addProblem(elastProb->getProblem(i));
    for (size_t i = 0; i < elastProb2->getNumProblems(); i++)
      addProblem(elastProb2->getProblem(i));

    addIterationInterface(elastProb);
    addIterationInterface(elastProb2);

    addTimeInterface(elastProb);
    addTimeInterface(elastProb2);

//     CouplingProblemStat::initialize(INIT_ALL | INIT_EXACT_SOLUTION);
	elastProb->initialize(INIT_ALL);
	elastProb2->initialize(INIT_ALL);
    dim = elastProb->getMesh()->getDim();

    initData();
    elastProb->initData();
    elastProb2->initData();
    fillCouplingBoundaryConditions();

    elastProb->fillOperators();
    elastProb2->fillOperators();

    elastProb->fillBoundaryConditions();
    elastProb2->fillBoundaryConditions();
  }

  
  void initData()
  { MSG("initData()\n");

    // ===== create coords vector =====
    for (int i = 0; i < Global::getGeo(WORLD); i++) {
      parametricCoords1[i] = new DOFVector<double>(elastProb->getFeSpace(0),
						  "parametric1 coords "+boost::lexical_cast<std::string>(i));
      parametricCoords2[i] = new DOFVector<double>(elastProb2->getFeSpace(0),
						  "parametric2 coords "+boost::lexical_cast<std::string>(i));
    }

    // ===== create parametric object =====
    parametric1 = new ParametricFirstOrder(&parametricCoords1);
    parametric2 = new ParametricFirstOrder(&parametricCoords2);
	
	double P = 0.2,L = 0.46,EI = 1.0;
	Parameters::get("beam->P", P);
	Parameters::get("beam->L", L);
	Parameters::get("beam->EI", EI);
	WorldVector<double> fixedPoint;
	Parameters::get("beam->fixedPoint", fixedPoint);
	beamDisplacement = new BeamDisplacement(P, L, EI, fixedPoint);
  }


  void fillCouplingBoundaryConditions()
  { MSG("fillCouplingBoundaryConditions()\n");
	elastProb->setDisplacement(beamDisplacement);
	elastProb2->setDisplacement(beamDisplacement);
  }


  /**
   * Solves the initial problems.Before this is done the mesh is refined by the Refinement-class
   * This provides local refinement functions for phaseField-refinement or signed-distance-refinement.
   * See MeshFunction_Level.h for some details.
   **/ 
  void solveInitialProblem(AdaptInfo *adaptInfo)
  { MSG("solveInitialProblem()\n");
  
	beamDisplacement->setTimePtr(adaptInfo->getTimePtr());
	beamDisplacement->setTauPtr(adaptInfo->getTimestepPtr());
  
//     size_t nVertices = 0;
//     WorldVector<double> c,c2;
//     Parameters::get("obstacle->num vertices",nVertices);
//     std::vector<WorldVector<double> > v(nVertices,c);
//     for (size_t i = 0; i < nVertices; i++)
//       Parameters::get("obstacle->vertex["+boost::lexical_cast<std::string>(i)+"]",v[i]);
//     v.push_back(v[0]);
//     
//     obstacle = new Polygon(v);
// 
//     Parameters::get("flag->num vertices",nVertices);
//     std::vector<WorldVector<double> > v2(nVertices,c);
//     for (size_t i = 0; i < nVertices; i++)
//       Parameters::get("flag->vertex["+boost::lexical_cast<std::string>(i)+"]",v2[i]);
//     v2.push_back(v2[0]);
//     
//     flag = new Polygon(v2);
// 
//     Parameters::get("connection->num vertices",nVertices);
//     std::vector<WorldVector<double> > v3(nVertices,c);
//     for (size_t i = 0; i < nVertices; i++)
//       Parameters::get("connection->vertex["+boost::lexical_cast<std::string>(i)+"]",v3[i]);
//     v3.push_back(v3[0]);
//     
//     signedDistFct = new MinWrapper(new Polygon(v2), new MinWrapper(obstacle,flag));
// 
//     refFunction1 = new SignedDistRefinement(elastProb->getMesh());
//     refinement1 = new RefinementLevelCoords2(
//       elastProb->getFeSpace(0),
//       refFunction1,
//       signedDistFct);
// 	
//     refFunction2 = new SignedDistRefinement(elastProb2->getMesh());
//     refinement2 = new RefinementLevelCoords2(
//       elastProb2->getFeSpace(0),
//       refFunction2,
//       signedDistFct);

    // initial refinement
//     refinement1->refine(10);
//     refinement2->refine(10);
//     flag->refine(10);

    // solve all initial problems of the problems added to the CouplingTimeInterface
    CouplingTimeInterface::solveInitialProblem(adaptInfo);
	
	DOFVector<WorldVector<double> > coords1(parametricCoords1[0]->getFeSpace(), "coords");
	parametricCoords1[0]->getFeSpace()->getMesh()->getDofIndexCoords(coords1);
	DOFVector<WorldVector<double> > coords2(parametricCoords2[0]->getFeSpace(), "coords");
	parametricCoords2[0]->getFeSpace()->getMesh()->getDofIndexCoords(coords2);
    for (int i = 0; i < Global::getGeo(WORLD); i++) {
	  transformDOF(&coords1, parametricCoords1[i], new AMDiS::Component2<>(i));
	  transformDOF(&coords2, parametricCoords2[i], new AMDiS::Component2<>(i));
	}
	
    // ===== enable parametric traverse =====
    elastProb->getMesh()->setParametric(parametric1);
    elastProb2->getMesh()->setParametric(parametric2);
  }


  /**
   * Called at the end of each timestep.
   * If the phase-field has changed the mesh is updated by the refinement-method
   **/
  void closeTimestep(AdaptInfo *adaptInfo)
  { MSG("closeTimestep()\n");
  
    CouplingTimeInterface::closeTimestep(adaptInfo);
	
	DOFVector<double> displacement(elastProb->getFeSpace(), "displacement");
	displacement.interpol(beamDisplacement);
	VtkWriter::writeFile(&displacement, "displacement.vtu");

//     flag->move(elastProb->getDeformation());
    updateMesh(elastProb->getSolution(), parametricCoords1, parametric1);
    updateMesh(elastProb2->getSolution(), parametricCoords2, parametric2);
//     refinement1->refine(5);
//     refinement2->refine(5);
  }

  void updateMesh(SystemVector* vec, WorldVector<DOFVector<double>*>& parametricCoords, ParametricFirstOrder* parametric)
  {
    FUNCNAME("ParametricSphere::buildAfterCoarsen()");
    MSG("calculation of parametric coordinates\n");
    const FiniteElemSpace* feSpace = vec->getFeSpace(0);
    int dim = feSpace->getMesh()->getDim();
    int dow = Global::getGeo(WORLD);
    WorldVector<double> newCoords;

    DegreeOfFreedom dof;
    WorldVector<double> x;

    const BasisFunction *basFcts = feSpace->getBasisFcts();
    int numBasFcts = basFcts->getNumber();
    std::vector<DegreeOfFreedom> localIndices(numBasFcts);
    DOFAdmin *admin = feSpace->getAdmin();

    // ===== disable parametric traverse =====
    feSpace->getMesh()->setParametric(NULL);

    std::map<DegreeOfFreedom, bool> visited;

    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(feSpace->getMesh(), -1,
					Mesh::CALL_LEAF_EL | Mesh::FILL_COORDS);
    while (elInfo) {
      basFcts->getLocalIndices(elInfo->getElement(), admin, localIndices);
      for (int i = 0; i < dim + 1; i++) {
	dof = localIndices[i];
	if (!visited[dof]) {
	  for (int j = 0; j < dow; j++)
	    (*(parametricCoords[j]))[dof] += (*vec->getDOFVector(j))[dof];

	  visited[dof] = true;
	}
      }
      elInfo = stack.traverseNext(elInfo);
    }

    // ===== enable parametric traverse =====
    feSpace->getMesh()->setParametric(parametric);
  }

protected:

  Elasticity_Type1 *elastProb;	// LinearElasticity baseProblem
  Elasticity_Type2 *elastProb2;	// LinearElasticity baseProblem

  /// DOFVector storing parametric coordinates.
  WorldVector<DOFVector<double>*> parametricCoords1;
  WorldVector<DOFVector<double>*> parametricCoords2;

  /// Parametrizes vertex coordinates while mesh traversal.
  ParametricFirstOrder *parametric1;
  ParametricFirstOrder *parametric2;

//   Polygon* flag;
//   Polygon* obstacle;
//   AbstractFunction<double, WorldVector<double> >* signedDistFct;

//   SignedDistRefinement* refFunction1;
//   SignedDistRefinement* refFunction2;
//   RefinementLevelCoords2* refinement1;
//   RefinementLevelCoords2* refinement2;

  BeamDisplacement* beamDisplacement;
  
  unsigned dim;		// dimension of the mesh
  unsigned dow;		// dimension of the world
};


#endif // MOVING_MESH_H
