/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/


#ifndef EXTENSIONS_KDTREE_NANOFLANN_DOF_H
#define EXTENSIONS_KDTREE_NANOFLANN_DOF_H

#include <nanoflann.hpp>

#include <cstdlib>
#include <iostream>

#include "AMDiS.h"

using namespace nanoflann;
using namespace AMDiS;

namespace experimental {

  /** A simple vector-of-vectors adaptor for nanoflann, without duplicating the storage.
    *  The i'th vector represents a point in the state space.
    *
    *  \tparam DIM If set to >0, it specifies a compile-time fixed dimensionality for the points in the data set, allowing more compiler optimizations.
    *  \tparam value_type The type of the point coordinates (typically, double or float).
    *  \tparam Distance The distance metric to use: nanoflann::metric_L1, nanoflann::metric_L2, nanoflann::metric_L2_Simple, etc.
    *  \tparam index_type The type for indices in the KD-tree index (typically, size_t of int)
    */
  template < class VectorOfVectorsType, 
	    typename value_type = double, 
	    int DIM = -1, 
	    class Distance = nanoflann::metric_L2_Simple, 
	    typename index_type = DegreeOfFreedom >
  struct KDTreeDOFVectorOfWorldVectorsAdaptor
  {
    typedef KDTreeDOFVectorOfWorldVectorsAdaptor< VectorOfVectorsType,
						value_type, 
						DIM, 
						Distance,
						index_type > self_t;
    typedef typename Distance::template traits<value_type, self_t>::distance_t metric_t;
    typedef KDTreeSingleIndexAdaptor< metric_t, 
				      self_t, 
				      DIM, 
				      index_type > index_t;

    index_t* index; //! The kd-tree index for the user to call its methods as usual with any other FLANN index.

    /// Constructor: takes a const ref to the vector of vectors object with the data points
    KDTreeDOFVectorOfWorldVectorsAdaptor(const int dimensionality,
				      VectorOfVectorsType &mat, 
				      const int leaf_max_size = 10) : m_data(mat)
    {
      const size_t dims = (*mat.begin()).getSize();
      assert(mat.getUsedSize() != 0 && dims != 0);
      if (DIM > 0 && static_cast<int>(dims) != DIM)
	throw std::runtime_error("Data set dimensionality does not match the 'DIM' template argument");
      index = new index_t( dims, *this /* adaptor */, nanoflann::KDTreeSingleIndexAdaptorParams(leaf_max_size, dims ) );
      index->buildIndex();
    }

    ~KDTreeDOFVectorOfWorldVectorsAdaptor() {
      delete index;
    }
    
    void reinit(const int leaf_max_size = 10)
    {
      delete index;
      const size_t dims = (*m_data.begin()).getSize();
      index = new index_t( dims, *this /* adaptor */, nanoflann::KDTreeSingleIndexAdaptorParams(leaf_max_size, dims ) );
      index->buildIndex();
    }

    VectorOfVectorsType &m_data;

    /** Query for the \a num_closest closest points to a given point (entered as query_point[0:dim-1]).
      *  Note that this is a short-cut method for index->findNeighbors().
      *  The user can also call index->... methods as desired.
      * \note nChecks_IGNORED is ignored but kept for compatibility with the original FLANN interface.
      */
    inline void query(const value_type *query_point, 
		      const size_t num_closest, 
		      index_type *out_indices, 
		      value_type *out_distances_sq, 
		      const int nChecks_IGNORED = 10) const
    {
      nanoflann::KNNResultSet<value_type, index_type> resultSet(num_closest);
      resultSet.init(out_indices, out_distances_sq);
      index->findNeighbors(resultSet, query_point, nanoflann::SearchParams());
    }

    /** @name Interface expected by KDTreeSingleIndexAdaptor
      * @{ */

    const self_t & derived() const {
      return *this;
    }
    self_t & derived() {
      return *this;
    }

    // Must return the number of data points
    inline size_t kdtree_get_point_count() const {
      return m_data.getUsedSize();
    }

    // Returns the distance between the vector "p1[0:size-1]" and the data point with index "idx_p2" stored in the class:
    inline value_type kdtree_distance(const value_type *p1, 
				      const index_type idx_p2, 
				      size_t size) const
    {
      value_type s = 0.0;
      for (size_t i = 0; i < size; i++) {
	const value_type d = p1[i] - m_data[idx_p2][i];
	s += d*d;
      }
      return s;
    }

    // Returns the dim'th component of the idx'th point in the class:
    inline value_type kdtree_get_pt(const index_type idx, size_t dim) const {
      return m_data[idx][dim];
    }

    // Optional bounding-box computation: return false to default to a standard bbox computation loop.
    //   Return true if the BBOX was already computed by the class and returned in "bb" so it can be avoided to redo it again.
    //   Look at bb.size() to find out the expected dimensionality (e.g. 2 or 3 for point clouds)
    template <class BBOX>
    bool kdtree_get_bbox(BBOX &bb) const {
      return false;
    }

    /** @} */
    
    
    /**
    * strategy and nrOfPoints are ignored
    *  == evalAtPoint_simple
    **/
    template<typename T>
    T evalAtPoint(const DOFVector<T>& vec, const WorldVector<double>& x, int strategy = 0, int nrOfPoints = -1)
    {
      const size_t nnp = 1;
      std::vector<DegreeOfFreedom> ret_indexes(nnp);
      std::vector<double> out_dists_sqr(nnp);
      nanoflann::KNNResultSet<double, DegreeOfFreedom> resultSet(nnp);
      
      double eps = 0.0;
      Parameters::get("KD-Tree->eps",eps);
      resultSet.init(&ret_indexes[0], &out_dists_sqr[0]);
      index->findNeighbors(resultSet, x.begin(), nanoflann::SearchParams(10, eps));
      
      T value; nullify(value);
      value = vec[ret_indexes[0]];
      return value;
    }

  }; // end of KDTreeDOFVectorOfWorldVectorsAdaptor

  typedef KDTreeDOFVectorOfWorldVectorsAdaptor<DOFVector<WorldVector<double> >, double >  KD_Tree_Dof;
  static std::map<const FiniteElemSpace*, std::pair<int, KD_Tree_Dof*> > kdtreeMap_Dof;

  inline KD_Tree_Dof* provideKDTree(const FiniteElemSpace* feSpace)
  {
    if (kdtreeMap_Dof.find(feSpace) != kdtreeMap_Dof.end()) {
      if (kdtreeMap_Dof[feSpace].first != feSpace->getMesh()->getChangeIndex()) {
	feSpace->getMesh()->getDofIndexCoords(kdtreeMap_Dof[feSpace].second->m_data);
	kdtreeMap_Dof[feSpace].second->reinit();
	kdtreeMap_Dof[feSpace].first = feSpace->getMesh()->getChangeIndex();
      }
    } else {
      DOFVector<WorldVector<double> >* coords = new DOFVector<WorldVector<double> >(feSpace, "coords");
      feSpace->getMesh()->getDofIndexCoords(*coords);
      
      KD_Tree_Dof* kdtree = new KD_Tree_Dof(Global::getGeo(WORLD), *coords);
      kdtreeMap_Dof[feSpace] = std::make_pair(feSpace->getMesh()->getChangeIndex(), kdtree);
    }

    return kdtreeMap_Dof[feSpace].second;
  }
  
} // end namespace

#endif // EXTENSIONS_KDTREE_NANOFLANN_DOF_H
