/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/
#include "CahnHilliard_.h"
// #include "Views.h"
#include "SignedDistFunctors.h"
#include "PhaseFieldConvert.h"

#include "HL_SignedDistTraverse.h"
#include "Recovery.h"

using namespace AMDiS;

CahnHilliard_::CahnHilliard_(const std::string &name_) :
  super(name_),
  useMobility(false),
  useReinit(false),
  doubleWell(0),
  gamma(1.0),
  eps(0.1),
  minusEps(-0.1),
  epsInv(10.0),
  minusEpsInv(-10.0),
  epsSqr(0.01),
  minusEpsSqr(-0.01)
{
  // parameters for CH
  Parameters::get(name_ + "->use mobility", useMobility); // mobility
  Parameters::get(name_ + "->gamma", gamma); // mobility
  Parameters::get(name_ + "->epsilon", eps); // interface width
  
  // type of double well: 0= [0,1], 1= [-1,1]
  Parameters::get(name_ + "->double-well type", doubleWell); 
  
  Parameters::get(name + "->use reinit", useReinit);

  // transformation of the parameters
  minusEps = -eps;
  epsInv = 1.0/eps;
  minusEpsInv = -epsInv;
  epsSqr = sqr(eps);
  minusEpsSqr = -epsSqr;
}


void CahnHilliard_::solveInitialProblem(AdaptInfo *adaptInfo) 
{
  Flag initFlag = initDataFromFile(adaptInfo);

  if (!initFlag.isSet(DATA_ADOPTED)) {
    int initialInterface = 0;
    Initfile::get(name + "->initial interface", initialInterface);
    double initialEps = eps;
    Initfile::get(name + "->initial epsilon", initialEps);

    if (initialInterface == 0) {
      /// horizontale Linie
      double a= 0.0, dir= -1.0;
      Initfile::get(name + "->line->pos", a);
      Initfile::get(name + "->line->direction", dir);
      prob->getSolution()->getDOFVector(1)->interpol(new Plane(a, dir));
    }
    else if (initialInterface == 1) {
      /// schraege Linie
      double theta = m_pi/4.0;
      prob->getSolution()->getDOFVector(1)->interpol(new PlaneRotation(0.0, theta, 1.0));
      transformDOFInterpolation(prob->getSolution()->getDOFVector(1),new PlaneRotation(0.0, -theta, -1.0), new AMDiS::Min<double>);
    }
    else if (initialInterface == 2) {
      /// Ellipse
      double a= 1.0, b= 1.0;
      Initfile::get(name + "->ellipse->a", a);
      Initfile::get(name + "->ellipse->b", b);
      prob->getSolution()->getDOFVector(1)->interpol(new Ellipse(a,b));
    }
    else if (initialInterface == 3) {
      /// zwei horizontale Linien
      double a= 0.75, b= 0.375;
      Initfile::get(name + "->lines->pos1", a);
      Initfile::get(name + "->lines->pos2", b);
      prob->getSolution()->getDOFVector(1)->interpol(new Plane(a, -1.0));
      transformDOFInterpolation(prob->getSolution()->getDOFVector(1),new Plane(b, 1.0), new AMDiS::Max<double>);
    }
    else if (initialInterface == 4) {
      /// Kreis
      double radius= 1.0;
      Initfile::get(name + "->kreis->radius", radius);
      prob->getSolution()->getDOFVector(1)->interpol(new Circle(radius));
    } else if (initialInterface == 5) {
      /// Rechteck
      double width = 0.5;
      double height = 0.3;
      WorldVector<double> center; center.set(0.5);
      Initfile::get(name + "->rectangle->width", width);
      Initfile::get(name + "->rectangle->height", height);
      Initfile::get(name + "->rectangle->center", center);
      prob->getSolution()->getDOFVector(1)->interpol(new Rectangle(width, height, center));
    }
    
    // TODO: Redistancing einfügen!
    if (useReinit) {
      FiniteElemSpace* feSpace = FiniteElemSpace::provideFeSpace(
	const_cast<DOFAdmin*>(getMesh()->getVertexAdmin()),
	Lagrange::getLagrange(getMesh()->getDim(), 1),
	getMesh(),
	"P1");
      DOFVector<double> tmp(feSpace, "tmp");
      tmp.interpol(prob->getSolution()->getDOFVector(1));
    
      HL_SignedDistTraverse reinit("reinit", getMesh()->getDim());
      reinit.calcSignedDistFct(adaptInfo, &tmp);
    
#ifndef HAVE_PARALLEL_DOMAIN_AMDIS
      Recovery recovery(L2_NORM, 1);
      recovery.recoveryUh(&tmp, *prob->getSolution()->getDOFVector(1));
#else
      prob->getSolution()->getDOFVector(1)->interpol(&tmp);
#endif
    }
    

    /// create phase-field from signed-dist-function
    if (doubleWell == 0) {
      forEachDOF(prob->getSolution()->getDOFVector(1),
        new SignedDistToPhaseField(initialEps));
    } else {
      forEachDOF(prob->getSolution()->getDOFVector(1),
        new SignedDistToCh(initialEps));
    }
  }
}


void CahnHilliard_::fillOperators()
{  
  const FiniteElemSpace* feSpace = prob->getFeSpace();
  int degree = feSpace->getBasisFcts()->getDegree();

  // c
  Operator *opChMnew = new Operator(feSpace, feSpace);
  opChMnew->addTerm(new Simple_ZOT);
  Operator *opChMold = new Operator(feSpace, feSpace);
  opChMold->addTerm(new VecAtQP_ZOT(prob->getSolution()->getDOFVector(1)));
  // -nabla*(grad(c))
  Operator *opChL = new Operator(feSpace, feSpace);
  opChL->addTerm(new Simple_SOT);
  
  // div(M(c)grad(mu)), with M(c)=gamma/4*(c^2-1)^2
  Operator *opChLM = new Operator(feSpace, feSpace);
  if (useMobility) {
    if (doubleWell == 0)
      opChLM->addTerm(new VecAtQP_SOT(
	prob->getSolution()->getDOFVector(1),
	new MobilityCH0(gamma, degree)));
    else
      opChLM->addTerm(new VecAtQP_SOT(
	prob->getSolution()->getDOFVector(1),
	new MobilityCH1(gamma, degree)));
  } else
    opChLM->addTerm(new Simple_SOT(gamma));
  
  // -2*c_old^3 + 3/2*c_old^2
  Operator *opChMPowExpl = new Operator(feSpace, feSpace);
  opChMPowExpl->addTerm(new VecAtQP_ZOT(
    prob->getSolution()->getDOFVector(1),
    new AMDiS::Pow<3>(-2.0, 3*degree)));
  if (doubleWell == 0) {
    opChMPowExpl->addTerm(new VecAtQP_ZOT(
      prob->getSolution()->getDOFVector(1),
      new AMDiS::Pow<2>(3.0/2.0, 2*degree)));
  }

  // -3*c_old^2 * c
  Operator *opChMPowImpl = new Operator(feSpace, feSpace);
  opChMPowImpl->addTerm(new VecAtQP_ZOT(
    prob->getSolution()->getDOFVector(1),
    new AMDiS::Pow<2>(-3.0, 2*degree)));
  if (doubleWell == 0) {
    opChMPowImpl->addTerm(new VecAtQP_ZOT(
      prob->getSolution()->getDOFVector(1),
      NULL, 3.0));
    opChMPowImpl->addTerm(new Simple_ZOT(-0.5));
  } else {
    opChMPowImpl->addZeroOrderTerm(new Simple_ZOT(1.0));
  }

  // mu + eps^2*laplace(c) + c - 3*(c_old^2)*c = -2*c_old^3 [+ BC]
  // ----------------------------------------------------------------------
  prob->addMatrixOperator(*opChMPowImpl,0,1);          /// < -3*c*c_old^2 , psi >
  prob->addMatrixOperator(*opChL,0,1, &minusEpsSqr);   /// < -eps^2*grad(c) , grad(psi) >
  prob->addMatrixOperator(*opChMnew,0,0);              /// < mu , psi >
  // . . . vectorOperators . . . . . . . . . . . . . . .
  prob->addVectorOperator(*opChMPowExpl,0);            /// < -2*c_old^3 , psi >

  
  // dt(c) = laplace(mu) - u*grad(c)
  // -----------------------------------
  prob->addMatrixOperator(*opChMnew,1,1); 		/// < c , psi >
  prob->addMatrixOperator(*opChLM,1,0, getTau());       /// < tau*grad(mu) , grad(psi) >
  // . . . vectorOperators . . . . . . . . . . . . . . .
  prob->addVectorOperator(*opChMold,1);   		/// < c^old , psi >
  
}

