/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/
#ifndef MTL_CH_PRECONDITIONER_H
#define MTL_CH_PRECONDITIONER_H

#include "AMDiS.h"
#include "BlockPreconditioner.h"

using namespace std;
using namespace AMDiS;   

template< typename MatrixType = MTLTypes::MTLMatrix, typename VectorType = MTLTypes::MTLVector >
struct MTLPreconCahnHilliard : BlockPreconditioner
{
  typedef BlockPreconditioner super;
 
  class Creator : public PreconditionCreator
  {
  public:
    virtual ~Creator() {}
    
    BasePreconditioner* create() { 
      return new MTLPreconCahnHilliard(this->name);
    }
  };  
  
  
  MTLPreconCahnHilliard(std::string name)
  : name(name), 
    eps(NULL), delta(NULL),
    solverM(NULL), solverMpL(NULL), runnerM(NULL), runnerMpL(NULL)
  {
    createSubSolver(name + "->subsolver M", solverM, runnerM);
    createSubSolver(name + "->subsolver MpL", solverMpL, runnerMpL);
  }
  
  
  virtual ~MTLPreconCahnHilliard()
  {
    exit();
    delete solverM;
    delete solverMpL;
  }
  
  
  virtual void init(const typename super::BlockMatrix& A, const MatrixType& fullMatrix);  
  
  virtual void solve(const VectorType& b, VectorType& x) const;
  
  
  virtual void exit()
  {
    runnerM->exit();
    runnerMpL->exit();
  }
  
  
  const MatrixType& getM() const
  {
    return A->getSubMatrix(0,0);
  }
  
  
  const MatrixType& getL() const
  { 
    return L; // gamma*tau*L
  }
  
  
  double getEps() const
  {
    return *eps;
  }
  
  
  double getDelta() const
  {
    return *delta;
  }
  
  
  void setData(double* eps_, double* delta_)
  {
    eps = eps_;
    delta = delta_;
  }
  
protected:      
  template <typename VectorX, typename VectorB>
  void solveM(VectorX& x, const VectorB& b) const
  {
    runnerM->solve(getM(), x, b);
  }
  
  
  template <typename VectorX, typename VectorB>
  void solveMpL(VectorX& x, const VectorB& b) const
  {
    runnerMpL->solve(MpL, x, b);
  }
  
  
  MatrixType M;
  MatrixType MpL;
  MatrixType L;
   
  mutable MTLVector y1;
  mutable MTLVector y2;
  
  LinearSolver* solverM;
  LinearSolver* solverMpL;
  
  MTL4Runner<MatrixType, VectorType>* runnerM;
  MTL4Runner<MatrixType, VectorType>* runnerMpL;
  
  std::string name;
  
  double* eps;
  double* delta;
};

#include "MTLPreconCahnHilliard.hh"

#endif // MTL_CH_PRECONDITIONER_H

