/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/

template< typename MatrixType, typename VectorType >
void MTLPreconCahnHilliard< MatrixType, VectorType >::init(const BlockMatrix& A_, const MTLMatrix& fullMatrix_)
{ FUNCNAME("MTLPreconCahnHilliard::init()");

  super::init(A_, fullMatrix_);
  
  assert(eps != NULL);
  assert(delta != NULL);
  
  const FiniteElemSpace* feSpace = (*A_.getOriginalMat())[0][0]->getFeSpace();
  DOFMatrix laplaceMatrix(feSpace, feSpace);
  Operator laplaceOp(feSpace, feSpace);
  Simple_SOT sot;
  laplaceOp.addTerm(&sot);
  laplaceMatrix.assembleOperator(laplaceOp);
    
  L.change_dim(num_rows(getM()), num_cols(getM()));
  L = laplaceMatrix.getBaseMatrix();
  
  // helper-matrix MpL = M + eps*sqrt(delta)*L
  MpL.change_dim(num_rows(getM()), num_cols(getM()));
  MpL = getM() + ( getEps()*sqrt(getDelta()) ) * getL();
  
  // temporary variables
  y1.change_dim(num_rows(getM()));
  y2.change_dim(num_rows(getM()));
  
  runnerM->init(A_, getM());
  runnerMpL->init(A_, MpL);
}


template< typename MatrixType, typename VectorType >
void MTLPreconCahnHilliard< MatrixType, VectorType >::solve(const MTLVector& b, MTLVector& x) const
{ FUNCNAME("MTLPreconCahnHilliard::solve()");

  x.change_dim(num_rows(b));
  
  const MTLVector b1(b[rows[0]]);
  const MTLVector b2(b[rows[1]]);
  
  MTLVector x1(x[rows[0]]);
  MTLVector x2(x[rows[1]]);
    
  double factor = getEps()/sqrt(getDelta());
  
  solveM(y1, b1);
  y2 = getL() * y1; 
  x1 = b2 - getDelta() * y2;
  
  solveMpL(y2, x1);
  x1 = getM() * y2;

  solveMpL(x2, x1);
  x1 = y1 - factor * (y2 - x2);
}
