/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/
#ifndef MTL_PFC_PRECONDITIONER_H
#define MTL_PFC_PRECONDITIONER_H

#include "AMDiS.h"
#include "BlockPreconditioner.h"

using namespace std;
using namespace AMDiS;    

template< typename Base >
struct PfcSchurMatrix
{
  PfcSchurMatrix() : super(NULL) {};
  PfcSchurMatrix(Base* super) : super(super)
  {
    m = num_rows(super->getM());
    y1.change_dim(m);
    y2.change_dim(m);
    y3.change_dim(m);
  }

  template <typename VectorIn, typename VectorOut, typename Assign>
  void mult(const VectorIn& b, VectorOut& x, Assign) const;

  template <typename VectorIn>
  mtl::vector::mat_cvec_multiplier<PfcSchurMatrix<Base>, VectorIn> operator*(const VectorIn& v) const
  {   return mtl::vector::mat_cvec_multiplier<PfcSchurMatrix<Base>, VectorIn>(*this, v);    }

  int m;
  
private:
  Base* super;
  
  mutable MTLVector y1;
  mutable MTLVector y2;
  mutable MTLVector y3;
};


struct MTLPreconPfc : BlockPreconditioner
{
  typedef BlockPreconditioner super;
 
  class Creator : public PreconditionCreator
  {
  public:
    virtual ~Creator() {}
    
    BasePreconditioner* create() { 
      return new MTLPreconPfc(this->name);
    }
  };  
  
  MTLPreconPfc(std::string name)
  : BlockPreconditioner(),
    name(name), 
    tau(NULL),
    solverM(NULL), solverM2(NULL), solverMpL(NULL), runnerM(NULL), runnerM2(NULL), runnerMpL(NULL),
    Pid(NULL), Pdiag(NULL),
    maxIterS(30), restartS(100)
  {
    createSubSolver(name + "->subsolver M", solverM, runnerM, "cg");
    createSubSolver(name + "->subsolver M2", solverM2, runnerM2, "cg");
    createSubSolver(name + "->subsolver MpL", solverMpL, runnerMpL, "cg");
    
    Parameters::get(name + "->subsolver S->solver->max iteration", maxIterS);
    Parameters::get(name + "->subsolver S->solver->restart", restartS);
  }
  
  virtual ~MTLPreconPfc()
  {
    delete solverM;
    delete solverMpL;
    
    if (Pid) { delete Pid; Pid = NULL; }
    if (Pdiag) { delete Pdiag; Pdiag = NULL; }
  }
  
  void setData(double* tau_)
  {
    tau = tau_;
  }
  
  virtual void init(const typename super::BlockMatrix& A, const MTLTypes::MTLMatrix& fullMatrix);
  
  virtual void exit()
  {
    runnerM->exit();
    runnerM2->exit();
    runnerMpL->exit();
    
    if (Pid) { delete Pid; Pid = NULL; }
    if (Pdiag) { delete Pdiag; Pdiag = NULL; }
  }
  
  virtual void solve(const MTLVector& b, MTLVector& x) const;  
    
  template <typename VectorX, typename VectorB>
  void solveM(VectorX& x, const VectorB& b, int nIter = -1) const
  {
    runnerM->solve(getM(), x, b);
  }
  template <typename VectorX, typename VectorB>
  void solveM2(VectorX& x, const VectorB& b) const
  {
    runnerM2->solve(getM(), x, b);
  }
  
  template <typename VectorX, typename VectorB>
  void solveMpK(VectorX& x, const VectorB& b) const
  {
    runnerMpL->solve(MpL, x, b);
  }
  
  template <typename SchurMatrix, typename VectorX, typename VectorB>
  void solveS(const SchurMatrix& S, VectorX& x, const VectorB& b) const
  { 
    itl::basic_iteration<double> iter(b, maxIterS, 0, 0);
    x = 0.0;
    itl::fgmres_householder(S, x, b, *Pdiag, iter, restartS);	// S*x = b
  }
  
  const MTLMatrix& getM() const { return A->getSubMatrix(2,2); }
  const MTLMatrix& getL() const { return A->getSubMatrix(2,1); }
  double getTau() const { return *tau; }
  
protected: 
  MTLMatrix MpL;
  PfcSchurMatrix<MTLPreconPfc> S;
   
  mutable MTLVector y0;
  mutable MTLVector y1;
  mutable MTLVector tmp;
  
  LinearSolver* solverM;
  LinearSolver* solverM2;
  LinearSolver* solverMpL;
  
  MTL4Runner<MTLTypes::MTLMatrix, MTLTypes::MTLVector>* runnerM;
  MTL4Runner<MTLTypes::MTLMatrix, MTLTypes::MTLVector>* runnerM2;
  MTL4Runner<MTLTypes::MTLMatrix, MTLTypes::MTLVector>* runnerMpL;
  
  itl::pc::identity<MTLTypes::MTLMatrix, double> *Pid;
  itl::pc::diagonal<MTLTypes::MTLMatrix, double> *Pdiag;
  
  std::string name;
  
  double* tau;
  
  int maxIterS;
  int restartS;
};

// some necessary functions that must be available for matrix-types

inline std::size_t size(const PfcSchurMatrix<MTLPreconPfc>& A) { return A.m * A.m; }
inline std::size_t num_rows(const PfcSchurMatrix<MTLPreconPfc>& A) { return A.m; }
inline std::size_t num_cols(const PfcSchurMatrix<MTLPreconPfc>& A) { return A.m; }

namespace mtl {
  template <>
  struct Collection<PfcSchurMatrix<MTLPreconPfc> >
  {
    typedef double value_type;
    typedef int    size_type;
  };

  namespace ashape {
    template <> struct ashape_aux<PfcSchurMatrix<MTLPreconPfc> > 
    {       typedef nonscal type;    };
  }
}

#include "MTLPreconPfc.hh"

#endif // MTL_PFC_PRECONDITIONER_H

