/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/
#include "PetscPreconPfc.h"

namespace AMDiS {

  using namespace std;
  int pcPfcMatMultSchur(Mat mat, Vec b, Vec x)
  {
    void *ctx;
    MatShellGetContext(mat, &ctx);
    PfcData* data = static_cast<PfcData*>(ctx);

    Vec y1, y2, y3;
    VecDuplicate(b, &y1);
    VecDuplicate(b, &y2);
    VecDuplicate(b, &y3);
    
    MatMult(data->matK, b, y1);		// y1 := K*b
    KSPSolve(data->kspM, y1, y2);	// M*y2 = y1
    MatMult(data->matK, y2, y3);	// y3 := K*y2
//     KSPSolve(data->kspM, y3, y2);	// M*y2 = y3
//     MatMult(data->matK, y2, y3);	// y3 := K*y2
    
    VecScale(y3, data->delta);		// y3 = delta*y3
    MatMultAdd(data->matM, b, y3, x);	// x = M*b + y3
    VecAXPY(x, -2.0*(data->delta), y1); // x := M*b - 2*delta*K*b + delta*B*b
    
    VecDestroy(&y1);
    VecDestroy(&y2);
    VecDestroy(&y3);
    return 0;
  }
  
  /// solve Pfc Preconditioner
  PetscErrorCode pcPfcShell(PC pc, Vec b, Vec x) // solve Px=b
  { FUNCNAME("pcPfcShell()");
  
    void *ctx;
    PCShellGetContext(pc, &ctx);
    PfcData* data = static_cast<PfcData*>(ctx);
    
    Vec b1, b2, b3, x1, x2, x3;
    VecNestGetSubVec(b, 0, &b1);
    VecNestGetSubVec(b, 1, &b2);
    VecNestGetSubVec(b, 2, &b3);

    VecNestGetSubVec(x, 0, &x1);
    VecNestGetSubVec(x, 1, &x2);
    VecNestGetSubVec(x, 2, &x3);


    // Hilfsvariablen
    Vec y1, y2, tmp;
    VecDuplicate(b1, &y1);
    VecDuplicate(b1, &y2);
    VecDuplicate(b1, &tmp);

    KSPSolve(data->kspM, b1, y1); 			// M*y1 = b1    
    MatMult(data->matK, y1, tmp); 			// tmp := K*y1
    VecAYPX(tmp, -(data->tau), b2);			// tmp := b2 - tau*tmp
    
    KSPSolve(data->kspMpK, tmp, y2); 			// (M+sqrt(tau)K)*y2 = tmp
    MatMult(data->matM, y2, tmp);			// tmp := M*y2

    KSPSolve(data->kspSchur, tmp, x2); 			// S*x2 = tmp
    VecCopy(x2, x1); 					// x1 := x2
    VecAXPBYPCZ(x1, 1.0, 1.0/(data->delta), -1.0/(data->delta), y1, y2);	// x1 = 1*y1 + factor*y2 - factor*x1

    MatMult(data->matK, x2, tmp); 			// tmp := K*x2
    VecAYPX(tmp, -1.0, b3);				// tmp := b3 - tmp
    KSPSolve(data->kspM2, tmp, x3); 			// M*x3 = tmp    

    VecDestroy(&y1);
    VecDestroy(&y2);
    VecDestroy(&tmp);
  }
  

  void PetscPreconPfc::init(const BlockMatrix& A, const Mat& matrix)
  {
    assert(tau != NULL);
    dynamic_cast<PetscSolver<PetscPreconPfc>*>(&oem)->setNestedVectors(true);
    PetscRunner::init(A, matrix);
    
    // init shell preconditioner
    PCSetType(getPc(), PCSHELL);
    PCShellSetApply(getPc(), pcPfcShell);
    PCShellSetContext(getPc(), &data);

    MatNestGetSubMat(matrix, 2, 2, &data.matM);
    MatNestGetSubMat(matrix, 2, 1, &data.matK);
    
    MatDuplicate(data.matM, MAT_COPY_VALUES, &MpK);
    MatAXPY(MpK, sqrt(*tau), data.matK, SAME_NONZERO_PATTERN);
    
    // schur-komplete-Matrix
    PetscInt numRows, numCols;
    MatGetSize(data.matM, &numRows, &numCols);
    MatCreateShell(PETSC_COMM_SELF, numRows, numCols, 
		    PETSC_DETERMINE, PETSC_DETERMINE,
		    &data, &matSchur);
    MatShellSetOperation(matSchur, MATOP_MULT, (void(*)(void))pcPfcMatMultSchur);

    KSPCreate(PETSC_COMM_SELF, &data.kspSchur);
    KSPSetOperators(data.kspSchur, matSchur, MpK, SAME_NONZERO_PATTERN);
    setSolver(data.kspSchur, "schur_", KSPFGMRES, PCJACOBI, 1e-6, 1e-8, 4);
    KSPSetFromOptions(data.kspSchur);
    
    // init sub-solvers
    createSubSolver(data.kspM, data.matM, "mass_");
    createSubSolver(data.kspM2, data.matM, "mass2_");
    createSubSolver(data.kspMpK, MpK, "MpK_");
    
    data.delta = sqrt(*tau);
    data.tau = (*tau);

    setSolver(data.kspM, "mass_", KSPCG, PCJACOBI, 0.0, 1e-14, 5);
    setSolver(data.kspM2, "mass2_", KSPCG, PCJACOBI, 0.0, 1e-14, 15);
    setSolver(data.kspMpK, "MpK_", KSPRICHARDSON, PCHYPRE, 0.0, 1e-14, 5);
    
  }


  void PetscPreconPfc::exit()
  {
    FUNCNAME("PetscPreconPfc::exit()");
    
    MatDestroy(&MpK);
    MatDestroy(&matSchur);
    
    KSPDestroy(&data.kspM);
    KSPDestroy(&data.kspM2);
    KSPDestroy(&data.kspMpK);
    KSPDestroy(&data.kspSchur);
    
    PetscRunner::exit();
  }
}
