/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/


#include "time/ExtendedRosenbrockAdaptInstationary.h"
#include "AdaptInfo.h"
#include "ProblemTimeInterface.h"
#include "VectorOperations.h"

namespace AMDiS {

  template<typename ProblemStationary>
  ExtendedRosenbrockAdaptInstationary<ProblemStationary>::ExtendedRosenbrockAdaptInstationary(std::string name,
							   ProblemStationary *problemStat,
							   AdaptInfo *info,
							   ProblemTimeInterface *problemInstat,
							   AdaptInfo *initialInfo,
							   time_t initialTimestamp)
    : AdaptInstationary(name, problemStat, info, problemInstat, initialInfo, initialTimestamp),
      rosenbrockStat(problemStat),
      firstTimestep(true),
      lastTimestepRejected(false),
      succRejection(false),
      fixFirstTimesteps(0),
      tau(1.0),
      dbgTimestepStudy(false)
  {
    FUNCNAME("ExtendedRosenbrockAdaptInstationary::ExtendedRosenbrockAdaptInstationary()");
    initConstructor(problemStat);
    rosenbrockStat->setOldTime(adaptInfo->getStartTime());
  }


  template<typename ProblemStationary>
  ExtendedRosenbrockAdaptInstationary<ProblemStationary>::ExtendedRosenbrockAdaptInstationary(std::string name,
							   ProblemStationary &problemStat,
							   AdaptInfo &info,
							   ProblemTimeInterface &problemInstat,
							   AdaptInfo &initialInfo,
							   time_t initialTimestamp)
    : AdaptInstationary(name, problemStat, info, problemInstat, initialInfo, initialTimestamp),
      rosenbrockStat(&problemStat),
      firstTimestep(true),
      lastTimestepRejected(false),
      succRejection(false),
      fixFirstTimesteps(0),
      tau(1.0),
      dbgTimestepStudy(false)
  {
    FUNCNAME("ExtendedRosenbrockAdaptInstationary::ExtendedRosenbrockAdaptInstationary()");
    initConstructor(&problemStat);
    rosenbrockStat->setOldTime(adaptInfo->getStartTime());
  }


  template<typename ProblemStationary>
  void ExtendedRosenbrockAdaptInstationary<ProblemStationary>::initConstructor(ProblemStationary *problemStat)
  {
    FUNCNAME("ExtendedRosenbrockAdaptInstationary::initConstructor()");

    adaptInfo->setRosenbrockMode(true);
    rosenbrockStat->setTau(&tau);
    
    Parameters::get(name + "->fix first timesteps", fixFirstTimesteps);
    Parameters::get(name + "->rosenbrock->timestep study", dbgTimestepStudy);
    Parameters::get(name + "->rosenbrock->timestep study steps", dbgTimesteps);
  }


  template<typename ProblemStationary>
  void ExtendedRosenbrockAdaptInstationary<ProblemStationary>::reset()
  {
    FUNCNAME("ExtendedRosenbrockAdaptInstationary::reset()");

    firstTimestep = true;
    lastTimestepRejected = false;
    succRejection = false;
    fixFirstTimesteps = 0;
  }

/*
  template<typename ProblemStationary>
  void ExtendedRosenbrockAdaptInstationary<ProblemStationary>::oneTimestep()
  {
    FUNCNAME("ExtendedRosenbrockAdaptInstationary::oneTimestep()");
    
    // estimate before first adaption
    if (adaptInfo->getTime() <= adaptInfo->getStartTime())
      problemIteration->oneIteration(adaptInfo, ESTIMATE);
    
    bool rejected = false;
    double timeTol = adaptInfo->getTimeTolerance(0);

    int studyTimestep = -1;

    do {
      if (dbgTimestepStudy) {
	if (studyTimestep >= 0)
	  adaptInfo->setTime(adaptInfo->getTime() - dbgTimesteps[studyTimestep]);

	studyTimestep++;
	TEST_EXIT(studyTimestep < dbgTimesteps.size())("Should not happen!\n");
	
	adaptInfo->setTimestep(dbgTimesteps[studyTimestep]);
      }

      // increment time
      adaptInfo->setTime(adaptInfo->getTime() + adaptInfo->getTimestep());
      problemTime->setTime(adaptInfo);
      tau = adaptInfo->getTimestep();
      
      INFO(info, 6)("time = %e   timestep = %e\n",
		    adaptInfo->getTime(), adaptInfo->getTimestep());
      
      adaptInfo->setSpaceIteration(0);
      
      // do the iteration
      problemIteration->beginIteration(adaptInfo);
      problemIteration->oneIteration(adaptInfo, FULL_ITERATION);
      problemIteration->endIteration(adaptInfo);

      double errorEst = adaptInfo->getTimeEst();
      INFO(info, 6)("overall time estimate = %e\n", errorEst);
      
      double newTimestep = 0.0;

      if (errorEst < timeTol || fixFirstTimesteps > 0 || firstTimestep) {
	double fac = 1.0;
	
	if (fixFirstTimesteps > 0) {
	  newTimestep = adaptInfo->getTimestep();
	} else {
	  if (firstTimestep || succRejection)
	    newTimestep = rosenbrockStat->getNewTimestep(timeTol, adaptInfo->getTimestep());
	  else
	    newTimestep = rosenbrockStat->getNewTimestep(timeTol, tauAcc, adaptInfo->getTimestep());
	}
	tauAcc = adaptInfo->getTimestep();
	
	lastTimestepRejected = false;
	succRejection = false;
      } else {
	if (lastTimestepRejected) {
	  succRejection = true;	  
	  rosenbrockStat->reduceOrder(tauRej, adaptInfo->getTimestep());
	} 
	newTimestep = rosenbrockStat->getNewTimestep(timeTol, adaptInfo->getTimestep(), false);

	tauRej = adaptInfo->getTimestep();	
	lastTimestepRejected = true;
      }
      
      
      if (errorEst < timeTol || fixFirstTimesteps 
	  || !(adaptInfo->getTimestep()>adaptInfo->getMinTimestep()) ) {
	INFO(info, 6)("Accepted timestep at time = %e   with timestep = %e\n",
		      adaptInfo->getTime()  - adaptInfo->getTimestep(),
		      adaptInfo->getTimestep());

	adaptInfo->setTimestep(newTimestep);

	rejected = false;

	for (int i = 0; i < adaptInfo->getSize(); i++)
	  INFO(info, 6)("time estimate for component %d = %e\n", 
			i, adaptInfo->getTimeEstSum(i));

        if (errorEst > timeTol)
	  MSG("Accepted timestep but tolerance not reached \n");

      } else {
	for (int i = 0; i < adaptInfo->getSize(); i++)
	  INFO(info, 6)("time estimate for component %d = %e\n", 
			i, adaptInfo->getTimeEstSum(i));

	INFO(info, 6)("Rejected timestep at time = %e   with timestep = %e\n",
		      adaptInfo->getTime()  - adaptInfo->getTimestep(),
		      adaptInfo->getTimestep());
	
	adaptInfo->setTime(adaptInfo->getTime() - adaptInfo->getTimestep());
	adaptInfo->setTimestep(newTimestep);

	rejected = true;
      }

      if (firstTimestep)
	firstTimestep = false;

      if (fixFirstTimesteps > 0)
	fixFirstTimesteps--;     
    } while (rejected || 
	     (dbgTimestepStudy && (studyTimestep + 1 < dbgTimesteps.size())));

    rosenbrockStat->acceptTimestep(adaptInfo);    

    adaptInfo->setLastProcessedTimestep(adaptInfo->getTimestep());
    adaptInfo->incTimestepNumber();
  }*/

  template<typename ProblemStationary>
  void ExtendedRosenbrockAdaptInstationary<ProblemStationary>::oneTimestep()
  {
    FUNCNAME("ExtendedRosenbrockAdaptInstationary::oneTimestep()");
    
    // estimate before first adaption
    if (adaptInfo->getTime() <= adaptInfo->getStartTime())
      problemIteration->oneIteration(adaptInfo, ESTIMATE);
    
    double timeTol = adaptInfo->getTimeTolerance(0);

    // increment time
    adaptInfo->setTime(adaptInfo->getTime() + adaptInfo->getTimestep());
    problemTime->setTime(adaptInfo);
    tau = adaptInfo->getTimestep();
    
    INFO(info, 6)("time = %e   timestep = %e\n",
		  adaptInfo->getTime(), adaptInfo->getTimestep());
    
    adaptInfo->setSpaceIteration(0);
    
    // do the iteration
    problemIteration->beginIteration(adaptInfo);
    problemIteration->oneIteration(adaptInfo, FULL_ITERATION);
    problemIteration->endIteration(adaptInfo);

    double errorEst = adaptInfo->getTimeEst();
    INFO(info, 6)("overall time estimate = %e\n", errorEst);
    
    double newTimestep = 0.0;
    double fac = 1.0;
      
    if (fixFirstTimesteps > 0) {
      newTimestep = adaptInfo->getTimestep();
    } else {
      if (firstTimestep)
	newTimestep = rosenbrockStat->getNewTimestep(timeTol, adaptInfo->getTimestep());
      else
	newTimestep = rosenbrockStat->getNewTimestep(timeTol, tauAcc, adaptInfo->getTimestep());
    }
    tauAcc = adaptInfo->getTimestep();   
    

    for (int i = 0; i < adaptInfo->getSize(); i++)
      INFO(info, 6)("time estimate for component %d = %e\n", 
		    i, adaptInfo->getTimeEstSum(i));
	
    INFO(info, 6)("Accepted timestep at time = %e   with timestep = %e\n",
		  adaptInfo->getTime()  - adaptInfo->getTimestep(),
		  adaptInfo->getTimestep());

    adaptInfo->setTimestep(newTimestep);

    if (errorEst > timeTol)
      MSG("Accepted timestep but tolerance not reached \n");

    if (firstTimestep)
      firstTimestep = false;

    if (fixFirstTimesteps > 0)
      fixFirstTimesteps--;     

    rosenbrockStat->acceptTimestep(adaptInfo);    

    adaptInfo->setLastProcessedTimestep(adaptInfo->getTimestep());
    adaptInfo->incTimestepNumber();
  }
}
