/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/



/** \file ExtendedRosenbrockStationary.h */

#ifndef MY_ROSENBROCKSTATIONARY_H
#define MY_ROSENBROCKSTATIONARY_H

#include "AMDiS_fwd.h"
#include "ExtendedProblemStat.h"
#include "SystemVector.h"
#include "time/RosenbrockMethod.h"

#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
#include "parallel/ParallelProblemStat.h"
#endif

using namespace AMDiS;

  struct MyRosenbrockBoundary 
  {
    MyRosenbrockBoundary(AbstractFunction<double, WorldVector<double> > * fct_,
			 AbstractFunction<double, WorldVector<double> > * fctDt_,
			 DOFVector<double> *vec_,
			 int row_,
			 int col_      
    ) : fct(fct_), fctDt(fctDt_), vec(vec_), row(row_), col(col_) {}
    
    AbstractFunction<double, WorldVector<double> > *fct; // f(t,x)
    AbstractFunction<double, WorldVector<double> > *fctDt; // dt[f](t,x)
    
    DOFVector<double> *vec;

    int row;
    
    int col;
  };


  class ExtendedRosenbrockStationary : public ExtendedProblemStat
  {
  public: // typedefs
    typedef ExtendedProblemStat super;

  public:
    ExtendedRosenbrockStationary(std::string name, int componentShift_ = 0)
      : super(name),
	componentShift(componentShift_),
	order(0),
	first(true),
	fixedTimestep(false),
	enforceAsmMat(false),
	minusOne(-1.0),
	invTauGamma(1.0),
	stageTime(0.0),
	errorEst(0.0),
	oldErrorEst(0.0)
    {
      Parameters::get(name + "->enforce matrix assembly",enforceAsmMat);
    }

    
    virtual void initialize(Flag initFlag,
			    ProblemStat *adoptProblem = NULL,
			    Flag adoptFlag = INIT_NOTHING);

    virtual Flag oneIteration(AdaptInfo *adaptInfo, Flag toDo);
    Flag stageIteration(AdaptInfo *adaptInfo, Flag flag, 
			bool asmMatrix, bool asmVector);
    
    virtual double estimateTimeError(AdaptInfo* adaptInfo);

    void acceptTimestep(AdaptInfo* adaptInfo);
    
    double getNewTimestep(double tol, double tau, bool restrict = true);      
    double getNewTimestep(double tol, double oldTau, double tau, bool restrict = true);

    void addOperator(Operator &op, int row, int col, 
		     double *factor = NULL, double *estFactor = NULL);

    void addJacobianOperator(Operator &op, int row, int col, 
			     double *factor = NULL, double *estFactor = NULL);

    void addTimeOperator(int i, int j, double factor = 1.0);

    // setting methods
    // _____________________________________________
    
    void setFixedTimestep(bool fixed)
    { 
      fixedTimestep = fixed;
    }

    void setTau(double *ptr)
    {
      tauPtr = ptr;
    }

    void setInvTauGamma(double t)
    {
      invTauGamma = t;
    }
    
    void setOldTime(double t)
    {
      oldTime = t;
    }
    
    void setStageTime(double t)
    {
      stageTime = t;
    }
	  
    void reduceOrder(double tauRej, double tau)
    {
      double newOrder = log(errorEst / oldErrorEst) / log(tau / tauRej);
      if (newOrder < order && newOrder > 0.0)
	order = newOrder;
    }
    
    // getting methods
    // _____________________________________________
    
    const RosenbrockMethod* getRosenbrockMethod()
    {
      return rm;
    }
    
    double* getTauPtr()
    {
      return tauPtr;
    }
    
    DOFVector<double>* getUnVec(int i)
    {
      return unVec->getDOFVector(i);
    }
    
    SystemVector* getUnVec()
    {
      return unVec;
    }

    DOFVector<double>* getStageSolution(int i)
    {
      return stageSolution->getDOFVector(i);
    }
    
    SystemVector* getStageSolution()
    {
      return stageSolution;
    }
    
    SystemVector* getStageSolutions(int i)
    {
      return stageSolutions[i];
    }

    DOFVector<double>* getTimeRhsVec(int i)
    {
      return timeRhsVec->getDOFVector(i);
    }
    
    SystemVector* getNewUn()
    {
      return newUn;
    }
    
    SystemVector* getLowSol()
    {
      return lowSol;
    }
    
    double getStageTime()
    {
      return stageTime;
    }
    
    double* getOldTimePtr()
    {
      return &oldTime;
    }
    
    double* getStageTimePtr()
    {
      return &stageTime;
    }
    
    double* getTauGammaI()
    {
      return &tauGammaI;
    }
    
    double* getInvTauGamma()
    {
      return &invTauGamma;
    }
    
    std::vector<MyRosenbrockBoundary>& getBoundaries()
    {
      return boundaries;
    }

    /// Adds a Dirichlet boundary condition, where the rhs is given by an AbstractFunction.
    void addDirichletBC(BoundaryType type, int row, int col,
			AbstractFunction<double, WorldVector<double> > *b,
			AbstractFunction<double, WorldVector<double> > *bDt = NULL);

    /// Adds a Dirichlet boundary condition, where the rhs is given by a DOFVector.
    void addDirichletBC(BoundaryType type, int row, int col,
			DOFVector<double> *vec)
    {
      FUNCNAME("ExtendedRosenbrockStationary::addDirichletBC()");

      ERROR_EXIT("Not yet supported!\n");
    }

    // special boundary conditions defined in ExtendedProblemStat
    // _________________________________________________________________
    
    void addDirichletBC(WorldVector<double> *pos, int row, int col,
				AbstractFunction<double, WorldVector<double> > *fct);

    void addDirichletBC(DofIndex *idx, int row, int col,
				AbstractFunction<double, WorldVector<double> > *fct);

    void addDirichletBC(AbstractFunction<double, WorldVector<double> > *signedDist, int row, int col,
				AbstractFunction<double, WorldVector<double> > *fct);
    
    void addDirichletBC(DOFVector<double> *signedDist, int row, int col,
				AbstractFunction<double, WorldVector<double> > *fct);

    void addDirichletBC(BoundaryType nr, AbstractFunction<bool, WorldVector<double> >* meshIndicator,
			      int row, int col,
			      AbstractFunction<double, WorldVector<double> > *fct);

    void addDirichletBC(AbstractFunction<bool, WorldVector<double> >* meshIndicator,
			      int row, int col,
			      AbstractFunction<double, WorldVector<double> > *fct);

  protected:
    void init();
    
    void reset()
    {
      first = true;
      stageSolution->set(0.0);
      unVec->set(0.0);
      for (int i = 0; i < rm->getStages(); i++)
	stageSolutions[i]->set(0.0);
      
      invTauGamma = 1.0;
    }

  protected:    
    /// Pointer to the Rosenbrock method that should be used.
    RosenbrockMethod *rm;

    SystemVector *stageSolution, *unVec, *timeRhsVec, *newUn, *tmp, *lowSol, *timeDrivative;

    std::vector<SystemVector*> stageSolutions;

    int componentShift;
    int order;
    
    bool first;
    bool fixedTimestep;
    bool enforceAsmMat;

    double minusOne;

    double *tauPtr;
    double invTauGamma;
    
    double stageTime; // t_n + alpha_i*tau
    double oldTime;
    double tauGammaI; // tau*gamma_i

    std::vector<MyRosenbrockBoundary> boundaries;

    double errorEst;
    double oldErrorEst;
    std::vector<double> errorEstWeights;
  };


#endif // MY_ROSENBROCKSTATIONARY_H
